/* ***********************************************
*  Bootstrapping Johansen's Cointegration Test
*  Statistics and the LR Statistic for Over-
*  Identifying Restrictions
*  --------------------------------------------
*  Written by Matthew Greenwood Sept 2008
*  Email: mjgn@lubs.leeds.ac.uk
*  --------------------------------------------
*  NB - much code adapted from Yongcheol Shin's
*  Long-Run Structural Model programme (2007)
*********************************************** */

output file = coin_bs.out reset;
Screen on; cls; OUTPUT OFF;


@ ******************* @
@ Initial information @
@ ******************* @

stt0=2; @ Sample adjustment for VAR lag length @
porder=2; @ VAR lag length @
stt=porder;
pyorder=porder;
case=4;


@ **************** @
@ Read the dataset @
@ **************** @

data = xlsreadm("coin_bs","b2:k165",8,""); @ 1967q1-2007q4 @

@* Original Exogenous I(1) variables, Xo @
ys = data[.,2]; @ Real translog FIML pot output assuming 0 unemployment @
poil = data[.,1]; @ West Texas Intermediate oil price, FRED @

Xo = poil~ys;
dXo = Xo-lag(Xo);
mx = cols(Xo);

@* Original Endogenous I(1) variables, Yo @
y = data[.,10]; @ Real GDP @
rus = data[.,6]; @ Federal funds rate, IFS @
inf = data[.,5]; @ Log--approx to GDP deflator inflation, IFS @
winf = data[.,4]; @ Productivity adjusted wage inflation,  BLS and author's calculations @
dsc = data[.,7]; @ Real debt service cost, FoF, IFS @
tobq = data[.,3]; @ Tobin's q @
invest = data[.,9]; @ Gross investment of nonfinancial corporate sector, FoF @
cf = data[.,8]; @ Real cash-flow measured by book value of internal funds, FoF @

Yo = tobq~winf~inf~rus~dsc~cf~invest~y;
dYo = Yo-lag(Yo);
my = cols(Yo);

Zo = Xo~Yo;
dZo = Zo-lag(Zo);
mz = cols(Zo);

@* dimensions @
yorder = cols(Yo);  @ number of endo I(1) variables @
xorder = cols(Xo);  @ number of exo I(1) variables @


@ ******************* @
@ Read the parameters @
@ ******************* @

ALP = xlsreadm("coin_bs","b2:e11",1,"");
    AlpY = ALP[3:10,.];

beta = xlsreadm("coin_bs","b2:e12",2,"");
    Bbeta = beta[1:10,.];
    PAI = -ALP*Bbeta';
    PaiY = -AlpY*Bbeta';

gtmp = xlsreadm("coin_bs","b2:k11",3,"");
    gamX = gtmp[1:2,.];
    gamY = gtmp[3:10,.];

lam = xlsreadm("coin_bs","b2:c9",4,"");

b0 = xlsreadm("coin_bs","b2:b11",5,"");
    b0y = b0[3:10,.];

b1 = xlsreadm("coin_bs","b2:b11",6,"");
    b1y = b1[3:10,.];

a0 = xlsreadm("coin_bs","b2:k11",9,"");
    ia0 = inv(a0);
    a0y = eye(8);
    ia0y = eye(8);

sample = rows(data);
sample = sample-stt0;
inpt = ones(sample,1); tr1 = seqa(0,1,sample); tr = seqa(1,1,sample); 
inpt = inpt'; tr1 = tr1'; tr = tr';


@ *************************** @
@  Big loop for bootstrapping @
@ *************************** @

"Specify the number of bootstrap samples"; n_iter = con(1,1);

@ for saving the cointest and LR stats later @
EIGB = zeros(my,n_iter);
TRB = zeros(my,n_iter);
MAXB = zeros(my,n_iter);
LRB = zeros(1,n_iter);
BCFLAG = zeros(1,n_iter);

"Select the bootstrapping method:
1. Parametric   2. Non-parametric"; boot_opt = con(1,1);

nboot = 1;
do until nboot > n_iter;


@ ****************************************************** @
@ Whiten and resample the estimated residuals, eX and eY @
@ ****************************************************** @

@ Reading residuals exported from CVAR programme @
resids = xlsreadm("coin_bs","b2:k163",7,"");
omegay = xlsreadm("coin_bs","d4:k11",10,""); @ cov matrix @


if boot_opt eq 1; @ PARAMETRIC METHOD @

    p = chol(omegay);
    p = p';
    invp = inv(p);
    res = rndn(my,sample);
    ranseed = 1357;
    resp = p*res;
    ehaty = resp;

else; @ NON-PARAMETRIC METHOD @
    
    @ Whitening - DOES NOT ALLOW for PARAMETER UNCERTAINTY @
    p = chol(omegay);
    p = p';
    invp = inv(p);
    residw = invp*resids[.,mx+1:mz]';
    residw = residw';
    
    @ Resampling whitened residuals @
    rnum = rndu(sample,1)*(sample);
    ranseed = 1234;
    ehat = zeros(sample,my);
    ik = 1;
        do until ik > sample;
        ehat[ik,.] = residw[ceil(rnum[ik,.]),.];
        ik = ik+1;
        endo;
    
    @ De-whitening resampled residuals @
    ehatY = p*ehat';

endif;


@ ************************************************ @
@ Generate bootstrap samples assuming model is DGP @
@ ************************************************ @

zo1 = lag(zo);
dzb = zeros(sample+stt0,mz);
dyb = zeros(sample,my);
yb = zeros(sample,my);

ib = 1;
    do until ib > sample;
    if ib eq 1;
       dyb[ib,.] = b0y' + (lam*dXo[ib+stt0,.]')' + (gamy*dZo[ib+stt0-1,.]')' + (alpY*beta'(Zo1[ib+stt0,.]~tr1[.,ib])')' + ehatY[.,ib]';
       yb[ib,.]=yo[2,.]+dyb[ib,.];
    elseif ib eq 2;
       dyb[ib,.] = b0y' + (lam*dXo[ib+stt0,.]')' + (gamy*(dXo[ib+stt0-1,.]~dyb[1,.])')' + (alpY*beta'(Xo[ib+stt0-1,.]~Yb[1,.]~tr1[.,ib])')' + ehatY[.,ib]';
       yb[ib,.]=yo[2,.]+sumc(dyb[1:ib,.])';
    else; @ ib > 2 @
       dyb[ib,.] = b0y' + (lam*dXo[ib+stt0,.]')' + (gamy*(dXo[ib+stt0-1,.]~dyb[ib-1,.])')' + (alpY*beta'(Xo[ib+stt0-1,.]~Yb[ib-1,.]~tr1[.,ib])')' + ehatY[.,ib]';
       yb[ib,.]=yo[2,.]+sumc(dyb[1:ib,.])';
    endif;
    ib = ib+1;
endo;

/* *********************************
* In Yongcheol's CVAR prog:
* allY = alpY~gamY~lam~inpt
* Yreg = beta'(lag(z)|tr)|dz|dx|inpt
********************************* */

@ Rename bootstrap samples for consistency with CVAR programme @
x=xo;
y=zeros(sample+stt0,my);
y[1:2,.]=yo[1:2,.];
y[3:sample+stt0,.]=yb;
z=xo~y;
Xdata = X; Ydata = Y; @ Used for forecasting and bootstrap @


@ *******************
* DATA TRANSFORMATION
* *******************
* dimensions of data matrices:
* DY: my by T, DX: mx by T, Z1: m by T, ADZLAGS: m(maxp-1) by T
* ************************************************************* @
{ DY, Z1, DX, ADZLAGS } = TRDATA(Y,X,yorder,xorder);

m = xorder + yorder; p = pyorder;
stt1 = 0; @ chk: remove it later @
@* sample adjustment @
DY = DY[.,stt+1:cols(DY)-stt1];
Z1 = Z1[.,stt+1:cols(Z1)-stt1];

if xorder eq 0;
else; DX = DX[.,stt+1:cols(DX)-stt1];
endif;
ActDX = DX; @ actual sample of DX for bootstap, where to put?? @

ADZLAGS = ADZLAGS[.,stt+1:cols(ADZLAGS)-stt1];
if p eq 1;
else; DZLAGS = ADZLAGS[1:m*(p-1),.];
endif;

@ **********************************
* CONSTRUCTION of ECM TERMS, ZS1 and 
* all other regressors, YY
  ********************************** @
@* Break dummies: provide information on break point @ 
@ if not, the default is the mid point @
brpoint = ceil(sample/2); 
brpoint = 138; @ in the original data: 2001q1 - greenspan gambit @

br = brpoint-stt; @ break point adjusted in sample deleting initial obs @
iduma = zeros(br-1,1);
idumb = ones(sample-br+1,1);
idum = iduma|idumb;
tduma = zeros(br-1,1);
tdumb = seqa(1,1,sample-br+1);
tdum = tduma|tdumb;

{ idum1, adidum, tdum1, adtdum} = TRDATA2(idum,tdum,stt);

idum1 = idum1[.,stt+1:cols(idum1)];
tdum1 = tdum1[.,stt+1:cols(tdum1)];
adidum = adidum[.,stt+1:cols(adidum)];
adtdum = adtdum[.,stt+1:cols(adtdum)];
didum = adidum[1:p,.]; @ including the contemporaneous one @
dtdum = adtdum[1:p,.]; @ including the contemporaneous one @

{ ZS1, YY } = CREG(case,Z1,inpt,TR1,idum1,tdum1,porder,
              DZlags,xorder,DX,TR,didum,dtdum);


@ ********************************** @
@ COINTEGRATION TEST & Johansen beta @
@ ********************************** @
{ COTEST, JBETA } = COINTEST(DY,ZS1,YY,yorder,xorder);

/*
@* printing and exporting cointegration test results @
output on;"";"COINTEGRATION TEST RESULTS:";"";
"eigenval   trace     max"; format /rd 8,4; COTEST;"";
*/

@* exactly identified (eigen) betas for r = 1,...,10: ms by r @
ms = rows(Jbeta);
E10beta = Jbeta[.,ms-9:ms];
E9beta = Jbeta[.,ms-8:ms];
E8beta = Jbeta[.,ms-7:ms];
E7beta = Jbeta[.,ms-6:ms];
E6beta = Jbeta[.,ms-5:ms];
E5beta = Jbeta[.,ms-4:ms];
E4beta = Jbeta[.,ms-3:ms];
E3beta = Jbeta[.,ms-2:ms];
E2beta = Jbeta[.,ms-1:ms];
E1beta = Jbeta[.,ms];

@* error correction terms: r by T @
E1xi = E1beta'ZS1; @ 1 by T @
E2xi = E2beta'ZS1; @ 2 by T @
E3xi = E3beta'ZS1; @ 3 by T @
E4xi = E4beta'ZS1; @ 4 by T @
E5xi = E5beta'ZS1; @ 5 by T @
E6xi = E6beta'ZS1; @ 6 by T @
E7xi = E7beta'ZS1; @ 7 by T @
E8xi = E8beta'ZS1; @ 8 by T @
E9xi = E9beta'ZS1; @ 9 by T @
E10xi = E10beta'ZS1; @ 10 by T @

rcoin = 4; @ # of cointegrating relations @

@* estimation of alpha and omega given Jbeta and max likelihood @
{ Jalpha, Jomega, JMLL } = AOest(E4beta,DY,ZS1,YY,yorder,sample,rcoin);

/*
"Johansen RRR estimates";
"beta"; E4beta; 
"alpha"; Jalpha;
"omega"; Jomega;
"maximum likelihood"; JMLL;
*/


@ *************************************************************
* Estimation of cointegrating vectors subject to non-homogenous 
* over-identifying restrictions using BS and NR algorithms
* *************************************************************
* Rmatrix*vec(Beta) = bvec
* where dimR = number of restrictions
*       DOF = dimR - rcoin^2: degrees of freedom
* You should specify these restrictions prior to estimation
* which are saved in procedure, RESTR1, RESTR2, ... .
* ********************************************************** @

@ m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           1    0    *    0      yscobb3
           0    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    0    *    0      t
* A total of 33 restrictions, so 17 over-identifying restrictions
* Later consider co-trending and co-breaking  @
        
kk = 33;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 33 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  bvec[10] = 0;
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;
Rmatrix[12,ms+4] = 1;  bvec[12] = 0;   
Rmatrix[13,ms+6] = 1;  bvec[13] = -1; 
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;
Rmatrix[18,ms+11] = 1;  bvec[18] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[19,2*ms+1] = 1;  bvec[19] = 0;  
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[24,3*ms+1] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+2] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+3] = 1;  bvec[26] = 0;
Rmatrix[27,3*ms+4] = 1;  bvec[27] = 1;
Rmatrix[28,3*ms+5] = 1;  bvec[28] = -1;  
Rmatrix[29,3*ms+7] = 1;  bvec[29] = 0; 
Rmatrix[30,3*ms+8] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+9] = 1;  bvec[31] = 0;
Rmatrix[32,3*ms+10] = 1;  bvec[32] = 0;
Rmatrix[33,3*ms+11] = 1;  bvec[33] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

@* data transformation @
MYY = eye(sample) - YY'invpd(YY*YY')*YY; @ idempotent @
vDY = vec(DY');


@ **************************** 
* Choice of initial estimates:
* **************************** @
{ beta0 } = INICOIN(ms,rcoin,E4beta,Rmatrix,bvec); @ ms by r @
/* beta0 = E4beta; */
@* estimation of alpha and omega given initial beta and MLL @
{ alpha0, omega0, MLL0 } = AOest(beta0,DY,ZS1,YY,yorder,sample,rcoin);

/*
"Initial estimates";
"beta"; beta0; 
"alpha"; alpha0;
"omega"; omega0;
"maximum likelihood"; MLL0;
*/

JO4beta = beta0;
JO4xi = JO4beta'ZS1; @ 4 by T @


@ ***********
* BS Algorithm
* *********** @
@ Tolerance level used in convergence criteria @
tol1 = 10^(-4);   @ check if Like(i)-Like(i-1)<tol1 @
tol2 = 10^(-4)*(ms*rcoin-dimR); @ check if |Theta1(i)-Theta1(i-1)|< tol2 @

maxiter = 5000; @ max number of iterations @
ij = 1;
do until ij > maxiter;

@* estimation of beta subject to restrictions @
   { beta1, U11 } = Best(alpha0,Omega0,DY,ZS1,YY,ms,rcoin,Rmatrix,bvec);


@ *********************************
* check the collinearity of beta's
* ********************************* @
   fflag = 0;
   if rank(beta1'ZS1*MYY*ZS1'beta1) < rcoin; fflag = 1;
      "beta's are collinear";
   endif;

@* estimation of alpha and omega given updated beta and MLL @
   { alpha1, omega1, MLL1 } = AOest(beta1,DY,ZS1,YY,yorder,sample,rcoin);

@* check if the convergence occurred @
   convt = abs(vec(beta1)-vec(beta0)); convtr = sumc(convt);
   convlike = MLL1-MLL0;

   cls; screen on; output off; format /rd 12,8;
   "Number of bootstrap iterations:";; nboot;
   "Number of Iterations:";; ij;
   "Restricted MLL: ";;MLL1;
   "MLL-MLL(-1):";;convlike;


   cflag = 0;
   if    convlike < tol1;
      if    convtr < tol2;  cflag = 1;
            @  output on; "converged after " ij " iterations"; @
            break;  @ convergence occurred @
      endif;
   endif;

@ update @
   alpha0 = alpha1; Omega0 = Omega1; beta0 = beta1; MLL0 = MLL1;

   ij = ij + 1;"";
endo;

/*
output on; format /rd 8,4;
"Converged estimates";
"beta"; beta1; 
"alpha"; alpha1;
"omega"; omega1;
"maximum likelihood"; MLL1;
"LR test"; 2*(JMLL-MLL1);
*/

O4beta = beta1;
O4xi = O4beta'ZS1; @ 4 by T @


@ ************************************************** @
@ save the cointest and LR test results if converged @
@ ************************************************** @

BCFLAG[.,nboot] = cflag;

if cflag eq 1;
    EIGB[.,nboot] = cotest[.,1];
    TRB[.,nboot] = cotest[.,2];
    MAXB[.,nboot] = cotest[.,3];
    LRB[.,nboot] = 2*(jmll-mll1);
else;
    EIGB[.,nboot] = zeros(my,1);
    TRB[.,nboot] = zeros(my,1);
    MAXB[.,nboot] = zeros(my,1);
    LRB[.,nboot] = 0;
endif;


@ ******************** @
@ Closing the big loop @
@ ******************** @

nboot = nboot + 1;
endo;


@ **************************************************** @
@  Export to Excel for computation of critical values  @
@ **************************************************** @

"Average of the LR statistics:"; sumr(LRB)/sumr(BCFLAG);
"Total number of converged estimates:"; sumr(bcflag);
"Convergence rate (%):"; 100*sumr(bcflag)/(n_iter);
"Average trace and max statistics:"; (sumr(TRB)/sumr(BCFLAG))~(sumr(MAXB)/sumr(BCFLAG));

@ Eigenvalues @
ac1 = {"EIG"};
acc1 = xlswritesa(ac1, "coinboot", "a1", 1, "");
accc1 = xlswritem(eigb', "coinboot", "b2", 1, "");

@ Trace Statistic @
ac2 = {"TRACE"};
acc2 = xlswritesa(ac2, "coinboot", "a1", 2, "");
accc2 = xlswritem(trb', "coinboot", "b2", 2, "");

@ Maximal Eigenvalue @
ac3 = {"MAX"};
acc3 = xlswritesa(ac3, "coinboot", "a1", 3, "");
accc3 = xlswritem(maxb', "coinboot", "b2", 3, "");

@ LR Statistic @
ac4 = {"LR"};
acc4 = xlswritesa(ac4, "coinboot", "a1", 4, "");
accc4 = xlswritem(lrb', "coinboot", "b2", 4, "");

end;


@* SUB PROCEDURES *@

@ ************************************* @
  proc (4) = TRDATA(Y,X,yorder,xorder);
@ ***************************************************************
*   Generate the appropriate Data matrices of DY, DX, Z1, DZLAGS,
*   to be used in the VEC model, depending on the order of VAR(p)
*   DY: First difference of endogenous I(1) variables
*   DX: First difference of exogenous I(1) variables
*   Z = X~Y
*   Z1: First lagged values of Z
*   ADZLAGS = DZ1~DZ2~...~DZ(p-1): T by m(p-1)
*   ************************************************************** @
local  MM,Y1,DY,Z,Z1,adzlags,dX,DZ,DZ1,DZ2,DZ3,DZ4,DZ5,DZ6,DZ7,DZ8,
       DZ9,DZ10,DZ11;

mm = yorder + xorder;
Y1 = LAG(Y); DY = Y - Y1;

if Xorder eq 0;   Z = Y; DX = 0; @ No exog variables: Txmy @
else;             Z = X~Y;       @ Txm @
                  DX = X - LAG(X);
endif;

Z1 = LAG(Z);     DZ = Z - Z1;    DZ1 = Lag(DZ);  DZ2 = Lag(DZ1);
DZ3 = Lag(DZ2);  DZ4 = Lag(DZ3); DZ5 = Lag(DZ4); DZ6 = Lag(DZ5);
DZ7 = Lag(DZ6);  DZ8 = Lag(DZ7); DZ9 = Lag(DZ8); DZ10 = Lag(DZ9);
DZ11 = Lag(DZ10);

@ all lagged values up to maximum lag order @
ADZLAGS = DZ1~DZ2~DZ3~DZ4~DZ5~DZ6~DZ7~DZ8~DZ9~DZ10~DZ11;

@* transpose the data for matching the specification *@
DY = DY';  DX = DX';  Z1 = Z1'; ADZLAGS = ADZLAGS';

retp(DY,Z1,DX,ADZLAGS);endp;

@ ********************************* @
  proc (4) = TRDATA2(idum,tdum,stt);
@ ********************************* @
@ idum1: lagged intercept shift dummy
* tdum1: lagged trend shift dummy
* ADidum = Didum~Didum1~...~Didum(p-1)
* ADtdum = Dtdum~Dtdum1~...~Dtdum(p-1)
* adding initial stt observations
* ************************************ @
local  stmp,idum1,Di0,Di1,Di2,Di3,Di4,Di5,Di6,Di7,Di8,Di9,Di10,Di11,
       tdum1,Dt0,Dt1,Dt2,Dt3,Dt4,Dt5,Dt6,Dt7,Dt8,Dt9,Dt10,Dt11;

stmp = zeros(stt,1);
idum = stmp|idum; tdum = stmp|tdum;

idum1 = LAG(idum); 
Di0 = idum - idum1; Di1 = Lag(Di0); Di2 = Lag(Di1); Di3 = Lag(Di2); 
Di4 = Lag(Di3); Di5 = Lag(Di4); Di6 = Lag(Di5); Di7 = Lag(Di6);  
Di8 = Lag(Di7); Di9 = Lag(Di8); Di10 = Lag(Di9); Di11 = Lag(Di10);
ADidum = Di0~Di1~Di2~Di3~Di4~Di5~Di6~Di7~Di8~Di9~Di10~Di11;

tdum1 = LAG(tdum); 
Dt0 = tdum - tdum1; Dt1 = Lag(Dt0); Dt2 = Lag(Dt1); Dt3 = Lag(Dt2); 
Dt4 = Lag(Dt3); Dt5 = Lag(Dt4); Dt6 = Lag(Dt5); Dt7 = Lag(Dt6);  
Dt8 = Lag(Dt7); Dt9 = Lag(Dt8); Dt10 = Lag(Dt9); Dt11 = Lag(Dt10);
ADtdum = Dt0~Dt1~Dt2~Dt3~Dt4~Dt5~Dt6~Dt7~Dt8~Dt9~Dt10~Dt11;

@* transpose the data for matching the specification *@
idum1 = idum1'; tdum1 = tdum1';
ADidum = ADidum'; ADtdum = ADtdum';

retp(idum1,ADidum,tdum1,ADtdum);endp;

@ ********************************************* @
  proc (3) = AOest(rbeta,DY,ZS1,YY,m,sample,r);
@ ********************************************* @
local  Ralpha,Romega,MYY,vDY,bzb,ibzb,bz,rkapa,rPAI,E0,MLL;

MYY = eye(sample) - YY'invpd(YY*YY')*YY; @ idempotent @
vDY = vec(DY');
bzb = rbeta'ZS1*MYY*ZS1'rbeta;
ibzb = inv(bzb); @ invpd ?? @
bz = rbeta'ZS1*MYY;
rkapa = eye(m).*.(ibzb*bz)*vDY;

ralpha = TRMAT(rkapa,r,m); @ r by m: @
ralpha = ralpha'; @ since kapa = vec(alpha') @

rPAI = ralpha*rbeta'; 
E0 = DY' - ZS1'rPAI';
rOmega = E0'MYY*E0./sample;

MLL = -(m*sample/2)*(1+ln(2*pi))-(sample/2)*ln(det(romega));

retp(Ralpha,Romega,MLL);endp;

@ ******************************************************** @
  proc (2) = Best(Ralpha,ROmega,DY,ZS1,YY,ms,r,Rmat,bvec);
@ ******************************************************** @
@* estimation of beta subject to restrictions @
local  Rbeta,U11,MYY,vDY,aoa,z1Mz1,Atheta,aom,z1M,Btheta,Jtheta,IJth,
       rth,rtheta;

MYY = eye(sample) - YY'invpd(YY*YY')*YY; @ idempotent @
vDY = vec(DY');
aoa = Ralpha'ROmega*Ralpha;
z1Mz1 = ZS1*MYY*ZS1';
Atheta = aoa.*.z1Mz1;
aom = Ralpha'ROmega;
z1M = ZS1*MYY;
Btheta = aom.*.z1M;
Jtheta = Atheta + Rmat'Rmat;
IJth = inv(Jtheta);
rth = IJth*Btheta*vDY;
rtheta = rth - IJth*Rmat'inv(Rmat*IJth*Rmat')*(Rmat*rth-bvec); 
rbeta = TRMAT(rtheta,ms,r); @ ms by r @
U11 = IJth - IJth*Rmat'Inv(Rmat*IJth*Rmat')*Rmat*IJth;

retp(Rbeta,U11);endp;

@ **************************************** @
  proc (1) = INICOIN(ms,r,bet,Rmat,bvec);
@ ****************************************
* Initial estimates of restricted cointegrating vectors:
* Transform Johansen beta using the whole restrictions
* ****************************************************** @
local  Q1,Q2,Ir,IQ2,theta0,beta0;

Ir = eye(rcoin);
Q1 = Ir.*.bet; Q2 = Rmat*Q1;
IQ2 = INV(Q2'Q2);
theta0 = Q1*IQ2*Q2'bvec;
beta0 = TRMAT(theta0,ms,rcoin); @ ms by r @

retp(beta0); endp;

@ ***************************************** @
  proc (2) = COINTEST(DY,Z1,YY,my,mx);
@ *****************************************
* (1) OLS of DY on YY to obtain R0 and then calculate
*     S00 = (1/T)R0*R0'
* (2) OLS of Z1 on YY to obtain R1 and then calculate
*     S11 = (1/T)R1*R1',      S01 = (1/T)R0*R1'
* (3) Then, calculate
*     Bmatirx = S11
*     Amatrix = S11 - S01'Invpd(S00)*S01
* for further details see Johansen (1988, 1991) and PSS (2000)
* ************************************************************* @
local  sample,URPAI,dreg,invYY,invS00,Amatrix,Bmatrix,
       S00,S01,S11,C0,C1,R0,R1,zorder,testres,
       pp,va,ve,dd,zi,sumva,maxva,Ltrace,Lmax,sumv,cvmax,cvtrace,
       vva,vaa,im1,im2,jbeta,cc;
sample = cols(dy);

@* Z1 = (Y1 X1) and any deterministic regressors @
zorder = rows(z1);  @ dimension of z1 @

@* Computation of S00, S11 and S10 @
if YY eq 0; R0 = DY; R1 = Z1;
else;
     InvYY = Invpd(YY*YY');
     C0 = DY*YY'InvYY;
     R0 = DY - C0*YY;
     C1 = Z1*YY'InvYY;
     R1 = Z1 - C1*YY;
endif;

S00 = R0*R0'./sample;
S11 = R1*R1'./sample;
S01 = R0*R1'./sample;

/* CC = invpd(S11)*s01'invpd(S00)*S01; save cc; */

@* Computation of Amatrix and Bmatrix @
InvS00 = invpd(S00);
Amatrix = S11 - S01'InvS00*S01;
Bmatrix = S11;

@* Unrestricted pi matrix @
URPAI = S01*invpd(S11);

@* Johansen Beta's using eigenvalue routine @
pp = chol(S11);
dd = inv(pp')*S01'invS00*S01*inv(pp);
{va,ve} = EIGHV(dd);

Jbeta = zeros(zorder,zorder);
zi = 1;
do until zi > zorder;
   Jbeta[.,zi] = inv(pp)*ve[.,zi];
   zi = zi + 1;
endo;

vva = zeros(zorder,1);
zi = 1;
do until zi > zorder;
   vva[zi,1] = va[zorder-zi+1,1];
   zi = zi + 1;
endo;

@* Trace and the Max Eigenvalue statistics @
vaa = zeros(zorder,1);
sumv = 0;
zi = 1;
do until zi > zorder;
   vaa[zi,1] = ln(1 - va[zi,1]);
   sumv = sumv + vaa[zi,1];
   zi = zi + 1;
endo;

sumva = zeros(my,1);
maxva = zeros(my,1);
sumva[1,1] = sumv;
maxva[1,1] = ln(1 - va[zorder,1]);
zi = 2;
do until zi > my;
   im1 = zorder + 2 - zi;
   sumva[zi,1] = sumva[zi-1,1] - vaa[im1,1];
   im2 = zorder + 1 - zi;
   maxva[zi,1] = ln(1 - va[im2,1]);
   zi = zi + 1;
endo;

Ltrace = zeros(my,1);  Lmax = zeros(my,1);
zi = 1;
do until zi > my;
   Ltrace[zi,1] = -sample*sumva[zi,1];
   Lmax[zi,1] = -sample*maxva[zi,1];
   zi = zi + 1;
endo;

VVA = VVA[1:my];
TESTRES = VVA~LTRACE~LMAX;

retp(TESTRES,JBETA);endp;

@ ***************************** @
  proc (1) = TRMAT(vect,ro,co);
@ ***************************** 
* Converting nm by 1 vector to n by m matrix 
* ****************************************** @
local matri,smj,rj,cj;

matri = zeros(ro,co);
smj = 1;
do until smj > co;
   rj = (smj-1)*ro + 1;
   cj = smj*ro;
   matri[.,smj] = vect[rj:cj];
   smj = smj + 1;
endo;

retp(Matri);endp;

@ *********************************************************** @
  proc (2) 
  = CREG(case,Z1,inpt,TR1,id1,td1,p,DZlags,mx,DX,TR,did,dtd);
@ *********************************************************** 
* CONSTRUCTION of ECM term, ZS1 and all other regressors, YY 
* ********************************************************** @
local zs1,yy;

if     case eq 1; ZS1 = Z1; 
elseif case eq 2; ZS1 = Z1|inpt;
elseif case eq 3; ZS1 = Z1;
elseif case eq 4; ZS1 = Z1|TR1;
elseif case eq 6; ZS1 = Z1|TR1|id1;
elseif case eq 7; ZS1 = Z1|TR1|td1;
endif;

if p eq 1; YY = 0;
else; YY = DZLAGS;
endif;

if YY eq 0;
   if mx eq 0;
   else; YY = DX;
   endif;
else;
   if mx eq 0;
   else; YY = YY|DX;
   endif;
endif;

if YY eq 0;
   if     case eq 3; YY = inpt;
   elseif case eq 4; YY = inpt;
   elseif case eq 5; YY = inpt|tr;
   elseif case eq 6; YY = inpt|did;
   elseif case eq 7; YY = inpt|dtd;
   endif;
else;
   if     case eq 3; YY = YY|inpt;
   elseif case eq 4; YY = YY|inpt;
   elseif case eq 5; YY = YY|inpt|tr;
   elseif case eq 6; YY = YY|inpt|did;
   elseif case eq 7; YY = YY|inpt|dtd; @ ?? @
   endif;
endif;

retp(ZS1,YY);endp;