/* ************************************************
*  Long-run Structural Macroeconometric Modelling:
*  An Application to KOREAN ECONOMY
*  --------------------------------
*  Written by Yongcheol Shin. Last updated May 2007
*  Contact to Y.SHIN@LEEDS.AC.UK
*  -------------------------------------------------
*  See Pesaran and Shin (2003, ER) and Garatt, Lee, Pesaran and Shin (2006, OUP)
*
*  This program provides estimation of both long-run cointegration 
*  and further short-run dynamics of vector error correction model.
*
*  This program includes:
*
*  A: The initial step loads and defines the data. It also
*     specifies some initial informations necessary for running 
*     the rest of the program.
*
*  B: The long-run cointegration test and estimation results including 
*     the cases with both exactly identified and over identified 
*     cointegrating relationships. In the latter case the LR tests
*     for the validity of LR theory restrictions being imposed. 
*
*  C: Further short-run dynamics:
*
*     a: VEC equations estimation results with diagnostic tests
*     b: Stability of the full VEC system
*     c: Persistence profiles (speed of convergence of LR equilibria
*     d: Generalized and orthogonalised impulse response functions
*     e: Structural impulse response functions
*     f: Permanent and transitory decompositions
*     g: Point (dynamic) forecasts with CIs
*     h: Scenario forecast analysis
*     i: In-sample forecast evaluation
*     j: Out-of-sample probability forecast
*
*  See Pesaran and Shin (1996, JoE), Pesaran and Shin (1998, EL),
*  GLPS (2003a,b, FJ, JASA), GLPS (2006, OUP)
*
*  This program only allows for future uncertainty.
* 
*  The companion program provides an option to compute the empirical
*  distributions of all statistics calculated above based on both parameter 
*  and future uncertainty.
*
*  Another companion program will also handle the model uncertainty in the 
*  evaluation of forecasting.
*
*  After running the program, you will be promted to select several options
*  appearing on screen.
**************************************************************************** */
new;  library pgraph;  graphset;

@* Specify the output file @

output file = cvarx.out reset;

/* ***************************************
*  STEP 1: DATA LOADING and Transformation
*  *************************************** */
@* Reading Excel file @
us1 = xlsreadm("data","b2:x165",1,""); @ 1967q1-2007q4 @

@ Definition of data
* 1967q1-2007q4, 164 quarterly observations: in order of 
* RYI,RYSC1I,RYSC2I,RYSC3I,RYST1OI,RYST2OI,RYST3OI,RYST1FI,RYST2FI,RYST3FI,RCFI,
* NCFI,TOBQI,RINI,RDSCI,NDSCI,RPOI,INF,INF2,WINF,FFQ
* Base year 2000=100, all vars are natural logs @

@* Exogenous I(1) variables, X @
ys = us1[.,10]; @ Real translog FIML pot output assuming 0 unemployment @
poil = us1[.,17]; @ West Texas Intermediate oil price, FRED @

X = poil~ys;

@* Endogenous I(1) variables, Y. @
y = us1[.,1]; @ Real GDP @
rus = us1[.,21]; @ Federal funds rate, IFS @
inf = us1[.,19]; @ Log--approx to GDP deflator inflation, IFS @
winf = us1[.,20]; @ Productivity adjusted wage inflation,  BLS and author's calculations @
dsc = us1[.,23]; @ Real debt service cost, FoF, IFS @
tobq = us1[.,13]; @ Tobin's q @
invest = us1[.,14]; @ Gross investment of nonfinancial corporate sector, FoF @
cf = us1[.,11]; @ Real cash-flow measured by book value of internal funds, FoF @

Y = tobq~winf~inf~rus~dsc~cf~invest~y; 

@* dimensions @
yorder = cols(y);  @ number of endo I(1) variables @
xorder = cols(x);  @ number of exo I(1) variables @

Xdata = X; Ydata = Y; @ Used for forecating and bootstrap @

Screen on; cls; OUTPUT OFF;
@* Type the data loading information below @
"You have just loaded the data with the following information:

   X =  poil~ys: 2 exo I(1) variables,
   y = tobq~winf~inf~rus~dsc~cf~invest~y, 8 endo I(1) variables,

In what follows provide the necessary information when prompted.";
"press any key"; wait;

@ *******************************************************************
* Some preliminary information are necessary for running the program, 
* but you may also be prompted to provide these on screen.
* ******************************************************** @

"Choose the VAR lag order (<=12)"; porder = con(1,1);
@ porder: VAR LAG ORDER <=12 @
porder = 2; pyorder = porder;
@ number of initial values discarded for VAR(2) @
stt = porder; @ currently 67q3-07q4, 162 obs @
@ you may change the value of stt @

@ *******************
* DATA TRANSFORMATION
* *******************
* dimensions of data matrices:
* DY: my by T, DX: mx by T, Z1: m by T, ADZLAGS: m(maxp-1) by T
* ************************************************************* @
{ DY, Z1, DX, ADZLAGS } = TRDATA(Y,X,yorder,xorder);

m = xorder + yorder; p = pyorder;
stt1 = 0; @ chk: remove it later @
@* sample adjustment @
DY = DY[.,stt+1:cols(DY)-stt1];
Z1 = Z1[.,stt+1:cols(Z1)-stt1];

if xorder eq 0;
else; DX = DX[.,stt+1:cols(DX)-stt1];
endif;
ActDX = DX; @ actual sample of DX for bootstap, where to put?? @

ADZLAGS = ADZLAGS[.,stt+1:cols(ADZLAGS)-stt1];
if p eq 1;
else; DZLAGS = ADZLAGS[1:m*(p-1),.];
endif;

/*
DY = DY[.,stt+1:cols(DY)-stt1];
Z1 = Z1[.,stt+1:cols(Z1)-stt1];

if xorder eq 0;
else; DX = DX[.,stt+1:cols(DX)-stt1];
endif;
ActDX = DX; @ actual sample of DX for bootstap @

ADZLAGS = ADZLAGS[.,stt+1:cols(ADZLAGS)-stt1];
if p eq 1;
else; DZLAGS = ADZLAGS[1:m*(p-1),.];
endif;
*/

sample = cols(dy); @ No. of observations @

@ *******************************************
* STEP 2: THE LONG-RUN COINTEGRATION ANALYSIS
* *******************************************
* We assume that you have already estimated ML estimates of
* the cointegrating vectors which are obtained via e.g. Microfit
* Then, you must provide this information below?? @

/* There are 7 cases depending on deterministic components:
  1. No intercepts and no trends.
  2. Restricted intercepts and no trends.
  3. Unrestricted intercepts and no trends.
  4. Unrestricted intercepts and restricted trends.
  5. Unrestricted intercepts and unrestricted trends.
  6. Case 4 plus restricted intercept shift dummies.
  7. Case 4 plus restricted trend shift dummies. */
case = 4; @ consider both cases 4 and 6. NB - case 6 nests case 4 as a special case @

@ **********************************
* CONSTRUCTION of ECM TERMS, ZS1 and 
* all other regressors, YY
  ********************************** @
inpt = ones(sample,1); tr1 = seqa(0,1,sample); tr = seqa(1,1,sample); 
inpt = inpt'; tr1 = tr1'; tr = tr';

@* Break dummies: provide information on break point @ 
@ if not, the default is the mid point @
brpoint = ceil(sample/2); 
brpoint = 138; @ in the original data: 2001q1 - greenspan gambit @

br = brpoint-stt; @ break point adjusted in sample deleting initial obs @
iduma = zeros(br-1,1);
idumb = ones(sample-br+1,1);
idum = iduma|idumb;
tduma = zeros(br-1,1);
tdumb = seqa(1,1,sample-br+1);
tdum = tduma|tdumb;

{ idum1, adidum, tdum1, adtdum} = TRDATA2(idum,tdum,stt);

idum1 = idum1[.,stt+1:cols(idum1)];
tdum1 = tdum1[.,stt+1:cols(tdum1)];
adidum = adidum[.,stt+1:cols(adidum)];
adtdum = adtdum[.,stt+1:cols(adtdum)];
didum = adidum[1:p,.]; @ including the contemporaneous one @
dtdum = adtdum[1:p,.]; @ including the contemporaneous one @

{ ZS1, YY } = CREG(case,Z1,inpt,TR1,idum1,tdum1,porder,
              DZlags,xorder,DX,TR,didum,dtdum);

@ You may type for YY below @
/* YY = DZLAGS|DX|inpt */

@ ********************************** @
@ COINTEGRATION TEST & Johansen beta @
@ ********************************** @
{ COTEST, JBETA } = COINTEST(DY,ZS1,YY,yorder,xorder);

@* printing and exporting cointegration test results @
output on;"";"COINTEGRATION TEST RESULTS:";"";
"eigenval   trace     max"; format /rd 8,4; COTEST;"";
"press any key"; wait;

@* Exporting the cointegration test results to Excel file @
out1 = xlswritem(COTEST,"cointest","b2",1,"");

@* Headers for Cointegration Test in US1out @
aa1 = {"cointest" "eigenval" "LRtrace" "LRmax"};
aa2 = {"r<=0", "r<=1", "r<=2", "r<=3", "r<=4", "r<=5", "r<=6", "r<=7", "r<=8", "r<=9", "r<=10"};
aaa1 = xlswritesa(aa1, "cointest", "a1", 1,"");
aaa2 = xlswritesa(aa2, "cointest", "a2", 1,"");
out1 = xlswritem(COTEST,"cointest","b2",1,"");

@* exactly identified (eigen) betas for r = 1,...,10: ms by r @
ms = rows(Jbeta);
E10beta = Jbeta[.,ms-9:ms];
E9beta = Jbeta[.,ms-8:ms];
E8beta = Jbeta[.,ms-7:ms];
E7beta = Jbeta[.,ms-6:ms];
E6beta = Jbeta[.,ms-5:ms];
E5beta = Jbeta[.,ms-4:ms];
E4beta = Jbeta[.,ms-3:ms];
E3beta = Jbeta[.,ms-2:ms];
E2beta = Jbeta[.,ms-1:ms];
E1beta = Jbeta[.,ms];

@* error correction terms: r by T @
E1xi = E1beta'ZS1; @ 1 by T @
E2xi = E2beta'ZS1; @ 2 by T @
E3xi = E3beta'ZS1; @ 3 by T @
E4xi = E4beta'ZS1; @ 4 by T @
E5xi = E5beta'ZS1; @ 5 by T @
E6xi = E6beta'ZS1; @ 6 by T @
E7xi = E7beta'ZS1; @ 7 by T @
E8xi = E8beta'ZS1; @ 8 by T @
E9xi = E9beta'ZS1; @ 9 by T @
E10xi = E10beta'ZS1; @ 10 by T @

@* Over identified (theory) betas for r = 3-5: later?? @
rcoin = 4; @ # of cointegrating relations @

@* estimation of alpha and omega given Jbeta and max likelihood @
{ Jalpha, Jomega, JMLL } = AOest(E4beta,DY,ZS1,YY,yorder,sample,rcoin);
"Johansen RRR estimates";
"beta"; E4beta; 
"alpha"; Jalpha;
"omega"; Jomega;
"maximum likelihood"; JMLL;

@ *************************************************************
* Estimation of cointegrating vectors subject to non-homogenous 
* over-identifying restrictions using BS and NR algorithms
* *************************************************************
* Rmatrix*vec(Beta) = bvec
* where dimR = number of restrictions
*       DOF = dimR - rcoin^2: degrees of freedom
* You should specify these restrictions prior to estimation
* which are saved in procedure, RESTR1, RESTR2, ... .
* ********************************************************** @

"Choose the restriction option (1-34)"; rest_op = con(1,1);

if rest_op eq 1; 
   { Rmatrix, bvec }= RESTR1(ms,rcoin);
elseif rest_op eq 2; 
   { Rmatrix, bvec }= RESTR2(ms,rcoin);
elseif rest_op eq 3; 
   { Rmatrix, bvec }= RESTR3(ms,rcoin);
elseif rest_op eq 4; 
   { Rmatrix, bvec }= RESTR4(ms,rcoin);
elseif rest_op eq 5; 
   { Rmatrix, bvec }= RESTR5(ms,rcoin);
elseif rest_op eq 6; 
   { Rmatrix, bvec }= RESTR6(ms,rcoin);
elseif rest_op eq 7; 
   { Rmatrix, bvec }= RESTR7(ms,rcoin);
elseif rest_op eq 8; 
   { Rmatrix, bvec }= RESTR8(ms,rcoin);
elseif rest_op eq 9; 
   { Rmatrix, bvec }= RESTR9(ms,rcoin);
elseif rest_op eq 10; 
   { Rmatrix, bvec }= RESTR10(ms,rcoin);
elseif rest_op eq 11; 
   { Rmatrix, bvec }= RESTR11(ms,rcoin);
elseif rest_op eq 12; 
   { Rmatrix, bvec }= RESTR12(ms,rcoin);
elseif rest_op eq 13; 
   { Rmatrix, bvec }= RESTR13(ms,rcoin);
elseif rest_op eq 14; 
   { Rmatrix, bvec }= RESTR14(ms,rcoin);
elseif rest_op eq 15; 
   { Rmatrix, bvec }= RESTR15(ms,rcoin);
elseif rest_op eq 16; 
   { Rmatrix, bvec }= RESTR16(ms,rcoin);
elseif rest_op eq 17; 
   { Rmatrix, bvec }= RESTR17(ms,rcoin);
elseif rest_op eq 18; 
   { Rmatrix, bvec }= RESTR18(ms,rcoin);
elseif rest_op eq 19; 
   { Rmatrix, bvec }= RESTR19(ms,rcoin);
elseif rest_op eq 20; 
   { Rmatrix, bvec }= RESTR20(ms,rcoin);
elseif rest_op eq 21; 
   { Rmatrix, bvec }= RESTR21(ms,rcoin);
elseif rest_op eq 22; 
   { Rmatrix, bvec }= RESTR22(ms,rcoin);
elseif rest_op eq 23; 
   { Rmatrix, bvec }= RESTR23(ms,rcoin);
elseif rest_op eq 24; 
   { Rmatrix, bvec }= RESTR24(ms,rcoin);
elseif rest_op eq 25; 
   { Rmatrix, bvec }= RESTR25(ms,rcoin);
elseif rest_op eq 26; 
   { Rmatrix, bvec }= RESTR26(ms,rcoin);
elseif rest_op eq 27; 
   { Rmatrix, bvec }= RESTR27(ms,rcoin);
elseif rest_op eq 28; 
   { Rmatrix, bvec }= RESTR28(ms,rcoin);
elseif rest_op eq 29; 
   { Rmatrix, bvec }= RESTR29(ms,rcoin);
elseif rest_op eq 30; 
   { Rmatrix, bvec }= RESTR30(ms,rcoin);
elseif rest_op eq 31; 
   { Rmatrix, bvec }= RESTR31(ms,rcoin);
elseif rest_op eq 32; 
   { Rmatrix, bvec }= RESTR32(ms,rcoin);
elseif rest_op eq 33; 
   { Rmatrix, bvec }= RESTR33(ms,rcoin);
elseif rest_op eq 34; 
   { Rmatrix, bvec }= RESTR34(ms,rcoin);
endif; 

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

@* data transformation @
MYY = eye(sample) - YY'invpd(YY*YY')*YY; @ idempotent @
vDY = vec(DY');

@ **************************** 
* Choice of initial estimates:
* **************************** @
{ beta0 } = INICOIN(ms,rcoin,E4beta,Rmatrix,bvec); @ ms by r @
/* beta0 = E4beta; */
@* estimation of alpha and omega given initial beta and MLL @
{ alpha0, omega0, MLL0 } = AOest(beta0,DY,ZS1,YY,yorder,sample,rcoin);

"Initial estimates";
"beta"; beta0; 
"alpha"; alpha0;
"omega"; omega0;
"maximum likelihood"; MLL0;

JO4beta = beta0;
JO4xi = JO4beta'ZS1; @ 4 by T @

@ ***********
* BS Algorithm
* *********** @
@ Tolerance level used in convergence criteria @
tol1 = 10^(-4);   @ check if Like(i)-Like(i-1)<tol1 @
tol2 = 10^(-4)*(ms*rcoin-dimR); @ check if |Theta1(i)-Theta1(i-1)|< tol2 @

maxiter = 10000; @ max number of iterations @
ij = 1;
do until ij > maxiter;

@* estimation of beta subject to restrictions @
   { beta1, U11 } = Best(alpha0,Omega0,DY,ZS1,YY,ms,rcoin,Rmatrix,bvec);

@ *********************************
* check the collinearity of beta's
* ********************************* @
   fflag = 0;
   if rank(beta1'ZS1*MYY*ZS1'beta1) < rcoin; fflag = 1;
      "beta's are collinear";
   endif;

@* estimation of alpha and omega given updated beta and MLL @
   { alpha1, omega1, MLL1 } = AOest(beta1,DY,ZS1,YY,yorder,sample,rcoin);

@* check if the convergence occurred @
   convt = abs(vec(beta1)-vec(beta0)); convtr = sumc(convt);
   convlike = MLL1-MLL0;

   cls; screen on; output off; format /rd 12,8;
   "Number of Iterations:";; ij;
   "Restricted MLL: ";;MLL1;
   "MLL-MLL(-1):";;convlike;

   if    convlike < tol1;
      if    convtr < tol2;  cflag = 1;
            output on; "converged after " ij " iterations"; 
            break;  @ convergence occurred @
      endif;
   endif;

@ update @
   alpha0 = alpha1; Omega0 = Omega1; beta0 = beta1; MLL0 = MLL1;

   ij = ij + 1;"";
endo;

output on; format /rd 8,4;
"Converged estimates";
"beta"; beta1; 
"alpha"; alpha1;
"omega"; omega1;
"maximum likelihood"; MLL1;
"LR test"; 2*(JMLL-MLL1);

O4beta = beta1;
O4xi = O4beta'ZS1; @ 4 by T @

@ ************************************
* STEP 3: Short-run Dynamic Estimation
* ************************************ @
@ ***********************
* Conditional VECM for dy
* *********************** @
@* r = 0, no cointegration @
E0Yreg = YY;   @ ?? by T @
k0 = yorder*rows(YY); @ number of total parameters @
{ E0ally, E0ey, E0omey, E0LL, E0AIC, E0SBC, E0HQ }
= VECdy(Dy,E0Yreg,sample,yorder,k0);

@* r = 1 @
E1Yreg = E1xi|YY;   @ ?? by T @
k1 = k0 + (ms*1-1^1);
{ E1ally, E1ey, E1Omey, E1LL, E1AIC, E1SBC, E1HQ }
= VECdy(Dy,E1Yreg,sample,yorder,k1);

@* r = 2 @
E2Yreg = E2xi|YY;   @ ?? by T @
k2 = k0 + (ms*2-2^2);
{ E2ally, E2ey, E2Omey, E2LL, E2AIC, E2SBC, E2HQ }
= VECdy(Dy,E2Yreg,sample,yorder,k2);

@* r = 3 @
E3Yreg = E3xi|YY;   @ ?? by T @
k3 = k0 + (ms*3-3^2);
{ E3ally, E3ey, E3Omey, E3LL, E3AIC, E3SBC, E3HQ }
= VECdy(Dy,E3Yreg,sample,yorder,k3);

@* r = 4 @
E4Yreg = E4xi|YY;   @ ?? by T @
k4 = k0 + (ms*4-4^2);
{ E4ally, E4ey, E4Omey, E4LL, E4AIC, E4SBC, E4HQ }
= VECdy(Dy,E4Yreg,sample,yorder,k4);

O4Yreg = O4xi|YY;   @ ?? by T @
k4 = k0 + (ms*4-dimR);
{ O4ally, O4ey, O4Omey, O4LL, O4AIC, O4SBC, O4HQ }
= VECdy(Dy,O4Yreg,sample,yorder,k4);

JO4Yreg = JO4xi|YY;   @ ?? by T @
k4 = k0 + (ms*4-dimR);
{ JO4ally, JO4ey, JO4Omey, JO4LL, JO4AIC, JO4SBC, JO4HQ }
= VECdy(Dy,JO4Yreg,sample,yorder,k4);

@* r = 5 @
E5Yreg = E5xi|YY;   @ ?? by T @
k5 = k0 + (ms*5-5^2);
{ E5ally, E5ey, E5Omey, E5LL, E5AIC, E5SBC, E5HQ }
= VECdy(Dy,E5Yreg,sample,yorder,k5);

@* r = 6 @
E6Yreg = E6xi|YY;   @ ?? by T @
k6 = k0 + (ms*6-6^2);
{ E6ally, E6ey, E6Omey, E6LL, E6AIC, E6SBC, E6HQ }
= VECdy(Dy,E6Yreg,sample,yorder,k6);

@* r = 7 @
E7Yreg = E7xi|YY;   @ ?? by T @
k7 = k0 + (ms*7-7^2);
{ E7ally, E7ey, E7Omey, E7LL, E7AIC, E7SBC, E7HQ }
= VECdy(Dy,E7Yreg,sample,yorder,k7);

@ *******************
  Marginal VAR for dx
  ******************* @
@* Two options:
   A. dx on inpt, dzlags
   B. dx on inpt, significant dzlags @

@* Case A: Unrestricted  @
if     case eq 1; XregA = dzlags; 
elseif case eq 2; XregA = dzlags;
elseif case eq 3; XregA = dzlags|inpt; 
elseif case eq 4; XregA = dzlags|inpt; 
elseif case eq 5; XregA = dzlags|inpt|tr; 
elseif case eq 6; XregA = dzlags|inpt|didum; 
elseif case eq 7; XregA = dzlags|inpt|dtdum; 
endif;
{ allxA, exA, OmexA, LLxA, AICxA, SBCxA, HQxA }
= MARdx(DX,XregA);

@* Case B: Restricted ?? @
if     case eq 1; XregB = dzlags; 
elseif case eq 2; XregB = dzlags;
elseif case eq 3; XregB = dzlags|inpt; 
elseif case eq 4; XregB = dzlags|inpt; 
elseif case eq 5; XregB = dzlags|inpt|tr; 
elseif case eq 6; XregB = dzlags|inpt; 
elseif case eq 7; XregB = dzlags|inpt; 
endif;
{ allxB, exB, OmexB, LLxB, AICxB, SBCxB, HQxB }
= MARdx(DX,XregB);

@ *************************************************
* FULL SYSTEM VAR COMBINING CONDITIONAL VECM FOR DY 
* AND MARGINAL VAR FOR DX
* *********************************************** @
@ maybe use EX4 for dy and A for dx as benchmark @
@ exact-identified beta with different cointegration rank or over-identified @
@ Abeta: all beta's including restricted deterministics @

"Choose which beta matrix to use for GIRs, OIRs and SIRs (0-9)"; e_op = con(1,1);

if e_op eq 0; 
   allY=E0allY; OmegaY=E0OmeY; eY=E0eY; Yreg=E0Yreg;
elseif e_op eq 1; 
   Abeta=E1beta; allY=E1allY; OmegaY=E1OmeY; eY=E1eY; Yreg=E1Yreg;
elseif e_op eq 2; 
   Abeta=E2beta; allY=E2allY; OmegaY=E2OmeY; eY=E2eY; Yreg=E2Yreg;
elseif e_op eq 3; 
   Abeta=E3beta; allY=E3allY; OmegaY=E3OmeY; eY=E3eY; Yreg=E3Yreg;
elseif e_op eq 4; 
   Abeta=E4beta; allY=E4allY; OmegaY=E4OmeY; eY=E4eY; Yreg=E4Yreg; 
elseif e_op eq 5; 
   Abeta=E5beta; allY=E5allY; OmegaY=E5OmeY; eY=E5eY; Yreg=E5Yreg;
elseif e_op eq 6; 
   Abeta=E6beta; allY=E6allY; OmegaY=E6OmeY; eY=E6eY; Yreg=E6Yreg;
elseif e_op eq 7; 
   Abeta=E7beta; allY=E7allY; OmegaY=E7OmeY; eY=E7eY; Yreg=E7Yreg;
elseif e_op eq 8; 
   Abeta=O4beta; allY=O4allY; OmegaY=O4OmeY; eY=O4eY; Yreg=O4Yreg;
elseif e_op eq 9; 
   Abeta=JO4beta; allY=JO4allY; OmegaY=JO4OmeY; eY=JO4eY; Yreg=JO4Yreg;
endif; 

rcoin = cols(Abeta); @ rank of coin vector @

@ ******************************************
* Conditional VECM estimation resutls for DY
* ******************************************
* ESTRES: coefficients~standard err~t-ratio
* DIAGRES1: RSQ|RBARSQ|SIGE|LL|AIC|SBC|HQ: 7x1 
* DIAGRES2: chiSC~chiFF~chiNO~chiHE 
*            pSC  ~pFF  ~pNO  ~pHE          
* ****************************************** @ 
VRES = eY;          @ Residuals @
VYHAT = ALLY*Yreg;  @ dyhat @
VREG = Yreg';       @ regressors: T by nreg @
nreg = cols(VREG);

ecmi = 1;
do until ecmi > yorder;

    VDEP = DY[ecmi,.]';       @ dependent variable: T by 1 @
    RESID = VRES[ecmi,.]';    @ residual: T by 1 @
    ECMCOEFF = ALLY[ecmi,.]'; @ coefficients @
    YHAT = VYHAT[ecmi,.]';    @ fitted value: T by 1 @

    { ESTRES, DIAGRES1, DIAGRES2 } 
    = VECM(ECMCOEFF,VDEP,VREG,RESID,YHAT,nreg);

    yout1 = xlswritem(ESTRES,"VECY","b2",ecmi,"");
    yout2 = xlswritem(DIAGRES1',"VECY","f2",ecmi,"");
    yout3 = xlswritem(DIAGRES2,"VECY","f5",ecmi,"");

@ Headers for VECY @
    bb1 = {"Coeff." "StdError" "t-ratio" "" "Rsquare" "R-Bar" "Sige" "LL" "AIC" "SBC" "HQ"};
    bb2 = {"SC" "FF" "NOR" "HET"};
    bb3 = {"Chi", "p"};
    bb4 = {"tobq"};
    bb5 = {"winf"};
    bb6 = {"inf"};
    bb7 = {"rus"};
    bb8 = {"dsc"};
    bb9 = {"cf"};
    bb10 = {"invest"};
    bb11 = {"y"};

    bbb1 = xlswritesa(bb1, "VECY", "b1", ecmi,"");
    bbb2 = xlswritesa(bb2, "VECY", "f4", ecmi,"");
    bbb3 = xlswritesa(bb3, "VECY", "e5", ecmi,"");
    bbb4 = xlswritesa(bb4, "VECY", "a1", 1,"");
    bbb5 = xlswritesa(bb5, "VECY", "a1", 2,"");
    bbb6 = xlswritesa(bb6, "VECY", "a1", 3,"");
    bbb7 = xlswritesa(bb7, "VECY", "a1", 4,"");
    bbb8 = xlswritesa(bb8, "VECY", "a1", 5,"");
    bbb9 = xlswritesa(bb9, "VECY", "a1", 6,"");
    bbb10 = xlswritesa(bb10, "VECY", "a1", 7,"");
    bbb11 = xlswritesa(bb11, "VECY", "a1", 8,"");

cls; output on;
"ECM Estimation Results for ";; ecmi;" Equation";
"**********************************************";"";
    PRECM(ESTRES,DIAGRES1,DIAGRES2);

    ecmi = ecmi + 1;
endo;

x_op = 1;
if x_op eq 1; 
   allX = allXA; Omegax = OmeXA; eX = eXA; Xreg=XregA;
elseif x_op eq 2; @ ?? @ 
   allX = allXB; Omegax = OmeXB; eX = eXB; Xreg=XregB;
endif;

@ **************************************
* Marginal VAR estimation results for DX 
* **************************************
* ESTRES: coefficients~standard err~t-ratio 
* DIAGRES1: RSQ|RBARSQ|SIGE|LL|AIC|SBC|HQ: 7x1 
* DIAGRES2: chiSC~chiFF~chiNO~chiHE 
             pSC  ~pFF  ~pNO  ~pHE  
* ****************************************** @ 
VRES = eX;          @ Residuals: n @
VXHAT = ALLX*Xreg;  @ dyhat: n @
VREG = Xreg';       @ common regressors:  @
nreg = cols(VREG);

ej = 1;
do until ej > xorder;

   VARCO = ALLX[ej,.]';  @ coefficients @
   VDEP = DX[ej,.]';     @ dependent variable @
   RESID = VRES[ej,.]';  @ residual @
   XHAT = VXHAT[ej,.]';  @ fitted value @

   { ESTRES, DIAGRES1, DIAGRES2 } 
   = VECM(VARCO,VDEP,VREG,RESID,XHAT,nreg);

    xout1 = xlswritem(ESTRES,"VARX","b2",ej,"");
    xout2 = xlswritem(DIAGRES1',"VARX","f2",ej,"");
    xout3 = xlswritem(DIAGRES2,"VARX","f5",ej,"");

@ Headers for VARX @
    cc1 = {"Coeff." "StdError" "t-ratio" "" "Rsquare" "R-Bar" "Sige" "LL" "AIC" "SBC" "HQ"};
    cc2 = {"SC" "FF" "NOR" "HET"};
    cc3 = {"Chi", "p"};
    cc4 = {"poil"};
    cc5 = {"ys"};
    ccc1 = xlswritesa(cc1, "VARX", "b1", ej,"");
    ccc2 = xlswritesa(cc2, "VARX", "f4", ej,"");
    ccc3 = xlswritesa(cc3, "VARX", "e5", ej,"");
    ccc4 = xlswritesa(cc4, "VARX", "a1", 1,"");
    ccc5 = xlswritesa(cc5, "VARX", "a1", 2,"");

cls; output on;
"Estimation Results for ";; ej;" Equation";
"***************************************************";"";
    PRECM(ESTRES,DIAGRES1,DIAGRES2);

   ej = ej + 1;
endo;

cls; output off;

Bbeta = Abeta[1:m,.]; 
@ m by r: deleting the last row(s) when restricted 
deterministics are present in cointegrating vectors @

@* Decomposition of parameters: assuming case 6 for now?? @

{AlpY,GamY,Lam,b0Y,b1Y,d0Y,d1Y,ddY,dtY} 
= decompY(ally,rcoin,pyorder,yorder,xorder,case,Abeta);

pxorder = pyorder;
{GamX,b0X,ddX,dtX} = decompX(allx,pxorder,yorder,xorder,case);

@* By construction we set the following parameters to zero @
AlpX = zeros(xorder,rcoin);
b1X = zeros(xorder,1);
d0X = zeros(xorder,1);
d1X = zeros(xorder,1);

@ NB: we are now using the same lag order @
if pxorder > pyorder;
   if pyorder eq 1; GamY = zeros(yorder,m*(pxorder-pyorder));
   else;       GAMY = GamY~zeros(yorder,m*(pxorder-pyorder));
   endif;
endif;

@ ************************* @
@ Contemporaneous MATRIX A0 @
@ ************************* @
A1 = eye(xorder)~zeros(xorder,yorder);
A2 = -Lam~eye(yorder);
A0 = A1|A2; IA0 = inv(A0); @ inverse @

@ ********************************************** @
@ CONSTRUCTION OF PARAMETERS for THE FULL SYSTEM @
@ ********************************************** @
ALP = AlpX|AlpY;
PAI = -ALP*Bbeta';  @ m by m @
gtmp = GamX|GamY; GAM = IA0*gtmp;
b0tmp = b0x|b0y;  b0 = IA0*b0tmp;
b1tmp = b1x|b1y;  b1 = IA0*b1tmp;
d0tmp = d0x|d0y;  d0 = IA0*d0tmp;
d1tmp = d1x|d1y;  d1 = IA0*d1tmp;
ddtmp = ddX|ddY;  dd = IA0*ddtmp;
dttmp = dtX|dtY;  dt = IA0*dttmp;

@ ****************** @
@ COVARIANCE MATRIX:
* OMEGA (ex and ey uncorrelated by construction) and
* COMEGA = IA0*Omega*IA0', which is correlated
* ************************************************** @
O1 = Omegax~zeros(xorder,yorder);
O2 = zeros(yorder,xorder)~Omegay;
Omega = O1|O2;
COmega = IA0*Omega*IA0'; @ reduced form COV @

@ ********************* @
@ FULL RESIDUAL VECTORS @
@ ********************* @
ez = ex|ey;

@ ******************************************************* @
@ Computation of A11 and W11 using Choleski decomposition:
* *******************************************************
* This is necessary for an analysis of structurally identified
* shocks such as monetary policy shock. Now assume the first 6 
* variables are recursive in the order of poil~ys~tobq~winf~
* inf~rus
* **************************************************** @
n_or = 6; @ No. of first block recursive variables @
Com11 = Comega[1:n_or,1:n_or]; @ z1 = (poild~ys~tobq~winf~inf~rus) @
P11 = chol(Com11); D11 = eye(n_or);
jj = 1;
do until jj > n_or;
   D11[jj,jj] = P11[jj,jj];
   jj = jj+1;
endo;
IA11 = P11'inv(D11);
W11 = D11.^2;  @ n_or by n_or variance matrix for structural errors @
A11 = inv(IA11);

@ *************************************************
* Computing DYNAMIC VMA PARAMETERS USING BIG VAR(1)
* ************************************************* @
cls; output off;
"Specify the forecast horizon.";"";
"HINT: type 40 here."; nhorz = con(1,1);

{ FAI, FF, BN } = DYPARA(pai,gam,porder,m,nhorz);

@ ********************************
* STEP 4: FURTHER DYNAMIC ANALYSIS
* ******************************** @
@ ***********************************
* 1. STABILITY OF THE FULL VEC SYSTEM
* *********************************** @
BEGSTAB:
@******@
output off;
"To evaluate stability of VEC system, you need to truncate the
infinite order sum of matrices. Specify the truncation order.";
"HINT: type 20 here.";""; ntrunc = con(1,1);

{ va, modva } = CHKSTAB(m,rcoin,porder,alp,Bbeta,gam,ntrunc);

@* Exporting the stability test results to Excel file @
out2 = xlswritem(modva,"Stability","b2",1,"");

@ ************************
*  2. PERSISTENCE PROFILES
* ************************ @
BEGPER:
@******@
hz = zeros(nhorz,rcoin);
ir = 1;
do until ir > rcoin;
   BETTA = BBETA[.,ir];
   { hz[.,ir] } = persis(m,nhorz,BETTA,COMEGA,BN);
   ir = ir + 1;
endo;
ini = ones(1,rcoin); hz = ini|hz;

@* Exporting the PP results to Excel file @
out3 = xlswritem(hz,"persist","b2",1,"");

@ Headers for Persistence Profiles of Cointegrating Vectors @
ee1 = {"persist" "CV1" "CV2" "CV3" "CV4"};
eee1 = xlswritesa(ee1, "persist", "a1", 1,"");

@ ****************************************
* 3. Generalised Impulse Response Analysis
* **************************************** @
BEGIMP:
@******@
cls; output off;
"This program provides both GIR and OIR impacts of shocks of all equations 
on all variables and cointegrating relations. Currently, the order of 
variables is given by poil~ys~tobq~sop~winf~inf~rus~dsc~cf~invest~y. Only
GIR is invariant to re-ordering. This order is the same as used in Structural
Impulse Response analysis.";"";

GIR = zeros((nhorz+1)*m,m); @ with respect to m shocks @
GIRZ = zeros((nhorz+1)*rcoin,m);
OIR = zeros((nhorz+1)*m,m);
OIRZ = zeros((nhorz+1)*rcoin,m);
ij = 1;
do until ij > m;
   sh_op = ij;
   { tmpGIR, tmpGIRZ, tmpOIR, tmpOIRZ }
   = Impulse(sh_op,m,nhorz,rcoin,BBETA,OMEGA,BN,IA0);

@* Exporting the results to Excel file @
   GIRres = tmpGIR~tmpGIRZ~tmpOIR~tmpOIRZ; 
   out4 = xlswritem(GIRres,"GIR","b3",sh_op,"");
   ij = ij + 1;
endo;

@ Headers for GIR and OIR @
    ff1 = {"GIR" "" "" "" "" "" "" "" "" "" "" "" "" "" "OIR"};
    ff2 = {"poil" "ys" "tobq" "winf" "inf" "rus" "dsc" "cf" "invest" "y" "CV1" "CV2" "CV3" "CV4"
          "poil" "ys" "tobq" "winf" "inf" "rus" "dsc" "cf" "invest" "y" "CV1" "CV2" "CV3" "CV4"};

    fffa1 = xlswritesa(ff1,"GIR","b1",1,"");
    fffa2 = xlswritesa(ff1,"GIR","b1",2,"");
    fffa3 = xlswritesa(ff1,"GIR","b1",3,"");
    fffa4 = xlswritesa(ff1,"GIR","b1",4,"");
    fffa5 = xlswritesa(ff1,"GIR","b1",5,"");
    fffa6 = xlswritesa(ff1,"GIR","b1",6,"");
    fffa7 = xlswritesa(ff1,"GIR","b1",7,"");
    fffa8 = xlswritesa(ff1,"GIR","b1",8,"");
    fffa9 = xlswritesa(ff1,"GIR","b1",9,"");
    fffa10 = xlswritesa(ff1,"GIR","b1",10,"");

    fffb1 = xlswritesa(ff2,"GIR","b2",1,"");
    fffb2 = xlswritesa(ff2,"GIR","b2",2,"");
    fffb3 = xlswritesa(ff2,"GIR","b2",3,"");
    fffb4 = xlswritesa(ff2,"GIR","b2",4,"");
    fffb5 = xlswritesa(ff2,"GIR","b2",5,"");
    fffb6 = xlswritesa(ff2,"GIR","b2",6,"");
    fffb7 = xlswritesa(ff2,"GIR","b2",7,"");
    fffb8 = xlswritesa(ff2,"GIR","b2",8,"");
    fffb9 = xlswritesa(ff2,"GIR","b2",9,"");
    fffb10 = xlswritesa(ff2,"GIR","b2",10,"");
   ij = ij + 1;

f1 = {"Shock", "poil"}; 
f2 = {"Shock", "ys"}; 
f3 = {"Shock", "tobq"}; 
f4 = {"Shock", "winf"};
f5 = {"Shock", "inf"};   
f6 = {"Shock", "rus"};
f7 = {"Shock", "dsc"};
f8 = {"Shock", "cf"};   
f9 = {"Shock", "invest"}; 
f10 = {"Shock", "y"}; 

af1 = xlswritesa(f1, "GIR", "a1", 1,"");
af2 = xlswritesa(f2, "GIR", "a1", 2,"");
af3 = xlswritesa(f3, "GIR", "a1", 3,"");
af4 = xlswritesa(f4, "GIR", "a1", 4,"");
af5 = xlswritesa(f5, "GIR", "a1", 5,"");
af6 = xlswritesa(f6, "GIR", "a1", 6,"");
af7 = xlswritesa(f7, "GIR", "a1", 7,"");
af8 = xlswritesa(f8, "GIR", "a1", 8,"");
af9 = xlswritesa(f9, "GIR", "a1", 9,"");
af10 = xlswritesa(f10, "GIR", "a1", 10,"");

@ ***************************************
* 4. Structural Impulse Response Analysis
* *************************************** @
BEGSIR:
@******@
cls; output off;
"We now compute structural impulse response functions wrt
shocks to oil price, potential output, Tobin's Q (stock market?),
profit share (investment prospects?), wage inflation, inflation and
the federal funds rate (monetary policy shock). We also consider shocks
to the interest rate intercept shift";
"press any key"; wait; "";

@ SIR: (h+1) by m, IRF of m variables
* SIRZ: (h+1) by rcoin, IRF of r cointegrating relationships
* where h is the forecast horizon.
* There are seven set of results with respect to shocks to
* O: oil price, YS: potential output, TQ: Tobin's Q, WI: wage
* inflation, IN: inflation, R: interest rate, I: intercept shift.
* ********************************************************** @

{ OSIR, OSIRZ, YSSIR, YSSIRZ, TQSIR, TQSIRZ, WISIR, WISIRZ, INSIR, INSIRZ, RSIR, RSIRZ, ISIR, ISIRZ }
= STIRF(m,Nhorz,rcoin,BBETA,COMEGA,BN,IA11,W11);

@* Exporting the results to Excel file @
sout1 = xlswritem(OSIR~OSIRZ,"SIR","b2",1,"");
sout2 = xlswritem(YSSIR~YSSIRZ,"SIR","b2",2,"");
sout3 = xlswritem(TQSIR~TQSIRZ,"SIR","b2",3,"");
sout4 = xlswritem(WISIR~WISIRZ,"SIR","b2",4,"");
sout5 = xlswritem(INSIR~INSIRZ,"SIR","b2",5,"");
sout6 = xlswritem(RSIR~RSIRZ,"SIR","b2",6,"");
sout7 = xlswritem(ISIR~ISIRZ,"SIR","b2",7,"");

@ Headers for SIR @
gg1 = {"poil" "ys" "tobq" "winf" "inf" "rus" "dsc" "cf" "invest" "y" "CV1" "CV2" "CV3" "CV4"};

sirj = 1;
do until sirj > 7;
    ggg1 = xlswritesa(gg1, "SIR", "b1", sirj,"");
    sirj = sirj + 1;
endo;

g1 = {"Shock", "poil"}; 
g2 = {"Shock", "ys"}; 
g3 = {"Shock", "tobq"}; 
g4 = {"Shock", "winf"};
g5 = {"Shock", "inf"};   
g6 = {"Shock", "rus"};
g7 = {"Shock", "idum"};

ag1 = xlswritesa(g1, "SIR", "a1", 1,"");
ag2 = xlswritesa(g2, "SIR", "a1", 2,"");
ag3 = xlswritesa(g3, "SIR", "a1", 3,"");
ag4 = xlswritesa(g4, "SIR", "a1", 4,"");
ag5 = xlswritesa(g5, "SIR", "a1", 5,"");
ag6 = xlswritesa(g6, "SIR", "a1", 6,"");
ag7 = xlswritesa(g7, "SIR", "a1", 7,"");


@ ********************
* 5. Central Forecasts
* ******************** @
@* h-step ahead forecasts (quarterly rate): h = 1,2,...,24,
*  4-quarter moving average and h-step average forecasts @
Zd = Xdata~Ydata; @ T+?? by m @ 
Zd = Zd[1:rows(ZD)-stt1,.]; @ chk delete it later @  
hh = tdum[rows(tdum)]; @ last observation for time dummy @
hh = sample-br +1;

FDZ = zeros(nhorz,m); @ forecasts of dz: n by m @
FZ  = zeros(nhorz,m); @ forecasts of z:  n by m @
j = 1;
do until j > nhorz;

@* data transformation for forecasting @
   { Z1T, DZlagsT } = TRDATA4(Zd,yorder,xorder,porder);

@* generating the regressors for VAR dx @
   k=j;
   { XregT } = REGXT(DZlagsT,case,sample,k,porder); 

@* forecast of DX at T+1: mx by 1 @
   DXT = allX*XregT; 

@* generating the regressors for VEC dy @
   { YregT } 
   = REGYT(Z1T,DZlagsT,DXT,case,porder,xorder,Abeta,sample,k,hh);

@* forecast of DY at T+1: my by 1 @
   DYT = allY*YregT; 

@* forecast of DZ and Z at T+j, m by 1 @
  DZT = DXT|DYT;  ZT = DZT + Z1T; 

@* saving forecasts of DZ and Z at j=1,...,n, m by 1 @
   FDZ[j,.] = DZT'; FZ[j,.] = ZT';

@* updating Z @
   ZD = ZD|ZT'; 

   j = j + 1;
endo;

@* 4-quarter MA and h-step average forecast @
ActZ = Xdata~Ydata; ActDZ = ActZ-lag(ActZ); @ original data @

{ a4fz, afz } = A4CHANGE(fz,actz,nhorz,m);     @ levels: n by m @
{ a4fdz, afdz } = A4change(fdz,actdz,nhorz,m); @ 1st differences @

@* Exporting the results to Excel file @
Fout1 = xlswritem(FZ,"Forecast","b2",1,"");
Fout2 = xlswritem(FDZ,"Forecast","b2",2,"");
Fout3 = xlswritem(A4FZ,"Forecast","b2",3,"");
Fout4 = xlswritem(A4FDZ,"Forecast","b2",4,"");
Fout5 = xlswritem(AFZ,"Forecast","b2",5,"");
Fout6 = xlswritem(AFDZ,"Forecast","b2",6,"");

@ Headers for Central Forecast @
hh1 = {"oilp" "ys" "tobq" "winf" "inf" "rus" "dsc" "cfr" "invest" "y"};

centj = 1;
do until centj > 6;
    hhh1 = xlswritesa(hh1, "Forecast", "b1", centj,"");
    centj = centj + 1;
endo;

h1 = {"Quarter", "Level"}; 
h2 = {"Quarter", "Dif."}; 
h3 = {"4-qt MA", "Level"};
h4 = {"4-qt MA", "Dif."};
h5 = {"h-step", "Average", "Level"};   
h6 = {"h-step", "Average", "Dif."}; 

ah1 = xlswritesa(h1, "Forecast", "a1", 1,"");
ah2 = xlswritesa(h2, "Forecast", "a1", 2,"");
ah3 = xlswritesa(h3, "Forecast", "a1", 3,"");
ah4 = xlswritesa(h4, "Forecast", "a1", 4,"");
ah5 = xlswritesa(h5, "Forecast", "a1", 5,"");
ah6 = xlswritesa(h6, "Forecast", "a1", 6,"");

@ ********************************************* 
* 6. In-Sample Probability Forecasts Evaluation  
* ********************************************* @
@ ************************************** 
* Short-run Dynamic Recursive Estimation
* **************************************
* one-step ahead forecasts for 8 quarters:
* currently 06q1, ..., 07q4 using the samples 
* 83q3-05q4, 83q3-06q1, ..., 83q3-07q3, recursively @

"We now carry out the in-sample probability forecasts evaluation
for the directional changes of one-step ahead forecasts. ";

"Specify the number of iterations for stochastic simulation.";"";
"HINT: type 1000 here."; n_stoc = con(1,1);

"Choose the draw option for stochastic simulation.
1. Parametric 2. Non-parametric.";"";
"HINT: type 1 here."; s_draw = con(1,1);

@ Original data @
Zdata = Xdata~Ydata;
DZdata = Zdata-lag(Zdata);
D2Zdata = DZdata-lag(DZdata);

@ No of quarters for in-sample forecasting evaluation: may change @ 
nloop = 8; 

@* One-step ahead forecasts: QUARTERLY and 4-QT: level, diff, 2nd diff @
RZT1 = zeros(nloop,m); A4RZT1 = zeros(nloop,m); ARZT1 = zeros(nloop,m);
RDZT1 = zeros(nloop,m); A4RDZT1 = zeros(nloop,m); ARDZT1 = zeros(nloop,m);
RD2ZT1 = zeros(nloop,m); A4RD2ZT1 = zeros(nloop,m); ARD2ZT1 = zeros(nloop,m);

@* Root mean squared errors for quarterly & 4-qt MA forecasts @
RMSF = zeros(nloop,m);   @ level @
RMSDF = zeros(nloop,m);  @ 1st difference @
RMSD2F = zeros(nloop,m); @ 2nd difference @
RMSA4F = zeros(nloop,m);
RMSA4DF = zeros(nloop,m);
RMSA4D2F = zeros(nloop,m);

@* Prob of directonal changes, see below for definiton of events @
SDP = zeros(nloop,m);   @ quarterly @
A4SDP = zeros(nloop,m); @ 4-qt MA @

@* integral transform @
SIP = zeros(nloop,m);   @ quarterly @
A4SIP = zeros(nloop,m); @ 4-qt MA @

@* The big iteration starts here @
ranseed = 1234;
li = 1;
do until li > nloop;  "replication" li;

@*initialising: careful @
   rini = rows(Ydata)-nloop+li;
   RY = Ydata[1:rini-1,.];
   RX = Xdata[1:rini-1,.];
   RZ = RX~RY;  @ recursive samples @
   RDZ = RZ-lag(RZ); RDZT = RDZ[rows(RDZ),.]; @ diff obs at T @
   RD2Z = RDZ-lag(RDZ); @ 2nd differences @
   actZT1 = Zdata[rini,.]; @ actual Z at T+h @
   actDZT1 = DZdata[rini,.]; @ actual DZ at T+h @
   actD2ZT1 = D2Zdata[rini,.]; @ actual D2Z at T+h @
   a4tmp = Zdata[rini-3:rini,.]; 
   a4aZT1 = meanc(a4tmp)'; @ actual A4Z at T+h @  
   a4tmp1 = DZdata[rini-3:rini,.]; 
   a4aDZT1 = meanc(a4tmp1)'; @ actual A4DZ at T+h @  
   a4tmp2 = D2Zdata[rini-3:rini,.]; 
   a4aD2ZT1 = meanc(a4tmp2)'; @ actual A4D2Z at T+h @  

@ actual data for in-sample evaluation: nloop by m @
   if li eq 1; 
      actDZEV = actDZT1; actD2ZEV = actD2ZT1;
      a4aDZEV = a4aDZT1; a4aD2ZEV = a4aD2ZT1;
   else; 
      actDZEV = actDZEV|actDZT1; actD2ZEV = actD2ZEV|actD2ZT1;
      a4aDZEV = a4aDZEV|a4aDZT1; a4aD2ZEV = a4aD2ZEV|a4aD2ZT1;
   endif;

@* DATA TRANSFORMATION *@
   { RDY, RZ1, RDX, RADZLAGS } = TRDATA(RY,RX,yorder,xorder);

@* sample adjustment @
   RDY = RDY[.,stt+1:cols(RDY)];
   RZ1 = RZ1[.,stt+1:cols(RZ1)];
   rsample = cols(rdy);
   RDX = RDX[.,stt+1:cols(RDX)];
   RADZLAGS = RADZLAGS[.,stt+1:cols(RADZLAGS)]; 

   if p eq 1;
   else; RDZLAGS = RADZLAGS[1:m*(p-1),.];
   endif;

@* CONSTRUCTION of RZS1 and all other regressors, RYY *@
   rinpt = ones(rsample,1); rtr1 = seqa(0,1,rsample); rtr = seqa(1,1,rsample); 
   rinpt = rinpt'; rtr1 = rtr1'; rtr = rtr';
   br = brpoint-stt; @ break point adjusted @
   iduma = zeros(br-1,1);
   idumb = ones(rsample-br+1,1);
   ridum = iduma|idumb;
   tduma = zeros(br-1,1);
   tdumb = seqa(1,1,rsample-br+1);
   rtdum = tduma|tdumb;
   { ridum1, radidum, rtdum1, radtdum} = TRDATA2(ridum,rtdum,stt);
   ridum1 = ridum1[.,stt+1:cols(ridum1)];
   rtdum1 = rtdum1[.,stt+1:cols(rtdum1)];
   radidum = radidum[.,stt+1:cols(radidum)];
   radtdum = radtdum[.,stt+1:cols(radtdum)];
   rdidum = radidum[1:p,.]; 
   rdtdum = radtdum[1:p,.]; 

   { RZS1, RYY } = CREG(case,RZ1,rinpt,RTR1,Ridum1,Rtdum1,p,
                   RDZlags,xorder,RDX,RTR,rdidum,rdtdum);

@* Cointegration test and Johansen eigenvectors *@
   { RCOTEST, RJBETA } = COINTEST(RDY,RZS1,RYY,yorder,xorder);

@* printing and exporting cointegration test results @
   output on;"";"COINTEGRATION TEST RESULTS";"";
   if     li eq 1; "over 67q3-04q4";"";
   elseif li eq 2; "over 67q3-05q1";"";
   elseif li eq 3; "over 67q3-05q2";"";
   elseif li eq 4; "over 67q3-05q3";"";
   elseif li eq 5; "over 67q3-05q4";"";
   elseif li eq 6; "over 67q3-06q1";"";
   elseif li eq 7; "over 67q3-06q2";"";
   elseif li eq 8; "over 67q3-06q3";"";
   endif;
   "eigenval   trace     max"; format /rd 8,4; RCOTEST;"";
 
@* exactly identified (eigen) betas for r = 1,...,7: ms by r @
   ms = rows(RJbeta);
   RE7beta = RJbeta[.,ms-6:ms];
   RE6beta = RJbeta[.,ms-5:ms];
   RE5beta = RJbeta[.,ms-4:ms];
   RE4beta = RJbeta[.,ms-3:ms];
   RE3beta = RJbeta[.,ms-2:ms];
   RE2beta = RJbeta[.,ms-1:ms];
   RE1beta = RJbeta[.,ms];

@* Initial estimates subject to OV restrictions @

{ RJO4beta } = INICOIN(ms,rcoin,RE4beta,Rmatrix,bvec); @ ms by r @

@* error correction terms: r by T @
   RE1xi = RE1beta'RZS1; @ 1 by T @
   RE2xi = RE2beta'RZS1; @ 2 by T @
   RE3xi = RE3beta'RZS1; @ 3 by T @
   RE4xi = RE4beta'RZS1; @ 4 by T @
   RE5xi = RE5beta'RZS1; @ 5 by T @
   RE6xi = RE6beta'RZS1; @ 6 by T @
   RE7xi = RE7beta'RZS1; @ 7 by T @
   RO4xi = O4beta'RZS1; @ 4 by T @
   RJO4xi = RJO4beta'RZS1; @ 4 by T @

@* Conditional VECM for rdy @
@* r = 0, no cointegration @
   RE0Yreg = RYY;   @ ?? by T @
   k0 = yorder*rows(YY); @ number of total parameters @
   { RE0ally, RE0ey, RE0omey, RE0LL, RE0AIC, RE0SBC, RE0HQ }
   = VECdy(RDy,RE0Yreg,rsample,yorder,k0);
@* r = 1 @
   RE1Yreg = RE1xi|RYY;   @ ?? by T @
   k1 = k0 + (ms*1-1^1);
   { RE1ally, RE1ey, RE1Omey, RE1LL, RE1AIC, RE1SBC, RE1HQ }
   = VECdy(RDy,RE1Yreg,Rsample,yorder,k1);
@* r = 2 @
   RE2Yreg = RE2xi|RYY;   @ ?? by T @
   k2 = k0 + (ms*2-2^2);
   { RE2ally, RE2ey, RE2Omey, RE2LL, RE2AIC, RE2SBC, RE2HQ }
   = VECdy(RDy,RE2Yreg,Rsample,yorder,k2);
@* r = 3 @
   RE3Yreg = RE3xi|RYY;   @ ?? by T @
   k3 = k0 + (ms*3-3^2);
   { RE3ally, RE3ey, RE3Omey, RE3LL, RE3AIC, RE3SBC, RE3HQ }
   = VECdy(RDy,RE3Yreg,Rsample,yorder,k3);
@* r = 4 @
   RE4Yreg = RE4xi|RYY;   @ ?? by T @
   k4 = k0 + (ms*4-4^2);
   { RE4ally, RE4ey, RE4Omey, RE4LL, RE4AIC, RE4SBC, RE4HQ }
   = VECdy(RDy,RE4Yreg,Rsample,yorder,k4);

   RO4Yreg = RO4xi|RYY;   @ ?? by T @
   k4 = k0 + (ms*4-4^2);
   { RO4ally, RO4ey, RO4Omey, RO4LL, RO4AIC, RO4SBC, RO4HQ }
   = VECdy(RDy,RO4Yreg,Rsample,yorder,k4);

   RJO4Yreg = RJO4xi|RYY;   @ ?? by T @
   k4 = k0 + (ms*4-4^2);
   { RJO4ally, RJO4ey, RJO4Omey, RJO4LL, RJO4AIC, RJO4SBC, RJO4HQ }
   = VECdy(RDy,RJO4Yreg,Rsample,yorder,k4);
@* r = 5 @
   RE5Yreg = RE5xi|RYY;   @ ?? by T @
   k5 = k0 + (ms*5-5^2);
   { RE5ally, RE5ey, RE5Omey, RE5LL, RE5AIC, RE5SBC, RE5HQ }
   = VECdy(RDy,RE5Yreg,Rsample,yorder,k5);
@* r = 6 @
   RE6Yreg = RE6xi|RYY;   @ ?? by T @
   k6 = k0 + (ms*6-6^2);
   { RE6ally, RE6ey, RE6Omey, RE6LL, RE6AIC, RE6SBC, RE6HQ }
   = VECdy(RDy,RE6Yreg,Rsample,yorder,k6);
@* r = 7 @
   RE7Yreg = RE7xi|RYY;   @ ?? by T @
   k7 = k0 + (ms*7-7^2);
   { RE7ally, RE7ey, RE7Omey, RE7LL, RE7AIC, RE7SBC, RE7HQ }
   = VECdy(RDy,RE7Yreg,Rsample,yorder,k7);

@* Marginal VAR for dx *@
@* Case A: Unrestricted  @
   if     case eq 1; RXregA = Rdzlags; 
   elseif case eq 2; RXregA = Rdzlags;
   elseif case eq 3; RXregA = Rdzlags|Rinpt; 
   elseif case eq 4; RXregA = Rdzlags|Rinpt; 
   elseif case eq 5; RXregA = Rdzlags|Rinpt|Rtr; 
   elseif case eq 6; RXregA = Rdzlags|Rinpt|Rdidum; 
   elseif case eq 7; RXregA = Rdzlags|Rinpt|Rdtdum; 
   endif;
   { RallxA, RexA, ROmexA, RLLxA, RAICxA, RSBCxA, RHQxA }
   = MARdx(RDX,RXregA);

@* Case B: Restricted ?? @
   if     case eq 1; RXregB = Rdzlags; 
   elseif case eq 2; RXregB = Rdzlags;
   elseif case eq 3; RXregB = Rdzlags|Rinpt; 
   elseif case eq 4; RXregB = Rdzlags|Rinpt; 
   elseif case eq 5; RXregB = Rdzlags|Rinpt|Rtr; 
   elseif case eq 6; RXregB = Rdzlags|Rinpt; 
   elseif case eq 7; RXregB = Rdzlags|Rinpt; 
   endif;
   { RallxB, RexB, ROmexB, RLLxB, RAICxB, RSBCxB, RHQxB }
   = MARdx(RDX,RXregB);

@* exactly-identified beta with different cointegration rank or over-identified @
@ Abeta: all beta's including restricted deterministics @

    if e_op eq 0; 
      RallY=RE0allY; ROmegaY=RE0OmeY; ReY=RE0eY; RYreg=RE0Yreg;
    elseif e_op eq 1; 
      RAbeta=RE1beta; RallY=RE1allY; ROmegaY=RE1OmeY; ReY=RE1eY; RYreg=RE1Yreg;
    elseif e_op eq 2; 
      RAbeta=RE2beta; RallY=RE2allY; ROmegaY=RE2OmeY; ReY=RE2eY; RYreg=RE2Yreg;
    elseif e_op eq 3; 
      RAbeta=RE3beta; RallY=RE3allY; ROmegaY=RE3OmeY; ReY=RE3eY; RYreg=RE3Yreg;
    elseif e_op eq 4; 
      RAbeta=RE4beta; RallY=RE4allY; ROmegaY=RE4OmeY; ReY=RE4eY; RYreg=RE4Yreg; 
    elseif e_op eq 5; 
      RAbeta=RE5beta; RallY=RE5allY; ROmegaY=RE5OmeY; ReY=RE5eY; RYreg=RE5Yreg;
    elseif e_op eq 6; 
      RAbeta=RE6beta; RallY=RE6allY; ROmegaY=RE6OmeY; ReY=RE6eY; RYreg=RE6Yreg;
    elseif e_op eq 7; 
      RAbeta=RE7beta; RallY=RE7allY; ROmegaY=RE7OmeY; ReY=RE7eY; RYreg=RE7Yreg;
    elseif e_op eq 8; 
      RAbeta=O4beta; RallY=RO4allY; ROmegaY=RO4OmeY; ReY=RO4eY; RYreg=RO4Yreg;
    elseif e_op eq 9; 
      RAbeta=RJO4beta; RallY=RJO4allY; ROmegaY=RJO4OmeY; ReY=RJO4eY; RYreg=RJO4Yreg;
    endif;
    rcoin = cols(RAbeta); @ rank of coin vector @

   if x_op eq 1; 
      RallX = RallXA; ROmegax = ROmeXA; ReX = ReXA; RXreg=RXregA;
   elseif x_op eq 2; @ ?? @ 
      RallX = RallXB; ROmegax = ROmeXB; ReX = ReXB; RXreg=XregB;
   endif;

 @* COVARIANCE MATRIX and FULL RESIDUAL VECTORS *@
   RO1 = ROmegax~zeros(xorder,yorder);
   RO2 = zeros(yorder,xorder)~ROmegay;
   ROmega = RO1|RO2;
   Rez = Rex|Rey;

@* 1-step ahead forecast @
   rhh = rsample-br+1; @ last observation for time dummy ?? @

@* data transformation for forecasting @
   { RZ1T, RDZlagsT } = TRDATA4(RZ,yorder,xorder,p);

@* generating the regressors for VAR dx @
   k=1;
   { RXregT } = REGXT(RDZlagsT,case,rsample,k,porder); 

@* forecast of DX at T+1: mx by 1 @
   RDXT1 = RallX*RXregT; 

@* generating the regressors for VEC dy @
   { RYregT } 
   = REGYT(RZ1T,RDZlagsT,RDXT1,case,p,xorder,RAbeta,rsample,k,rhh);
 
@* forecast of DY at T+1: my by 1 @
   RDYT1 = RallY*RYregT; 

@* forecast of DZ and Z at T+1, m by 1 @
   tmpdZ = RDXT1|RDYT1; RDZT1[li,.] = tmpdZ'; 

@* forecast of Z at T+1, m by 1 @
   tmpZ = tmpdz+RZ1T; RZT1[li,.] = tmpZ';

@* forecast of D2Z at T+1, m by 1: RDZT: RDZ at T: 1 by m @
   tmpd2Z = tmpdZ-RDZT'; RD2ZT1[li,.] = tmpd2Z';

/* "chk: loop";; li; "tmpdz~tmpd2z"; tmpdz~tmpd2z; wait; "press any key";
   if li eq 1; save tmpz tmpdz tmpd2z;
   endif; */

@* 4-quarter MA and h-step average forecast: 1 by m @
   rnh=1; @ horizon of 1 only @
   { A4RZT1[li,.], ARZT1[li,.] } = A4CHANGE(tmpZ',RZ,rnh,m);     @ levels @
   { A4RDZT1[li,.], ARDZT1[li,.] } = A4CHANGE(tmpdZ',RDZ,rnh,m); @ 1st diff @
   { A4RD2ZT1[li,.], ARD2ZT1[li,.] } = A4CHANGE(tmpd2Z',RD2Z,rnh,m); @ 2nd diff @

@* RMSE: sqrt(E(yhat - y)^2) based on quaretrly forecasts @
   RMSF[li,.] = abs(RZT1[li,.]-actZT1);
   RMSDF[li,.] = abs(RDZT1[li,.]-actDZT1);
   RMSD2F[li,.] = abs(RD2ZT1[li,.]-actD2ZT1);
   RMSA4F[li,.] = abs(A4RZT1[li,.]-a4aZT1);
   RMSA4DF[li,.] = abs(A4RDZT1[li,.]-a4aDZT1);
   RMSA4D2F[li,.] = abs(A4RD2ZT1[li,.]-a4aD2ZT1);

@ ****************************
* Probability Events Forecasts
* **************************** @
@ ******************************************
* We consider the following 9 single events:
*
* for probability of directional changes
*
* E1: oil prices: pr(d2po>0)
* E2: foreign interest rates: pr(drs>0)
* E3: foreign output: pr(d2ys>0)
* E4: exchange rates: pr(de>0)
* E5: interest rates: pr(dr>0)
* E6: inflation: pr(dinf>0)
* E7: output: pr(d2y>0)
* E8: relative prices: pr(dpps>0)
* E9: real balance: pr(d2hy>0)
*
* for probability integral transforms
*
* E1: oil prices: pr(d2po > actd2po)
* E2: foreign interest rates: pr(drs > actdrs)
* E3: foreign output: pr(d2ys > actd2ys)
* E4: exchange rates: pr(de > actde)
* E5: interest rates: pr(dr > actdr)
* E6: inflation: pr(dinf > actdinf)
* E7: output: pr(d2y > actd2y)
* E8: relative prices: pr(dpps > actdpps)
* E9: real balance: pr(d2hy > actd2hy)
* ********************************************* *@
@* Choleski decomposition of Omega @
   Rupz = Chol(Romega); Rlpz = Rupz'; Rvhat = Inv(Rlpz)*Rez;
 
@ Drawing m by n_stoc vector of N(0,1) for future errors,
* which are symmetric such that mean = 0 by construction @
   rchk = n_stoc/2;
   rtmp1 = rndn(m,rchk); rtmp2 = -rtmp1;
   Rtmp = rtmp1~rtmp2; @ Rtmp = rndn(m,n_stoc) ?? @

   sumDF = zeros(1,m); sumA4DF = zeros(1,m); 
   sumIF = zeros(1,m); sumA4IF = zeros(1,m);
   ej = 1;
   do until ej > n_stoc;

@* draws of future errors at T+1: parametric one @
      Rrtmp = Rtmp[.,ej];

@ 1-step ahead forecast plus forecast error: quarterly: 1 by m @
      {RSF, RDSF } = stocfor(RZ,rsample,rnh,xorder,yorder,p,case,rnh,
                     RAbeta,RallY,RallX,RLpz,s_draw,Rvhat,Rrtmp);

@* forecast of RD2SF at T+1, 1 by m @
      RD2SF = RDSF-RDZT; 

@* 4-qt MA @
      { A4RSF, ARSF } = A4CHANGE(RSF,RZ,rnh,m);         @ levels @
      { A4RDSF, ARDSF } = A4CHANGE(RDSF,RDZ,rnh,m);     @ diff @
      { A4RD2SF, ARD2SF } = A4CHANGE(RD2SF,RD2Z,rnh,m); @ 2nd diff @

@* 9 probility events @
      {fdd, fdd2, a4fdd, a4fdd2, fid, fid2, a4fid, a4fid2 }
      = PREV2(RDSF,RD2SF,A4RDSF,A4RD2SF,actdzt1,actd2zt1,a4adzt1,a4ad2zt1);

@ collecting the relevant events: needs checking?? @
      DFLAG = fdd2[1]~fdd[2]~fdd2[3]~fdd[4:6]~fdd2[7:10];
      A4DFLAG = a4fdd2[1]~a4fdd[2]~a4fdd2[3]~a4fdd[4:6]~a4fdd2[7:10];
      IFLAG = fid2[1]~fid[2]~fid2[3]~fid[4:6]~fid2[7:10];
      A4IFLAG = a4fid2[1]~a4fid[2]~a4fid2[3]~a4fid[4:6]~a4fid2[7:10];

      sumDF = sumDF+DFLAG; sumA4DF=sumA4DF+A4DFLAG;
      sumIF = sumIF+IFLAG; sumA4IF=sumA4IF+A4IFLAG;

      ej=ej+1;
   endo;
   SDP[li,.] = sumDF./n_stoc; A4SDP[li,.] = sumA4DF./n_stoc;
   SIP[li,.] = sumIF./n_stoc; A4SIP[li,.] = sumA4IF./n_stoc;

   li=li+1;

endo;

@ *****************************
* In-sample Forecast evaluation
* *****************************
* We provide the following in-sample forecast evaluation results:
*
* 1. Forecast evaluation statistics: UD DD DU UU, where the first letter denotes 
*    the direction of forecast and the second the direction of actual outcome.
*
* 2. Hit ratio = (DD+UU)/(UD+DD+DU+UU)
*
* 3. Kuipers Score = H-F, where H = UU/(UU+UD) is the proportion of ups that
*    were correctly predicted to occur, and F = DU/(DU+DD) is the proportion 
*    of downs that were incorrectly predicted.
*
* 4. Pesaran-Timmerman statistic: PT=(phat-pstar)/(V(phat)-V(pstar))^(1/2),
*    where phat is the proportion of correctly predicted movements, pstar
*    the estimated probability of correctly predicting the event under the
*    null that forecasts and realisations are independently distributed and 
*    V(phat) and V(pstar) are the consistent variance estimates.
*
* 5. Kolmogorov-Smirnov test: KS = sup_x|F_n(x)-U(x)|, where F_n(x) is empirical 
*    CDF of probability integral transforms and U(x) is the CDF of iid U[0,1].
* **************************************************************************** @
@ Constructing the relevant point forecast results for
* quarterly and 4-qt MA, matching with prob events @

   CFEV = RD2ZT1[.,1:3]~RDZT1[.,4:6]~RD2ZT1[.,7:10];
   A4CFEV = A4RD2ZT1[.,1:3]~A4RDZT1[.,4:6]~A4RD2ZT1[.,7:10];
   actdat = actD2ZEV[.,1:3]~actDZEV[.,4:6]~actD2ZEV[.,7:10];
   a4actdat = a4aD2ZEV[.,1:3]~a4aDZEV[.,4:6]~a4aD2ZEV[.,7:10];
   RMSEF = RMSD2F[.,1:3]~RMSDF[.,4:6]~RMSD2F[.,7:10];
   RMSEA4F = RMSA4D2F[.,1:3]~RMSA4DF[.,4:6]~RMSA4D2F[.,7:10];


"Quarterly rates";
"***************";"";
format /rd 6,2;
"Point forecasts (annual percent) of dz or d2z in the column order of";
"   d2po      d2ys     d2qs      dwinf     dinf     drus    d2dsc    d2cf    d2inv d2y";
400.*CFEV;
"";
"Actual data (annual percent) of dz or d2z in the column order of";
"   d2po      d2ys     d2qs      dwinf     dinf     drus    d2dsc    d2cf    d2inv d2y";
400.*actdat;
"RMSE (quarterly rates) of dz or d2z in the column order of";
"   d2po      d2ys     d2qs      dwinf     dinf     drus    d2dsc    d2cf    d2inv d2y";
format /rd 6,3; RMSEF;

@ directional changes @

{ HITFOR, SUMDR, HRFOR, KSFOR, PTF } = HITCF(CFEV,actdat,nloop,m);

hhh0 = { "quarter", "direct'n", "change" };
hout0 = xlswritesa(hhh0, "HFOR","a1",1,"");
hhh1 = { "UD" "DD" "DU" "UU" };
hout1a = xlswritesa(hhh1, "HFOR","b1",1,"");
hout1b = xlswritem(HITFOR,"HFOR","b2",1,"");
hhh2 = { "Hit" "Rate" };
hout2a = xlswritesa(hhh2,"HFOR","b21",1,"");
hout2  = xlswritem(HRFOR,"HFOR","d21",1,"");
hhh3 = { "Kuipers" "Score" };
hout3a = xlswritesa(hhh3,"HFOR","b23",1,"");
hout3  = xlswritem(KSFOR,"HFOR","d23",1,"");
hhh4 = { "PT" "Test" };
hout4a = xlswritesa(hhh4,"HFOR","b25",1,"");
hout4  = xlswritem(PTF,  "HFOR","d25",1,"");

@ integral tranform @

{ HITIPR, SUMIPR, HRIPR, KSIPR, PTS } = HITPR(SDP,actdat,nloop,m);
{ sks } = KS(SIP,nloop,m);
"";

hhh0 = { "quarter", "integral", "transfom" };
hout0 = xlswritesa(hhh0, "HFOR","a1", 2,"");
hout1a = xlswritesa(hhh1, "HFOR","b1",2,"");
hout1b = xlswritem(HITIPR,"HFOR","b2",2,"");
hout2a = xlswritesa(hhh2,"HFOR","b21",2,"");
hout2  = xlswritem(HRIPR,"HFOR","d21",2,"");
hhh3 = { "Kuipers" "Score" };
hout3a = xlswritesa(hhh3,"HFOR","b23",2,"");
hout3  = xlswritem(KSIPR,"HFOR","d23",2,"");
hhh4 = { "PT" "Test" };
hout4a = xlswritesa(hhh4,"HFOR","b25",2,"");
hout4  = xlswritem(PTS,  "HFOR","d25",2,"");
hhh5 = { "KS" "Test" };
hout5a = xlswritesa(hhh5,"HFOR","b27",2,"");
hout5  = xlswritem(sks,  "HFOR","d27",2,"");

"4-Quarter MA";
"**************";"";format /rd 6,2;
"Point forecasts (annual percent) of dz or d2z in the column order of";
"   d2po      d2ys     d2qs      dwinf     dinf     drus    d2dsc    d2cf    d2inv d2y";
400.*A4CFEV;
"";
"Actual data (annual percent) of dz or d2z in the column order of";
"   d2po      d2ys     d2qs      dwinf     dinf     drus    d2dsc    d2cf    d2inv d2y";
400.*a4actdat;
"RMSE (quarterly rates) of dz or d2z in the column order of";
"   d2po      d2ys     d2qs      dwinf     dinf     drus    d2dsc    d2cf    d2inv d2y";
format /rd 6,3; RMSEA4F;

@ directional change @

{ HITFOR, SUMDR, HRFOR, KSFOR, PTF } = HITCF(A4CFEV,a4actdat,nloop,m);

hhh0 = { "4-qt MA", "direct'n", "change" };
hout0 = xlswritesa(hhh0, "HFOR","a1",3,"");
hhh1 = { "UD" "DD" "DU" "UU" };
hout1a = xlswritesa(hhh1, "HFOR","b1",3,"");
hout1b = xlswritem(HITFOR,"HFOR","b2",3,"");
hhh2 = { "Hit" "Rate" };
hout2a = xlswritesa(hhh2,"HFOR","b21",3,"");
hout2  = xlswritem(HRFOR,"HFOR","d21",3,"");
hhh3 = { "Kuipers" "Score" };
hout3a = xlswritesa(hhh3,"HFOR","b23",3,"");
hout3  = xlswritem(KSFOR,"HFOR","d23",3,"");
hhh4 = { "PT" "Test" };
hout4a = xlswritesa(hhh4,"HFOR","b25",3,"");
hout4  = xlswritem(PTF,  "HFOR","d25",3,"");

@ integral transform @

{ HITSPR, SUMIPR, HRSPR, KSSPR, PTS } = HITPR(A4SDP,a4actdat,nloop,m);
{ sks } = KS(A4SIP,nloop,m);

hhh0 = { "4-qt MA", "integral", "transfom" };
hout0 = xlswritesa(hhh0, "HFOR","a1", 4,"");
hout1a = xlswritesa(hhh1, "HFOR","b1",4,"");
hout1b = xlswritem(HITIPR,"HFOR","b2",4,"");

hout2a = xlswritesa(hhh2,"HFOR","b21",4,"");
hout2  = xlswritem(HRIPR,"HFOR","d21",4,"");

hhh3 = { "Kuipers" "Score" };
hout3a = xlswritesa(hhh3,"HFOR","b23",4,"");
hout3  = xlswritem(KSIPR,"HFOR","d23",4,"");

hhh4 = { "PT" "Test" };
hout4a = xlswritesa(hhh4,"HFOR","b25",4,"");
hout4  = xlswritem(PTS,  "HFOR","d25",4,"");

hhh5 = { "KS" "Test" };
hout5a = xlswritesa(hhh5,"HFOR","b27",4,"");
hout5  = xlswritem(sks,  "HFOR","d27",4,"");

@ **************************************
* 7. Out of Sample Probability Forecasts
* **************************************
* See above for the definition of the 10 events
* *********************************************
* Currently, future uncertainty only 
* ********************************** @

Zdata = Xdata~Ydata; @ T+?? by m @
DZdata = Zdata - lag(Zdata); 
DZdataT = DZdata[rows(DZdata),.]; @ obs at T @ 

@ Empirical distribution of forecasts: future uncertainty only @
nhm = nhorz*m;  
SFZ = zeros(nhm,n_stoc); SFDZ = zeros(nhm,n_stoc);    @ quarterly rates @     
SA4FZ = zeros(nhm,n_stoc); SA4FDZ = zeros(nhm,n_stoc);@ 4-qt MA @     
SAFZ = zeros(nhm,n_stoc); SAFDZ = zeros(nhm,n_stoc);  @ h-step avg @     

/*
@* Probability forecasts: future uncertainty only @
SPE1 = zeros(nhorz,rows(thdp)); @ E1 @
SPE2 = zeros(nhorz,rows(thdy)); @ E2 @
SPE3 = zeros(nhorz,1);  @ E3 @
SPE4 = zeros(nhorz,1);  @ E4 @
SPE5 = zeros(nhorz,1);  @ E5 @
SPE6 = zeros(nhorz,1);  @ E6 @
SPE7 = zeros(nhorz,1);  @ E7 @
SPE8 = zeros(nhorz,1);  @ E8 @
SPE9 = zeros(nhorz,1);  @ E9 @
SPE10 = zeros(nhorz,1); @ E10 @
SPECA = zeros(nhorz,6); @ 6 CA events @
*/

@ Choleski decomposition of Omgea @
Upz = Chol(Omega); Lpz = Upz'; vhat = Inv(Lpz)*ez;
   
rndseed 2222;
sj = 1;
do until sj > n_stoc;
cls; output off; "replication" sj;

@ Drawing the m by nhorz vector of N(0,1) for future errors which
* are perfectly symmetric with zero mean by construction @

/*
   if     sj eq 1; rtmp = rndn(m,nhorz); rtmp0 = rtmp;
   elseif sj eq 2; rtmp = -rtmp;         rtmp1 = rtmp;
   else;
      if rtmp1 eq -rtmp0; rtmp = rndn(m,nhorz);
      else; rtmp = -rtmp;
      endif;
      rtmp0 = rtmp1; rtmp1 = rtmp; @ updating @
   endif;
*/

@ forecasts plus forecast errors @
   rtmp = rndn(m,nhorz);
   {SZ, Sdz} = stocfor(ZData,sample,nhorz,xorder,yorder,porder,case,hh,
               Abeta,allY,allX,Lpz,s_draw,vhat,rtmp);

@ 4-quarter MA and h-step average forecast @
   { a4sz, asz } = A4CHANGE(sz,actz,nhorz,m);     @ levels: n by m @
   { a4sdz, asdz } = A4change(sdz,actdz,nhorz,m); @ 1st differences @

@ vectorising & saving: nh*m by 1 vector @
   SFZ[.,sj] = vec(SZ); SFDZ[.,sj] = vec(SDZ); 
   SA4FZ[.,sj] = vec(A4SZ); SA4FDZ[.,sj] = vec(A4SDZ); 
   SAFZ[.,sj] = vec(ASZ); SAFDZ[.,sj] = vec(ASDZ); 

/*
@ Probility evaluation for 10 events for dp and dy and for 6 CA events @
   { flagE1, flagE2, flagE3, flagE4, flagE5, flagE6, flagE7, flagE8, flagE9, flagE10, flagCA }
   = PREVENT(SZ,SDZ,actz,actdz,nhorz,m,DZdataT,n_im,n_ex,n_dp,n_y,THDP,THDY,EVDP,EVDY);

   SPE1 = SPE1 + flagE1;
   SPE2 = SPE2 + flagE2;
   SPE3 = SPE3 + flagE3;
   SPE4 = SPE4 + flagE4;
   SPE5 = SPE5 + flagE5;
   SPE6 = SPE6 + flagE6;
   SPE7 = SPE7 + flagE7;
   SPE8 = SPE8 + flagE8;
   SPE9 = SPE9 + flagE9;
   SPE10 = SPE10 + flagE10;
   SPECA = SPECA + flagCA;
*/

   sj = sj + 1;
endo;


@ ***************************************************** @
@ Empirical mean, variance and CIs of central forecasts @
@ ***************************************************** @
@ mean: empirical mean: (all n by m)
  med: empirical median 
  std: empirical standard deviation 
  L90: 90% lower confidence bound 
  U90: 90% upper confidence bound 
  L95: 95% lower confidence bound 
  U95: 95% upper confidence bound @

{ FZmean, FZmed, FZstd, FZL90, FZU90, FZL95, FZU95 }
= EMPDIST(SFZ,nhorz,m,n_stoc);

{ FDZmean, FDZmed, FDZstd, FDZL90, FDZU90, FDZL95, FDZU95 }
= EMPDIST(SFDZ,nhorz,m,n_stoc);

{ A4FZmean, A4FZmed, A4FZstd, A4FZL90, A4FZU90, A4FZL95, A4FZU95 }
= EMPDIST(SA4FZ,nhorz,m,n_stoc);

{ A4FDZmea, A4FDZmed, A4FDZstd, A4FDZL90, A4FDZU90, A4FDZL95, A4FDZU95 }
= EMPDIST(SA4FDZ,nhorz,m,n_stoc);

{ AFZmean, AFZmed, AFZstd, AFZL90, AFZU90, AFZL95, AFZU95 }
= EMPDIST(SAFZ,nhorz,m,n_stoc);

{ AFDZmean, AFDZmed, AFDZstd, AFDZL90, AFDZU90, AFDZL95, AFDZU95 }
= EMPDIST(SAFDZ,nhorz,m,n_stoc);

@ Exporting the results to Excel file @

SFZ1 = xlswritem(FZmean,"SFOR","b2",1,"");
SFZ2 = xlswritem(FZmed, "SFOR","b2",2,"");
SFZ3 = xlswritem(FZstd, "SFOR","b2",3,"");
SFZ4 = xlswritem(FZL90, "SFOR","b2",4,"");
SFZ5 = xlswritem(FZU90, "SFOR","b2",5,"");
SFZ6 = xlswritem(FZL95, "SFOR","b2",6,"");
SFZ7 = xlswritem(FZU95, "SFOR","b2",7,"");

SFDZ1 = xlswritem(FDZmean,"SFOR","b2",8,"");
SFDZ2 = xlswritem(FDZmed, "SFOR","b2",9,"");
SFDZ3 = xlswritem(FDZstd, "SFOR","b2",10,"");
SFDZ4 = xlswritem(FDZL90, "SFOR","b2",11,"");
SFDZ5 = xlswritem(FDZU90, "SFOR","b2",12,"");
SFDZ6 = xlswritem(FDZL95, "SFOR","b2",13,"");
SFDZ7 = xlswritem(FDZU95, "SFOR","b2",14,"");

SA4FZ1 = xlswritem(A4FZmean,"SA4FOR","b2",1,"");
SA4FZ2 = xlswritem(A4FZmed, "SA4FOR","b2",2,"");
SA4FZ3 = xlswritem(A4FZstd, "SA4FOR","b2",3,"");
SA4FZ4 = xlswritem(A4FZL90, "SA4FOR","b2",4,"");
SA4FZ5 = xlswritem(A4FZU90, "SA4FOR","b2",5,"");
SA4FZ6 = xlswritem(A4FZL95, "SA4FOR","b2",6,"");
SA4FZ7 = xlswritem(A4FZU95, "SA4FOR","b2",7,"");

SA4FDZ1 = xlswritem(A4FDZmea,"SA4FOR","b2",8,"");
SA4FDZ2 = xlswritem(A4FDZmed,"SA4FOR","b2",9,"");
SA4FDZ3 = xlswritem(A4FDZstd,"SA4FOR","b2",10,"");
SA4FDZ4 = xlswritem(A4FDZL90,"SA4FOR","b2",11,"");
SA4FDZ5 = xlswritem(A4FDZU90,"SA4FOR","b2",12,"");
SA4FDZ6 = xlswritem(A4FDZL95,"SA4FOR","b2",13,"");
SA4FDZ7 = xlswritem(A4FDZU95,"SA4FOR","b2",14,"");

SAFZ1 = xlswritem(AFZmean,"SAFOR","b2",1,"");
SAFZ2 = xlswritem(AFZmed,"SAFOR","b2",2,"");
SAFZ3 = xlswritem(AFZstd,"SAFOR","b2",3,"");
SAFZ4 = xlswritem(AFZL90,"SAFOR","b2",4,"");
SAFZ5 = xlswritem(AFZU90,"SAFOR","b2",5,"");
SAFZ6 = xlswritem(AFZL95,"SAFOR","b2",6,"");
SAFZ7 = xlswritem(AFZU95,"SAFOR","b2",7,"");

SAFDZ1 = xlswritem(AFDZmean,"SAFOR","b2",8,"");
SAFDZ2 = xlswritem(AFDZmed, "SAFOR","b2",9,"");
SAFDZ3 = xlswritem(AFDZstd, "SAFOR","b2",10,"");
SAFDZ4 = xlswritem(AFDZL90, "SAFOR","b2",11,"");
SAFDZ5 = xlswritem(AFDZU90, "SAFOR","b2",12,"");
SAFDZ6 = xlswritem(AFDZL95, "SAFOR","b2",13,"");
SAFDZ7 = xlswritem(AFDZU95, "SAFOR","b2",14,"");


@ **************************************************
* Data export for cointegration and IRF bootstraps
* ************************************************ @

@ Alpha @
ab1 = {"ALP" "Eq.1" "Eq.2" "Eq.3" "Eq.4"};
abb1 = xlswritesa(ab1, "coin_bs", "a1", 1, "");
abbb1 = xlswritem(alp, "coin_bs", "b2", 1, "");

@ Beta @
ab2 = {"BETA" "Eq.1" "Eq.2" "Eq.3" "Eq.4"};
abb2 = xlswritesa(ab2, "coin_bs", "a1", 2, "");
abbb2 = xlswritem(o4beta, "coin_bs", "b2", 2, "");

@ Gamma @
ab3 = {"GAMMA" "vardx1" "vardx2" "vec1" "vec2" "vec3" "vec4" "vec5" "vec6" "vec7" "vec8"};
abb3 = xlswritesa(ab3, "coin_bs", "a1", 3, "");
abbb3 = xlswritem(gtmp, "coin_bs", "b2", 3, "");

@ Lambda @
ab4 = {"LAMBDA" "poil" "ys"};
abb4 = xlswritesa(ab4, "coin_bs", "a1", 4, "");
abbb4 = xlswritem(lam, "coin_bs", "b2", 4, "");

@ Constant coefficients @
ab5 = {"CONST" "coef"};
abb5 = xlswritesa(ab5, "coin_bs", "a1", 5, "");
abbb5 = xlswritem(b0x|b0y, "coin_bs", "b2", 5, "");

@ Trend coefficients @
ab6 = {"TREND" "coef"};
abb6 = xlswritesa(ab6, "coin_bs", "a1", 6, "");
abbb6 = xlswritem(b1x|b1y, "coin_bs", "b2", 6, "");

@ Residuals @
ab7 = {"RESID" "vardx1" "vardx2" "vec1" "vec2" "vec3" "vec4" "vec5" "vec6" "vec7" "vec8"};
abb7 = xlswritesa(ab7, "coin_bs", "a1", 7, "");
abbb7 = xlswritem(eZ', "coin_bs", "b2", 7, "");

@ Data used in estimation @
ab8 = {"DATA" "poil" "ys" "tobq" "winf" "inf" "rus" "dsc" "cf" "invest" "y"};
abb8 = xlswritesa(ab8, "coin_bs", "a1", 8, "");
abbb8 = xlswritem(zdata, "coin_bs", "b2", 8, "");

@ A0 @
ab9 = {"A0"};
abb9 = xlswritesa(ab9, "coin_bs", "a1", 9, "");
abbb9 = xlswritem(a0, "coin_bs", "b2", 9, "");

@ Omega @
ab10 = {"Omega"};
abb10 = xlswritesa(ab10, "coin_bs", "a1", 10, "");
abbb10 = xlswritem(omega, "coin_bs", "b2", 10, "");


"Finished";

END;

@* SUB PROCEDURES *@

@ ************************************* @
  PROC (4) = TRDATA(Y,X,yorder,xorder);
@ ***************************************************************
*   Generate the appropriate Data matrices of DY, DX, Z1, DZLAGS,
*   to be used in the VEC model, depending on the order of VAR(p)
*   DY: First difference of endogenous I(1) variables
*   DX: First difference of exogenous I(1) variables
*   Z = X~Y
*   Z1: First lagged values of Z
*   ADZLAGS = DZ1~DZ2~...~DZ(p-1): T by m(p-1)
*   ************************************************************** @
Local  MM,Y1,DY,Z,Z1,adzlags,dX,DZ,DZ1,DZ2,DZ3,DZ4,DZ5,DZ6,DZ7,DZ8,
       DZ9,DZ10,DZ11;

mm = yorder + xorder;
Y1 = LAG(Y); DY = Y - Y1;

IF Xorder eq 0;   Z = Y; DX = 0; @ No exog variables: Txmy @
ELSE;             Z = X~Y;       @ Txm @
                  DX = X - LAG(X);
ENDIF;

Z1 = LAG(Z);     DZ = Z - Z1;    DZ1 = Lag(DZ);  DZ2 = Lag(DZ1);
DZ3 = Lag(DZ2);  DZ4 = Lag(DZ3); DZ5 = Lag(DZ4); DZ6 = Lag(DZ5);
DZ7 = Lag(DZ6);  DZ8 = Lag(DZ7); DZ9 = Lag(DZ8); DZ10 = Lag(DZ9);
DZ11 = Lag(DZ10);

@ all lagged values up to maximum lag order @
ADZLAGS = DZ1~DZ2~DZ3~DZ4~DZ5~DZ6~DZ7~DZ8~DZ9~DZ10~DZ11;

@* transpose the data for matching the specification *@
DY = DY';  DX = DX';  Z1 = Z1'; ADZLAGS = ADZLAGS';

RETP(DY,Z1,DX,ADZLAGS);ENDP;

@ ********************************* @
  PROC (4) = TRDATA2(idum,tdum,stt);
@ ********************************* @
@ idum1: lagged intercept shift dummy
* tdum1: lagged trend shift dummy
* ADidum = Didum~Didum1~...~Didum(p-1)
* ADtdum = Dtdum~Dtdum1~...~Dtdum(p-1)
* adding initial stt observations
* ************************************ @
Local  stmp,idum1,Di0,Di1,Di2,Di3,Di4,Di5,Di6,Di7,Di8,Di9,Di10,Di11,
       tdum1,Dt0,Dt1,Dt2,Dt3,Dt4,Dt5,Dt6,Dt7,Dt8,Dt9,Dt10,Dt11;

stmp = zeros(stt,1);
idum = stmp|idum; tdum = stmp|tdum;

idum1 = LAG(idum); 
Di0 = idum - idum1; Di1 = Lag(Di0); Di2 = Lag(Di1); Di3 = Lag(Di2); 
Di4 = Lag(Di3); Di5 = Lag(Di4); Di6 = Lag(Di5); Di7 = Lag(Di6);  
Di8 = Lag(Di7); Di9 = Lag(Di8); Di10 = Lag(Di9); Di11 = Lag(Di10);
ADidum = Di0~Di1~Di2~Di3~Di4~Di5~Di6~Di7~Di8~Di9~Di10~Di11;

tdum1 = LAG(tdum); 
Dt0 = tdum - tdum1; Dt1 = Lag(Dt0); Dt2 = Lag(Dt1); Dt3 = Lag(Dt2); 
Dt4 = Lag(Dt3); Dt5 = Lag(Dt4); Dt6 = Lag(Dt5); Dt7 = Lag(Dt6);  
Dt8 = Lag(Dt7); Dt9 = Lag(Dt8); Dt10 = Lag(Dt9); Dt11 = Lag(Dt10);
ADtdum = Dt0~Dt1~Dt2~Dt3~Dt4~Dt5~Dt6~Dt7~Dt8~Dt9~Dt10~Dt11;

@* transpose the data for matching the specification *@
idum1 = idum1'; tdum1 = tdum1';
ADidum = ADidum'; ADtdum = ADtdum';

RETP(idum1,ADidum,tdum1,ADtdum);ENDP;

@ *********************************************************** @
  PROC (2) 
  = CREG(case,Z1,inpt,TR1,id1,td1,p,DZlags,mx,DX,TR,did,dtd);
@ *********************************************************** 
* CONSTRUCTION of ECM term, ZS1 and all other regressors, YY 
* ********************************************************** @
Local zs1,yy;

IF     case eq 1; ZS1 = Z1; 
Elseif case eq 2; ZS1 = Z1|inpt;
Elseif case eq 3; ZS1 = Z1;
Elseif case eq 4; ZS1 = Z1|TR1;
Elseif case eq 6; ZS1 = Z1|TR1|id1;
Elseif case eq 7; ZS1 = Z1|TR1|td1;
ENDIF;

if p eq 1; YY = 0;
else; YY = DZLAGS;
endif;

if YY eq 0;
   if mx eq 0;
   else; YY = DX;
   endif;
else;
   if mx eq 0;
   else; YY = YY|DX;
   endif;
endif;

if YY eq 0;
   IF     case eq 3; YY = inpt;
   ELSEIF case eq 4; YY = inpt;
   ELSEIF case eq 5; YY = inpt|tr;
   ELSEIF case eq 6; YY = inpt|did;
   ELSEIF case eq 7; YY = inpt|dtd;
   ENDIF;
else;
   IF     case eq 3; YY = YY|inpt;
   ELSEIF case eq 4; YY = YY|inpt;
   ELSEIF case eq 5; YY = YY|inpt|tr;
   ELSEIF case eq 6; YY = YY|inpt|did;
   ELSEIF case eq 7; YY = YY|inpt|dtd; @ ?? @
   ENDIF;
endif;

RETP(ZS1,YY);ENDP;

@ ***************************************** @
  Proc (2) = COINTEST(DY,Z1,YY,my,mx);
@ *****************************************
* (1) OLS of DY on YY to obtain R0 and then calculate
*     S00 = (1/T)R0*R0'
* (2) OLS of Z1 on YY to obtain R1 and then calculate
*     S11 = (1/T)R1*R1',      S01 = (1/T)R0*R1'
* (3) Then, calculate
*     Bmatirx = S11
*     Amatrix = S11 - S01'Invpd(S00)*S01
* For further details see Johansen (1988, 1991) and PSS (2000)
* ************************************************************* @
local  sample,URPAI,dreg,invYY,invS00,Amatrix,Bmatrix,
       S00,S01,S11,C0,C1,R0,R1,zorder,testres,
       pp,va,ve,dd,zi,sumva,maxva,Ltrace,Lmax,sumv,cvmax,cvtrace,
       vva,vaa,im1,im2,jbeta,cc;
sample = cols(dy);

@* Z1 = (Y1 X1) and any deterministic regressors @
zorder = rows(z1);  @ dimension of z1 @

@* Computation of S00, S11 and S10 @
IF YY eq 0; R0 = DY; R1 = Z1;
ELSE;
     InvYY = Invpd(YY*YY');
     C0 = DY*YY'InvYY;
     R0 = DY - C0*YY;
     C1 = Z1*YY'InvYY;
     R1 = Z1 - C1*YY;
ENDIF;

S00 = R0*R0'./sample;
S11 = R1*R1'./sample;
S01 = R0*R1'./sample;

/* CC = invpd(S11)*s01'invpd(S00)*S01; save cc; */

@* Computation of Amatrix and Bmatrix @
InvS00 = invpd(S00);
Amatrix = S11 - S01'InvS00*S01;
Bmatrix = S11;

@* Unrestricted pi matrix @
URPAI = S01*invpd(S11);

@* Johansen Beta's using eigenvalue routine @
pp = chol(S11);
dd = inv(pp')*S01'invS00*S01*inv(pp);
{va,ve} = EIGHV(dd);

Jbeta = zeros(zorder,zorder);
zi = 1;
do until zi > zorder;
   Jbeta[.,zi] = inv(pp)*ve[.,zi];
   zi = zi + 1;
endo;

vva = zeros(zorder,1);
zi = 1;
do until zi > zorder;
   vva[zi,1] = va[zorder-zi+1,1];
   zi = zi + 1;
endo;

@* Trace and the Max Eigenvalue statistics @
vaa = zeros(zorder,1);
sumv = 0;
zi = 1;
do until zi > zorder;
   vaa[zi,1] = ln(1 - va[zi,1]);
   sumv = sumv + vaa[zi,1];
   zi = zi + 1;
endo;

sumva = zeros(my,1);
maxva = zeros(my,1);
sumva[1,1] = sumv;
maxva[1,1] = ln(1 - va[zorder,1]);
zi = 2;
do until zi > my;
   im1 = zorder + 2 - zi;
   sumva[zi,1] = sumva[zi-1,1] - vaa[im1,1];
   im2 = zorder + 1 - zi;
   maxva[zi,1] = ln(1 - va[im2,1]);
   zi = zi + 1;
endo;

Ltrace = zeros(my,1);  Lmax = zeros(my,1);
zi = 1;
do until zi > my;
   Ltrace[zi,1] = -sample*sumva[zi,1];
   Lmax[zi,1] = -sample*maxva[zi,1];
   zi = zi + 1;
endo;

VVA = VVA[1:my];
TESTRES = VVA~LTRACE~LMAX;

RETP(TESTRES,JBETA);ENDP;

@ ************************************* @
  Proc (7) = VECdy(dep,reg,sample,my,k);
@ ************************************* @
Local all,e,omega,AIC,SBC,HQ,LL;

all = dep*reg'*invpd(reg*reg'); @ parameters: my by ?? @
e = dep - all*reg;              @ residuals: my by T @
omega = e*e'./(sample-rows(reg)); @ covariance: my by my @
/* omega = e*e'./(sample-rows(reg)-1); @ Mfit: restricted trend? @ */

LL = -(sample/2)*ln(det(e*e'./sample))-(sample*my/2)*(1+ln(2*pi));
@ log-likelihood function @
AIC = LL - k;
SBC = LL - (k/2)*ln(sample);
HQ  = LL - k*ln(ln(sample));

RETP(all,e,omega,LL,AIC,SBC,HQ);ENDP;

@ ************************** @
  Proc (7) = MARdx(dep,reg);
@ ************************** @
local all,e,omega,AIC,SBC,HQ,LL,k,mx,obs;

mx = rows(dep); obs = cols(dep);
all = dep*reg'*invpd(reg*reg'); @ parameters: mx by ?? @
e = dep - all*reg;              @ residuals: mx by T @
omega = e*e'./(obs-rows(reg)); @ covariance: my by my @

if mx eq 1;
   LL = -(obs/2)*ln(e'e/obs)-(obs/2)*(1+ln(2*pi));
else;
   LL = -(obs/2)*ln(det(e*e'./obs))-(obs*mx/2)*(1+ln(2*pi));
endif;

k = rows(all)*cols(all);
AIC = LL - k;
SBC = LL - (k/2)*ln(obs);
HQ  = LL - k*ln(ln(obs));

RETP(all,e,omega,LL,AIC,SBC,HQ);ENDP;

@ ********************************************* @
  Proc (3) = VECM(ECMco,DEP,REG,RES,YHAT,nreg);
@ ********************************************* 
  Computation of estimation and diagnostic test results 
* ***************************************************** @
local diagres,sige,invreg,var,se,tratio,ti,estres;

@* computation of R2 and diagnostic statistics using VECM @
{ DIAGRES1, DIAGRES2 } = DIAGNOS(dep,reg,res,yhat,nreg);

sige = diagres1[3]^2;       @ sig^2 @
invreg = invpd(reg'reg);
var = sige.*invreg;
se = sqrt(diag(var));

tratio = zeros(rows(ecmco),1);
ti = 1;
do until ti > rows(ecmco);
   tratio[ti] = ecmco[ti]/se[ti];
   ti = ti + 1;
endo;

ESTRES = ecmco~se~tratio;

RETP(ESTRES,DIAGRES1,DIAGRES2); ENDP;

@ ****************************************** @
  PROC (2) = DIAGNOS(dep,reg,res,yhat,nreg);
@ **************************************************
* This procedure computes the following for ECM eq:
* dy(i) = phi(i)*y(i,-1) + theta(i)*x(i) + e(i),??
* (1) R-squared and Rbar-squared
* (2) Maximized log-likelihood, AIC, SBC, HQ
* (3) Diagnostic statistics such as
* Godfrey's test of residual serial correlation: chi(4)
* Ramsey's RESET test of functional form: chi(1)
* Jarque-Bera's test of the normality: chi(2)
* Test of Heterosckedasticity: chi(1)
* see Microfit manual
* ********************************************** @
Local obs,sig2,sige,LL,AIC,SBC,HQ,Rbar,RSQ,ysq2,rbarsq,
      e,x,p,mx,w,p2,p1,num,den,mu1,e2,mu2,e3,mu3,e4,mu4,
      yhat2,re,ure,R2,tmpa,tmpb,tmpc,hdep,hconst,hindep,
      chiSC,chiFF,chiNO,chiHE,diagres1,diagres2,
      tmp2a,tmp2b,wtmp,pSC,pFF,pNO,pHE;

@* computation of the likelihood, AIC, SC, HQ and Rbar square @
obs = rows(res);
sig2 = res'res/obs;
sige = res'res/(obs-nreg);   @ variance estimate with dof @
sige = sqrt(sige);           @ standard error of regression @

LL = -(obs/2)*(1+ln(2*pi*sig2));   @ log-likelihood @

AIC = LL - nreg;                   @ Akaike @
SBC = LL - (nreg/2)*ln(obs);       @ Schwarz @
HQ  = LL - nreg*ln(ln(obs));       @ Hannan-Quinn @

ysq2 = sumc((dep-meanc(dep))^2)/obs;  @ sum of squared dmeaned dep @
RSQ = 1-(sig2/ysq2);                  @ R-squared @
RBARSQ = 1-((obs-1)/(obs-nreg))*(1-RSQ); @ adjyusted R2 @

@* computation of diagnostic test statistics using the OLS residuals @
e = res;   @ Tx1 residual @
x = reg;   @ Txk indep vble @
p = 4;     @ lags used in SC test @

@* LM test for serial correlation @
@ computing W matrix for SC @
w = zeros(obs,p);   @ Txp @
p2 = 1;
do until p2 > p;
   w[obs,p+1-p2] = e[obs-p+p2-1,1];
   p1 = 1;
   do until p1 > obs-1;
      if obs-p+p2-1-p1 > 0;
         w[obs-p1,p+1-p2] = e[obs-p+p2-1-p1,1];
      endif;
      p1 = p1 + 1;
   endo;
   p2 = p2 + 1;
endo;
wtmp = w'w - w'x*invpd(x'x)*x'w;
num = e'w*invpd(wtmp)*w'e;
den = e'e/obs;
chiSC = num/den;        @ statistic @
pSC = cdfchic(chiSC,4); @ p-value @

@* LM test for functional form @
w = yhat.^2;
wtmp = w'w - w'x*invpd(x'x)*x'w;
num = e'w*invpd(wtmp)*w'e;
chiFF = num/den;
pFF = cdfchic(chiFF,1); @ p-value @

@* LM test for normality @
mu1 = sumc(e)/obs;
e2 = e.^2;
mu2 = sumc(e2)/obs;
e3 = e.^3;
mu3 = sumc(e3)/obs;
e4 = e.^4;
mu4 = sumc(e4)/obs;
tmpa = mu3^2/(6*mu2^3);
tmpb = (mu4/mu2^2 - 3)^2;
tmpc = 3*mu1^2/(2*mu2) - (mu3*mu1/mu2^2);
chiNO = obs*(tmpa+(1/24)*tmpb) + obs*tmpc;
pNO = cdfchic(chiNO,2); @ p-value @

@* LM test for heterockedasticity @
hdep   = e2;
hconst = ones(obs,1);
yhat2 = yhat.^2;
hindep = hconst~yhat2;
ure = hdep - hindep*invpd(hindep'hindep)*hindep'hdep; @ unrestricted res @
re = hdep - hconst*invpd(hconst'hconst)*hconst'hdep;  @ restricted res @
R2 = (re're - ure'ure)/re're;
chiHE = obs*R2;
pHE = cdfchic(chiHE,1); @ p-value @

DIAGRES1 = RSQ|RBARSQ|SIGE|LL|AIC|SBC|HQ; @ 7x1 @
tmp2a = chiSC~chiFF~chiNO~chiHE; 
tmp2b = pSC~pFF~pNO~pHE; 
DIAGRES2 = tmp2a|tmp2b; @ 2x4 @

RETP(DIAGRES1,DIAGRES2);ENDP;

@ ********************************* @
  Proc (0) = PRECM(EST,DIAG1,DIAG2);
@ ********************************* @
local pval1,pval2,pval3,pval4,rr2,rr3;

"    Coeff.   Std. Err  T-ratio";"";
call printfm(est,1~1~1,"lf"~10~4|"lf"~10~4|"lf"~10~4);"";
"Diagnostic Statistics";"";
" RSQ    RBARSQ   SIGE     LL    AIC    SBC     HQ";"";
rr2 = diag1';
call printfm(rr2,1~1~1~1~1~1~1,"lf"~7~3|"lf"~7~3|"lf"~7~3|"lf"~7~2|"lf"~7~2
               |"lf"~7~2|"lf"~7~2);"";
rr3 = diag2;
"Ch-SC  Ch-FF  Ch-NO  Ch-HE";"";
call printfm(rr3,1~1~1~1,"lf"~7~2|"lf"~7~2|"lf"~7~2|"lf"~7~2);"";"";

retp;endp;

@ ******************************************** @
  Proc (9) = decompY(ally,r,p,my,mx,case,beta);
@ ******************************************** @
@ Decomposition into separate parameter blocks for VEC for dy 
* Remind: the order of regressors in VECM is:
* ecm(-1)|dzlags|DX|deterministics where
* the dimensions (row) are r, m(p-1), mx, and 1,2 or 3 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummies 
* whose dimension is equal to p 
* ***************************** @
Local Alp,Gam,Lam,b0,b1,d0,d1,dd,dt,i1,i2,m;

m = my + mx;
if r eq 0;
   Alp = zeros(my,r); @ no coin @
else;
   i1 = 1; i2 = i1+r-1;
   Alp = allY[.,i1:i2];
endif;

if p eq 1; 
   Gam = 0; @ lag order @
else; 
   i1 = r+1;  i2 = i1+(p-1)*m-1;
   Gam = allY[.,i1:i2];  @ my by m(p-1)  @
endif;

if mx eq 0;  @ xorder @
   Lam = 0;
else;  
   i1 = r+(p-1)*m+1; i2 = i1+mx-1;
   Lam = allY[.,i1:i2];  @ my by mx @
endif;

b0 = zeros(my,1);  @ intercepts @
b1 = zeros(my,1);  @ trend coefficients @
d0 = zeros(my,1);  @ intercept dummy coefficients @
d1 = zeros(my,1);  @ trend dummy coefficients @
dd = zeros(my,p);  @ coefficients of FD int dum @
dt = zeros(my,p);  @ coefficients of FD tr dum @

if     case eq 2;  
       if r eq 0;
       else;  b0 = alp*beta[m+1,.]';@ restricted trends @
       endif;
elseif case eq 3;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; 
elseif case eq 4;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]'; @ restricted trends @
       endif;
elseif case eq 5;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; b1 = allY[.,i1+1];   
elseif case eq 6;
       i1 = r+(p-1)*m+mx+1; 
       b0 = allY[.,i1]; dd = allY[.,i1+1:i1+p]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]';  @ restricted trends @
             d0 = alp*beta[m+2,.]';  @ restricted int dum @
       endif;
elseif case eq 7;
       i1 = r+(p-1)*m+mx+1;
       b0 = allY[.,i1]; dt = allY[.,i1+1:i1+p]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]';  @ restricted trends @
             d1 = alp*beta[m+2,.]';  @ restricted int dum @
       endif;
endif;

RETP(Alp,Gam,Lam,b0,b1,d0,d1,dd,dt);ENDP;

@ ************************************** @
  Proc (4) = decompX(allx,p,my,mx,case);
@ ************************************** @
@ Decomposition into separate parameter blocks for VAR for dx:
* Remind: the order of regressors in VAR is: dzlags|deterministics 
* where the dimensions (row) are m(p-1) and 1,2 or plus p. 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummies 
* whose dimension is equal to p 
* ***************************** @
Local Gam,b0,b1,dd,dt,i1,i2,m;

m = my+mx;
if p eq 1; @ lag order @
   Gam = 0; 
else; 
   i1 = 1;  i2 = i1+(p-1)*m-1;
   Gam = allX[.,i1:i2];  @ mx by m(p-1)  @
endif;

b0 = zeros(mx,1);  @ intercepts @
b1 = zeros(mx,1);  @ trends @
dd = zeros(mx,p);  @ coefficients of FD int dum @
dt = zeros(mx,p);  @ coefficients of FD tr dum @

if     case eq 3;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 4;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 5;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; b1 = allX[.,i1+1];   
elseif case eq 6;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; dd = allX[.,i1+1:i1+p]; 
elseif case eq 7;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; dt = allX[.,i1+1:i1+p]; 
endif;

RETP(Gam,b0,dd,dt);ENDP;

@ ******************************** @
  PROC(3) = DYPARA(pai,gam,p,m,n);
@ ********************************
* Computation of MA parameters
* FAI: FAI(1),...,FAI(p), m by m*p
* FF: F matrix, mp by mp augmented matrix
* BN: m by m MA matrix
* **************************************** @
local  pgam,kk,jj,fai,ff,bn,ki,fi,ktmp1,ktmp2,ktmp3,
       Ftmp1,Ftmp2,Ftmp3,bi,bi1,btemp;

@* Computation of Fai matrix @
if p eq 1; PGAM = PAI;
else;      PGAM = PAI~GAM;   @ m by pm @
endif;

if p eq 1; KK = -eye(m); JJ = eye(m);
else;
   Ktmp1 = -eye(m*p);
   Ktmp2 = zeros(p,p);
   ki = 1;
   do until ki > p-1;
      Ktmp2[ki+1,ki] = 1;
      ki = ki + 1;
   endo;
   Ktmp3 = Ktmp2.*.eye(m);

   KK = Ktmp1 + Ktmp3;
   JJ = eye(m)~zeros(m,m*(p-1));
endif;
FAI = PGAM*KK + JJ;  @ FAI1,...,FAIp: m by m*p @

if p eq 1;  FF = FAI;
else;
   ftmp1 = FAI|zeros((p-1)*m,m*p);
   ftmp2 = zeros(p,p);
   fi = 1;
   do until fi > p-1;
      ftmp2[fi+1,fi] = 1;
      fi = fi + 1;
   endo;
   ftmp3 = ftmp2.*.eye(m);
   FF = ftmp1 + ftmp3;
endif;

@* Computation of Bn matrix (MA) @
BN = zeros(m,n*m);  @ MA coefficients @
btemp = eye(m*p);
bi = 1;
do until bi > n;
   btemp = FF*btemp;
   bi1 = (bi-1)*m+1;
   BN[.,bi1:bi1+m-1] = JJ*btemp*JJ';
   bi = bi + 1;
endo;

RETP(FAI,FF,BN);ENDP;

@ *********************************************** @
  Proc (2) = CHKSTAB(m,rcoin,p,alpha,beta,gam,n);
@ *********************************************** @
local  Amat,bet,aj,nGam,Gam0,Gamj,Amatj,eigstab,ve,Ir,Zr,gtmp,
       ggtmp1,ggtmp2,ggtmp,atmp1,atmp2,atmp,AN,GN,BigA,va,jg,gg,gi,
       gi1,gi2,rj0,neig,vatmp,modva,modvatmp;

Ir = eye(rcoin); Zr = zeros(rcoin,rcoin);
if p eq 1;
   Amat = Ir + beta'alpha;
   va = eig(Amat);
   va = sortc(va,1);
   va = rev(vatmp);  @ descending order @
   { modva } = MODULUS(va);

elseif p > 1;
   if p eq 2;     JG = eye(m);
                       GG = Gam;
   elseif p > 2;  JG = eye(m)~zeros(m,m*(p-2));
                       GGtmp1 = eye(m*(p-2));
                       GGtmp2 = zeros(m*(p-2),m);
                       GGtmp = GGtmp1~GGtmp2;
                       GG = Gam|GGtmp;
   endif;

   GN = zeros(m,m*n);  @ MA matrix: G1,G2,...,GN @
   gtmp = eye(m*(p-1));
   gi = 1;
   do until gi > n;
      gtmp = GG*gtmp;
      gi1 = (gi-1)*m+1;
      GN[.,gi1:gi1+m-1] = JG*gtmp*JG';
      gi = gi + 1;
   endo;

   AN = zeros(rcoin,rcoin*n);  @ A1,A2,...,AN @
   AN[.,1:rcoin] = Ir + beta'alpha;  @ A1 @
   gi = 2;
   do until gi > n;
      gi1 = (gi-1)*rcoin+1;
      gi2 = (gi-2)*m+1;
      AN[.,gi1:gi1+rcoin-1] = beta'GN[.,gi2:gi2+m-1]*alpha;
      gi = gi + 1;
   endo;

   atmp1 = eye(rcoin*(n-1));
   atmp2 = zeros(rcoin*(n-1),rcoin);
   atmp = atmp1~atmp2;
   BigA = AN|atmp;   @ Big A matrix: rl by rl @

   va = zeros(rcoin*n,n);  @ eigenvalues of BigA @
   modva = zeros(rcoin*n,n);  @ modulus of BigA @
   aj = 1;
   do until aj > n;
      rj0 = aj*rcoin;
      Amat = BigA[1:rj0,1:rj0];  @ rl by rl @
      neig = rows(Amat);  @ number of eigenvalues @
      vatmp = eig(Amat);
      vatmp = sortc(vatmp,1);
      vatmp = rev(vatmp);  @ descending order @
      { modvatmp } = MODULUS(vatmp);
      va[1:neig,aj] = vatmp;
      modva[1:neig,aj] = modvatmp;
      aj = aj + 1;
   endo;
endif;

RETP(va,modva);ENDP;

@ *************************** @
  Proc (1) = MODULUS(vatemp);
@ *************************** @
local rva,iva,vj,modvatmp;

rva = real(vatemp);  @ real part of eigenvalues @
iva = imag(vatemp);  @ imaginary part of eigenvalues @

modvatmp = zeros(rows(vatemp),1);
vj = 1;
do until vj > rows(vatemp);
   modvatmp[vj] = sqrt(rva[vj]^2 + iva[vj]^2);
   vj = vj + 1;
endo;

modvatmp = sortc(modvatmp,1);
modvatmp = rev(modvatmp);  @ descending order @

retp(modvatmp);endp;

@ ************************************* @
   PROC (1) = persis(m,n,BETA,OMEGA,BN);
@ *************************************
*  Calculate persistence profiles of LR relations,
*  hz(i) = (beta'Bi*Omega*Bi'beta)/(beta'Omega*beta)
*  Here Omega is correlated.
* ************************************************** @
LOCAL   hz,DENOM,jp,temp,temp1,ro,co,vect,j1;

BETA = BETA.*1000;    @ hz is invariant to scale @
DENOM = beta'OMEGA*beta;

hz = zeros(n,1);
temp = zeros(m,m);
jp = 1;
do until jp > n;
   j1 = (jp-1)*m+1;
   temp = BN[.,j1:j1+m-1];
   temp1 = beta'(temp*OMEGA*temp')*beta;
   hz[jp,1] = temp1/Denom;
   jp = jp + 1;
endo;

RETP(hz);ENDP;

@ ******************************************************* @
  PROC (4) = Impulse(sh_op,m,n,RCOIN,BETA,gOMEGA,BN,IA0);
@ *******************************************************
* GIR (OIR) is the generalized (orthogonalised) impulse 
* responses of m variables, whilst GIRZ (OIRZ) is GIR (OIR) 
* of r cointegrating relations with respect to shocks in jth 
* equation, with h = 1,2,... and j = 1 to m.
* ******************************************************* @
LOCAL   evec,sigjj,gir0,gir,girz0,girz,zi,j1,bj,chk1,chk2,
        upp,lpp,ipp,oir0,oir,oirz0,oirz;

evec = zeros(m,1);
evec[sh_op] = 1;    @ selection vector @
sigjj = evec'gomega*evec;

GIR0  = IA0*gOmega*evec./sqrt(sigjj);    @ impact impulse @
GIRZ0 = beta'IA0*gOmega*evec./sqrt(sigjj);

UPP = Chol(gomega); LPp = UPp';
Ipp = Inv(LPp);
OIR0  = IA0*LPP*evec;           @ impact impulse @
OIRZ0 = beta'IA0*LPP*evec;

GIR = zeros(m,n); GIRZ = zeros(rcoin,n);     
OIR = zeros(m,n); OIRZ = zeros(rcoin,n);     
zi = 1;
DO until zi > n;
   j1 = (zi-1)*m + 1;
   Bj = BN[.,j1:j1+m-1];
   GIR[.,zi]  = Bj*IA0*gOmega*evec./sqrt(sigjj);
   GIRZ[.,zi] = beta'Bj*IA0*gOmega*evec./sqrt(sigjj);
   OIR[.,zi]  = Bj*IA0*LPP*evec;
   OIRZ[.,zi] = beta'Bj*IA0*LPP*evec;
   zi = zi + 1;
ENDO;

GIR = GIR0~GIR;   GIRZ = GIRZ0~GIRZ;
OIR = OIR0~OIR;   OIRZ = OIRZ0~OIRZ;

@ transposing the results such that nh+1 (or r) by m @
GIR = GIR'; GIRZ = GIRZ'; OIR = OIR'; OIRZ = OIRZ';

RETP(GIR,GIRZ,OIR,OIRZ);ENDP;

@ ************************************************* @
  proc (14) = STIRF(m,n,r,BETA,VOMEGA,BN,IA11,W11);
@ *************************************************
  SIR is structural IRF of m variables,
  SIRZ is structural IRF of r coin relations wrt
  shocks in jth equation, with n = 1,2,... and j = 1 to m.
* ******************************************************** @
local A11,ovec,ysvec,tqvec,psvec,wivec,invec,rvec, ivec,zi,j1,bj,
      saie,sair,deltar,osir0a,osir0b,osir0,osirz0,yssir0a,yssir0b,
      yssir0,yssirz0,tqsir0a,tqsir0b,tqsir0,tqsirz0,wisir0a,wisir0b,
      wisir0,wisirz0,insir0a,insir0b,insir0,insirz0,rsir0a,rsir0b,
      rsir0,rsirz0,isir0a,isir0b,isir0,isirz0,OSIR,OSIRZ,YSSIR,
      YSSIRZ,TQSIR,TQSIRZ,PSSIR,PSSIRZ,WISIR,WISIRZ,INSIR,INSIRZ,
      RSIR,RSIRZ,ISIR,ISIRZ;

@ Now n_or = 6, when it changes, the proc should change accordingly. @ 
A11 = Inv(IA11);

@* Oil price shock @
ovec = zeros(6,1); ovec[1] = W11[1,1];
OSIR0A = IA11*ovec;
OSIR0B = Vomega[7:m,1];
OSIR0 = OSIR0A|OSIR0B; @ m by 1 @
OSIR0 = OSIR0./sqrt(ovec[1]);
OSIRZ0 = beta'OSIR0;

@* Potential output (productivity?) shock @
ysvec = zeros(6,1); ysvec[2] = W11[2,2];
YSSIR0A = IA11*ysvec;
YSSIR0B = Vomega[7:m,2] + A11[2,1].*Vomega[7:m,1];
YSSIR0 = YSSIR0A|YSSIR0B;
YSSIR0 = YSSIR0./sqrt(ysvec[2]);
YSSIRZ0 = beta'YSSIR0;

@* Tobin's Q (stock market?) shock @
tqvec = zeros(6,1); tqvec[3] = W11[3,3];  
TQSIR0A = IA11*tqvec;
TQSIR0B = Vomega[7:m,3] + A11[3,2].*Vomega[7:m,2] 
        + A11[3,1].*Vomega[7:m,1];
TQSIR0 = TQSIR0A|TQSIR0B;
TQSIR0 = TQSIR0./sqrt(tqvec[3]);
TQSIRZ0 = beta'TQSIR0;

@* Wage inflation shock @
wivec = zeros(6,1); wivec[4] = W11[4,4];
WISIR0A = IA11*wivec;
WISIR0B = Vomega[7:m,4] + A11[4,3].*Vomega[7:m,3] 
       + A11[4,2].*Vomega[7:m,2] + A11[4,1].*Vomega[7:m,1];
WISIR0 = WISIR0A|WISIR0B;
WISIR0 = WISIR0./sqrt(wivec[4]);
WISIRZ0 = beta'WISIR0;

@* Inflation shock @
invec = zeros(6,1); invec[5] = W11[5,5];
INSIR0A = IA11*invec;
INSIR0B = Vomega[7:m,5] + A11[5,4].*Vomega[7:m,4] 
       + A11[5,3].*Vomega[7:m,3] + A11[5,2].*Vomega[7:m,2] 
       + A11[5,1].*Vomega[7:m,1];
INSIR0 = INSIR0A|INSIR0B;
INSIR0 = INSIR0./sqrt(invec[5]);
INSIRZ0 = beta'INSIR0;

@* Interest rate (monetary policy) shock @
rvec = zeros(6,1); rvec[6] = W11[6,6];
RSIR0A = IA11*rvec;
RSIR0B = Vomega[7:m,6] + A11[6,5].*Vomega[7:m,5] 
       + A11[6,4].*Vomega[7:m,4] + A11[6,3].*Vomega[7:m,3] 
       + A11[6,2].*Vomega[7:m,2] + A11[6,1].*Vomega[7:m,1];
RSIR0 = RSIR0A|RSIR0B;
RSIR0 = RSIR0./sqrt(rvec[6]);
RSIRZ0 = beta'RSIR0;

@* interest intercept shift @
ISIR0 = zeros(m,1);
ISIR0[6] = sqrt(W11[6,6]); @ exogenous impact @
ISIRZ0 = beta'ISIR0;

OSIR = zeros(m,n);  OSIRZ = zeros(r,n);
YSSIR = zeros(m,n); YSSIRZ = zeros(r,n);
TQSIR = zeros(m,n); TQSIRZ = zeros(r,n);
PSSIR = zeros(m,n);  PSSIRZ = zeros(r,n);
WISIR = zeros(m,n);  WISIRZ = zeros(r,n);
INSIR = zeros(m,n);  INSIRZ = zeros(r,n);
RSIR = zeros(m,n);  RSIRZ = zeros(r,n);
ISIR = zeros(m,n);  ISIRZ = zeros(r,n);
zi = 1;
do until zi > n;
   j1 = (zi-1)*m + 1;
   Bj = BN[.,j1:j1+m-1];
   OSIR[.,zi] = Bj*OSIR0;   OSIRZ[.,zi] = beta'Bj*OSIR0;
   YSSIR[.,zi] = Bj*YSSIR0; YSSIRZ[.,zi] = beta'Bj*YSSIR0;
   TQSIR[.,zi] = Bj*TQSIR0; TQSIRZ[.,zi] = beta'Bj*TQSIR0;
   WISIR[.,zi] = Bj*WISIR0;   WISIRZ[.,zi] = beta'Bj*WISIR0;
   INSIR[.,zi] = Bj*INSIR0;   INSIRZ[.,zi] = beta'Bj*INSIR0;
   RSIR[.,zi] = Bj*RSIR0;   RSIRZ[.,zi] = beta'Bj*RSIR0;
   ISIR[.,zi] = Bj*ISIR0;   ISIRZ[.,zi] = beta'Bj*ISIR0;
zi = zi + 1;
ENDO;

OSIR = OSIR0~OSIR;    OSIRZ = OSIRZ0~OSIRZ;
YSSIR = YSSIR0~YSSIR; YSSIRZ = YSSIRZ0~YSSIRZ;
TQSIR = TQSIR0~TQSIR; TQSIRZ = TQSIRZ0~TQSIRZ;
WISIR = WISIR0~WISIR;    WISIRZ = WISIRZ0~WISIRZ;
INSIR = INSIR0~INSIR;    INSIRZ = INSIRZ0~INSIRZ;
RSIR = RSIR0~RSIR;    RSIRZ = RSIRZ0~RSIRZ;
ISIR = ISIR0~ISIR;    ISIRZ = ISIRZ0~ISIRZ;

@* transposing the results such that nh by m (or r) @

OSIR = OSIR';   OSIRZ = OSIRZ';
YSSIR = YSSIR'; YSSIRZ = YSSIRZ';
TQSIR = TQSIR'; TQSIRZ = TQSIRZ';
WISIR = WISIR';   WISIRZ = WISIRZ';
INSIR = INSIR';   INSIRZ = INSIRZ';
RSIR = RSIR';   RSIRZ = RSIRZ';
ISIR = ISIR';   ISIRZ = ISIRZ';

retp(OSIR,OSIRZ,YSSIR,YSSIRZ,TQSIR,TQSIRZ,WISIR,WISIRZ,INSIR,INSIRZ,RSIR,RSIRZ,ISIR,ISIRZ);
ENDP;

@ ****************************** @
  proc (2) = TRDATA4(Zd,my,mx,p);
@ ****************************** @
local  z,Z1,adzlags,DZ,DZ1,DZ2,DZ3,DZ4,DZ5,DZ6,DZ7,DZ8,
       DZ9,DZ10,DZ11,ZT,DZLAGST,m;
z=zd;
Z1 = LAG(Z);     DZ = Z - Z1;    DZ1 = Lag(DZ);  DZ2 = Lag(DZ1);
DZ3 = Lag(DZ2);  DZ4 = Lag(DZ3); DZ5 = Lag(DZ4); DZ6 = Lag(DZ5);
DZ7 = Lag(DZ6);  DZ8 = Lag(DZ7); DZ9 = Lag(DZ8); DZ10 = Lag(DZ9);
DZ11 = Lag(DZ10);

@ all lagged values up to maximum lag order @
m = my+mx;
ADZLAGS = DZ~DZ1~DZ2~DZ3~DZ4~DZ5~DZ6~DZ7~DZ8~DZ9~DZ10;
DZLAGST = ADZLAGS[rows(ADZLAGS),1:m*(p-1)]'; @ m(p-1) by 1 @
ZT = Z[rows(Z),.]';  @ m by 1 @

RETP(ZT,DZLAGST);ENDP;

@ ************************************* @
  Proc (1) = REGXT(DZT,case,sample,k,p); 
@ ************************************* @
Local XregT,k1;

@* Case A: Unrestricted  @
k1 = sample+k; @ T+k-1 @
XregT = DZT; @ m(p-1) by 1 at T @
if     case eq 3; XregT = XregT|1; 
elseif case eq 4; XregT = XregT|1; 
elseif case eq 5; XregT = XregT|1|k1; 
elseif case eq 6; XregT = XregT|1|zeros(p,1); 
elseif case eq 7; XregT = XregT|1|ones(p,1); 
endif;

RETP(XregT);ENDP; 

@ ****************************************************** @
  proc (1) = REGYT(Z1T,DZLT,DXT,case,p,mx,beta,sampl,k,h); 
@ ****************************************************** @
Local k1,k2,k3,ZS1T,YregT;

k1 = sampl+k; @ T+k @
k2 = sampl+k-1; @ T+k-1 @
k3 = h+k-1; @ Tdum+k-1 @

ZS1T = Z1T;
IF     case eq 2; ZS1T = ZS1T|1;
Elseif case eq 3; ZS1T = ZS1T;
Elseif case eq 4; ZS1T = ZS1T|k2;
Elseif case eq 6; ZS1T = Z1T|k2|1;
Elseif case eq 7; ZS1T = Z1T|k2|k3;
ENDIF;

YregT = beta'ZS1T; @ r by 1 @

if p eq 1; YregT = YregT|DXT;
else; YregT = YregT|DZLT|DXT;
endif;

IF     case eq 3; YregT = YregT|1;
ELSEIF case eq 4; YregT = YregT|1;
ELSEIF case eq 5; YregT = YregT|1|k1;
ELSEIF case eq 6; YregT = YregT|1|zeros(p,1);
ELSEIF case eq 7; YregT = YregT|1|ones(p,1); @ ?? @
ENDIF;

RETP(YregT);ENDP; 

@ ******************************** @
  Proc (2) = A4CHANGE(lhat,z,n,m);
@ ******************************** @
local a4hat,ahat,zi,zt4,suml,a4tmp,zj,ztmp,obs;

@* Forecasts of annual 4-qt MA and average over n @
obs = rows(z);
zt4 = z[obs-3:obs,.];  @ initial 4 by m @
A4hat = zeros(n,m); Ahat = zeros(n,m);  suml = zeros(1,m);
zi = 1;
do until zi > n;
   zt4 = zt4[2:4,.]|Lhat[zi,.];
   A4tmp = zeros(1,m);
   zj = 1;
   do until zj > m;
      ztmp = zt4[.,zj];
      A4tmp[zj] = meanc(ztmp);
      zj = zj + 1;
   endo;
   A4hat[zi,.] = A4tmp;
   sumL = sumL + Lhat[zi,.];
   Ahat[zi,.] = sumL./zi;
   zi = zi + 1;
endo;

RETP(a4hat,ahat); ENDP;

@ ************************************************************************* @
  Proc (2) = stocfor(ZD,sample,n,mx,my,p,case,hh,beta,allY,allX,Lpz,s,v,rn);
@ ************************************************************************* 
* h-step ahead point forecasts plus future forecast errors
* rn is m by n vector of N(0,1) 
* ***************************** @
local m,sdz,sz,h,k,Z1T,DZlagsT,XregT,DXT,YregT,DYT,DZT,ZT,rhat,rxhat,ryhat,rnum;

m = mx + my;
SDZ = zeros(n,m); @ forecasts of dz: n by m @
SZ  = zeros(n,m); @ forecasts of z:  n by m @
h = 1;
do until h > n;

@* draws of future errors (uncertainty): jointly @
   if     s eq 1; rhat = rn[.,h];
   elseif s eq 2; rnum = rndu(1,1)*sample;
                  rhat = v[.,ceil(rnum)];
   endif;
   rhat = lpz*rhat;  @ m by 1 @ 
/*  rhat = zeros(m,1); */
   rxhat = rhat[1:mx]; ryhat = rhat[mx+1:m];

@* data transformation for forecasting @
   { Z1T, DZlagsT } = TRDATA4(Zd,my,mx,p);

@* generating the regressors for VAR dx @
   k=h;
   { XregT } = REGXT(DZlagsT,case,sample,k,p); 

@* forcast of DX at T+1: mx by 1 @
   DXT = allX*XregT + rxhat; 

@* generating the regressors for VEC dy @
   { YregT } 
   = REGYT(Z1T,DZlagsT,DXT,case,p,mx,beta,sample,k,hh);

@* forcast of DY at T+1: my by 1 @
   DYT = allY*YregT + ryhat; 

@* forcast of DZ and Z at T+j, m by 1 @
   DZT = DXT|DYT;  ZT = DZT + Z1T; 

@* saving forcasts of DZ and Z at h=1,...,n, m by 1 @
   SDZ[h,.] = DZT'; SZ[h,.] = ZT';

@* updating Z @
   ZD = ZD|ZT'; 

   h = h + 1;
endo;

RETP(SZ,SDZ);ENDP;

@ ********************************************************** @
  Proc (8) = PREV2(dz,d2z,a4dz,a4d2z,adz,ad2z,a4adz,a4ad2z);
@ ********************************************************** 
* Probility evaluation for 9 events for directional change and
* integral transforms: see the def for 9 events above 
* Currently, the variables in the order of
* po,rs,ps,e,r,dp,y,pps,hy
* ************************ @
local fdd,fdd2,a4fdd,a4fdd2,fid,fid2,a4fid,a4fid2,fj;

m = cols(dz); @ dimension @

@ increase the precision @
dz = 10000.*dz; d2z = 10000.*d2z;
a4dz = 10000.*a4dz; a4d2z = 10000.*a4d2z;
adz = 10000.*adz; ad2z = 10000.*ad2z;
a4adz = 10000.*a4adz; a4ad2z = 10000.*a4ad2z;

@ dummies for directional changes and integral transforms @
fdd = zeros(1,m);   @ direction of diff, quarterly @
fdd2 = zeros(1,m);  @ direction of 2nd diff @
a4fdd = zeros(1,m);   @ direction of diff, 4-qt MA @
a4fdd2 = zeros(1,m);  @ direction of 2nd diff @
fid = zeros(1,m);   @ integral transform of diff, quarterly @
fid2 = zeros(1,m);  @ integral transform of 2nd diff @
a4fid = zeros(1,m);   @ integral transform of diff, 4-qt MA @
a4fid2 = zeros(1,m);  @ integral transform of 2nd diff @

fj=1;
do until fj > m;

   if dz[fj] > 0; fdd[fj]=1;
   endif;
   if d2z[fj] > 0; fdd2[fj]=1;
   endif;
   if a4dz[fj] > 0; a4fdd[fj]=1;
   endif;
   if a4d2z[fj] > 0; a4fdd2[fj]=1;
   endif;

   if dz[fj] > adz[fj]; fid[fj]=1;
   endif;
   if d2z[fj] > ad2z[fj]; fid2[fj]=1;
   endif;
   if a4dz[fj] > a4adz[fj]; a4fid[fj]=1;
   endif;
   if a4d2z[fj] > a4ad2z[fj]; a4fid2[fj]=1;
   endif;

   fj=fj+1;
endo;

RETP(fdd,fdd2,a4fdd,a4fdd2,fid,fid2,a4fid,a4fid2);ENDP;

@ ****************************** @
  proc (5) = HITCF(fo,act,n,m);
@ ******************************
*  Counting hits and misses:
*  1 = UD, 2 = DD, 3 = DU, 4 = UU
*  for: central forecasts
*  **************************** @
Local  UD,DD,DU,UU,HPR,i,j,HITFOR,den,num,tmp1,tmp2,
       hrfor,ksfor,ptf,sumud,sumdd,sumdu,sumuu,sumdr;

HPR = zeros(n,m);  @ hit and miss: 1=UD, 2=DD, 3=DU, 4=UU @
UD = zeros(m,1); DD = zeros(m,1); DU = zeros(m,1); UU = zeros(m,1);

i=1;
do until i > n;  @ number of horizons @
   j=1;
   do until j > m; @ number of events @

      IF fo[i,j] > 0;
         if act[i,j] < 0; hpr[i,j]=1; ud[j] = ud[j]+1;
         else;            hpr[i,j]=4; uu[j] = uu[j]+1;
         endif;
      ELSE;
         if act[i,j] < 0; hpr[i,j]=2; dd[j] = dd[j]+1;
         else;            hpr[i,j]=3; du[j] = du[j]+1;
         endif;
      ENDIF;

      j=j+1;
   endo;
   i=i+1;
endo;

@* SUMMARY results for ups and downs @
HITFOR = UD~DD~DU~UU;  @ m by 4 @

sumUD = sumc(UD);
sumDD = sumc(DD);
sumDU = sumc(DU);
sumUU = sumc(UU);

@* hit ratio = (DD+UU)/(UD+DD+DU+UU) @
den = sumUD + sumDD + sumDU + sumUU;
num = sumDD + sumUU;
hrfor = num/den;

@ Kuipers Score = H-F,
* where H = UU/(UU+UD) and F = DU/(DU+DD) 
* *************************************** @
tmp1 = sumUU/(sumUU+sumUD);
tmp2 = sumDU/(sumDU+sumDD);
ksfor = tmp1-tmp2;

@* Pesaran-Timmerman test @
{ ptf } = PT(UD,DD,DU,UU);

format /rd 4,0;
"Forecast Evaluation";
"  UD   DD   DU   UU";
HITFOR;
"Sum";
sumdr=sumud~sumdd~sumdu~sumuu;sumdr;"";
format /rd 6,3;"hit rate: ";; hrfor;
"Kuipers score: ";; ksfor;
"Pesaran-Timmerman test: ";; ptf;"";

retp(HITFOR,SUMDR,HRFOR,KSFOR,PTF);endp;

@ ****************************** @
  proc (5) = HITPR(spr,act,n,m);
@ ******************************
* Counting hits and misses:
* 1 = UD, 2 = DD, 3 = DU, 4 = UU
* pr: prob forecast with future and/or parameter uncertainty
* ************************************************************ @
Local  UD,DD,DU,UU,HPR,i,j,HITSPR,den,num,tmp1,tmp2,
       hrSPR,ksSPR,pts,sumud,sumdd,sumdu,sumuu,sumipr;

HPR = zeros(n,m);  @ hit and miss: 1=UD, 2=DD, 3=DU, 4=UU @
UD = zeros(m,1); DD = zeros(m,1); DU = zeros(m,1); UU = zeros(m,1);

i=1;
do until i > n;  @ number of horizons @
   j = 1;
   do until j > m; @ number of events @

      IF spr[i,j] > 0.5;
         if act[i,j] < 0; hpr[i,j]=1; ud[j] = ud[j] + 1;
         else;            hpr[i,j]=4; uu[j] = uu[j] + 1;
         endif;
      ELSE;
         if act[i,j] < 0; hpr[i,j] = 2; dd[j] = dd[j] + 1;
         else;            hpr[i,j] = 3; du[j] = du[j] + 1;
         endif;

      ENDIF;

      j=j+1;
   endo;
   i=i+1;
endo;

@* SUMMARY results for ups and downs @
HITSPR = UD~DD~DU~UU;  @ m by 4 @

sumUD = sumc(UD);
sumDD = sumc(DD);
sumDU = sumc(DU);
sumUU = sumc(UU);

@* hit ratio = (DD+UU)/(UD+DD+DU+UU) @
den = sumUD + sumDD + sumDU + sumUU;
num = sumDD + sumUU;
hrSPR = num/den;

@ Kuipers Score = H-F,
* where H = UU/(UU+UD) and F = DU/(DU+DD) 
* *************************************** @
tmp1 = sumUU/(sumUU+sumUD);
tmp2 = sumDU/(sumDU+sumDD);
ksSPR = tmp1-tmp2;

@* Pesaran-Timmerman test @
{ pts } = PT(UD,DD,DU,UU);

sumipr=sumud~sumdd~sumdu~sumuu; 

format /rd 4,0;
"Forecast Evaluation";
"  UD   DD   DU   UU"; HITSPR;
"Sum"; sumipr;"";
format /rd 6,3;"hit rate: ";; hrSPR;
"Kuipers score: ";; ksSPR;
"Pesaran-Timmerman test: ";; pts;"";

retp(HITSPR,SUMIPR,HRSPR,KSSPR,PTS);endp;

@ *************************** @
   PROC (1) = PT(UD,DD,DU,UU);
@ ***************************
* Computing the Pesaran-Timmerman test
* ************************************ @
Local n,p,py,px,ps,varp,varps,pttest;

n = sumc(UD) + sumc(DD) + sumc(DU) + sumc(UU);

p = (sumc(DD)+sumc(UU))/n;
py = (sumc(UD)+sumc(UU))/n;
px = (sumc(DU)+sumc(UU))/n;
ps = py*px + (1-py)*(1-px);

varp = ps*(1-ps)/n;
varps = (((2*py-1)^2)*px*(1-px)/n) + (((2*px-1)^2)*py*(1-py)/n)
      + (4*py*px*(1-py)*(1-px)/n^2);

pttest = (p-ps)/sqrt(varp-varps);

RETP(pttest);ENDP;

@ ************************ @
   proc (1) = KS(IPR,n,m);
@ ************************
* Computing the Kolmogorov-Smirnov test:
* KS = sup_x|F_n(x)-U(x)|, where F_n(x) is the empirical CDF of 
* the probability integral transforms and U(x) is the CDF of U[0,1].
* ****************************************************************** @
Local  unif,schk,bchk,chk,semp,bemp,cemp,i,j,sks,bks,cks,nm;

nm = n*m;
unif = seqa(1,1,nm); unif = unif./nm; @ uniform cdf @
chk = unif; @ for checking @
schk = vec(ipr); schk = sortc(schk,1);

semp = zeros(nm,1);
cemp = zeros(nm,1);
i=1;
do until i > nm;
   j=1;
   do until j > nm;

      if schk[i] < unif[j];
      else;  semp[i] = semp[i] + 1;
      endif;
      if chk[i] < unif[j];
      else; cemp[i] = cemp[i] + 1;
      endif;

      j=j+1;
   endo;

   i=i+1;
endo;

semp = semp./nm; cemp = cemp./nm;
sks = semp - unif; cks = cemp - unif;
sks = maxc(abs(sks)); cks = maxc(abs(cks));

format /rd 6,3;
"Kolmogorov-Smirnov statistic";;sks;

retp(sks);endp;

@ **************************************************** @
  Proc (7) = PREVENT(sz,sdz,actz,actdz,n,m,DZT,popgr);
@ **************************************************** @
local a4sz,asz,a4sdz,asdz,flag1,flag2,flag3,flag4,flag5,flag6,flag7,
      hat1,hat2,hat3,inidy;

@* constructing 4-quarter moving averages and h-step avg @
{ a4sz, asz }   = A4CHANGE(sz,actz,n,m);   @ levels: n by m @
{ a4sdz, asdz } = A4change(sdz,actdz,n,m); @ 1st differences @

@* Probility evaluation for 7 events: E1-E7 @
@* E1: 4-qt inflation with 12 thresholds, 6th in level @
@* E3: PR( 1.5% < 4qt MA dp < 3%) @
hat1 = a4sz[.,6];   @ n by 1 vec of CPI inflation forecasts @
{ flag1, flag3 } = EVALE1(hat1);  @ n by ?? & n by 1 @

@* E2: 4-qt output growth with 10 thresholds, 7th in 1st difference @
@* E5: Bad growth Pr(4-qt dy <2%) @
hat2 = a4sdz[.,7];   @ n by 1 vec of 4qt MA growth forecasts @
/* hat2 = hat2 + ones(nhorz,1).*popgrowth;  @ ?? plus pop growth @ */
{ flag2, flag5 } = EVALE2(hat2);  @ n by ?? @

@* E4: Recession: Pr(quarterly dy <0 and dy(-1) 0) @
inidy = DZT[7];    @ output growth at T @
hat3 = sdz[.,7];   @ quarterly @
/* inidy = DZT[7] + popgrowth;      @ total output growth at T @
hat3 = sdz[.,7]+ ones(nhorz,1).*popgrowth;   @ quarterly @ */
{ flag4 } = EVALE4(hat3,inidy); @ n by 1 @

@* E6: Joint event of no recession and inflation target @
{ flag6 } = EVALE6(hat1,hat3,inidy); @ n by 1 @

@* E7: Joint event of good growth and inflation target:
   Pr(4 qt MA dy > 2% and 1.5% < 4-qt dp < 3.5%) @
{ flag7 } = EVALE7(hat1,hat2); @ n by 1 @

RETP(flag1,flag2,flag3,flag4,flag5,flag6,flag7);ENDP;

@ *********************** @
  Proc (2) = EVALE1(hat);
@ ************************ 
* Prob evaluation of inflation < a% 
* ********************************* @
local n,thr,thr1,thr2,flag,ti,tj,flag1;

@* thresholds as % per annum @
thr = seqa(0,0.5,14); @ 0-6.5% @
thr1 = 1.5; thr2 = 3.5; @ between 1.5 and 3.5% @
thr = thr./400;  @ quarterly rates @
thr1 = thr1./400; thr2 = thr2./400;  
 
n = rows(hat);  @ forecast horizon @
flag = zeros(n,rows(thr));  @ number of thresholds @
flag1 = zeros(n,1);
ti = 1;
do until ti > n;
   tj = 1;
   do until tj > rows(thr);
      if hat[ti] < thr[tj];  flag[ti,tj] = 1;  @ pr(dp < a%) @
      endif;
      tj = tj + 1;
   endo;

@ pr(1.5% < dp < 3.5%) @
   if hat[ti] > thr1;  if hat[ti] < thr2; flag1[ti] = 1;
                       endif;
   endif;
   ti = ti + 1;
endo;

RETP(flag,flag1);ENDP;

@ *********************** @
  Proc (2) = EVALE2(hat);
@ ************************ 
* Prob evaluation of output growth < a% 
* ************************************* @
local n,thr,thr1,flag,flag1,ti,tj;

@* thresholds as % per annum @
thr = seqa(0,1,10); @ 0-9% @
thr1 = 2; @ below 2%: bad growth @ 
thr = thr./400;  @ quarterly rates @
thr1 = thr1./400;

n = rows(hat);  @ forecast horizon @
flag = zeros(n,rows(thr));  @ No. of thresholds @
flag1 = zeros(n,1);
ti = 1;
do until ti > n;
   tj = 1;
   do until tj > rows(thr);
      if hat[ti] < thr[tj];  flag[ti,tj] = 1;
      endif;
      tj = tj + 1;
   endo;
@ pr(g < 2%): bad growth @
   if hat[ti] < thr1;  flag1[ti] = 1;
   endif;
   ti = ti + 1;
endo;

RETP(flag,flag1);ENDP;

@ *************************** @
  Proc (1) = EVALE4(hat,dy0);
@ *************************** 
* Prob evaluation of recession 
* **************************** @
local n,thr,flag,ti;

thr = 0; thr = thr./400;  @ quarterly rate threshold @
n = rows(hat);  @ forecast horizon @
flag = zeros(n,1);
ti = 1;
do until ti > n;
   if ti eq 1;
      if hat[ti] < thr;
         if dy0 < thr;       flag[ti] = 1;
         endif;
      endif;
   else;
      if hat[ti] < thr;
         if hat[ti-1] < thr; flag[ti] = 1;
         endif;
      endif;
   endif;
   ti = ti + 1;
endo;

RETP(flag);ENDP;

@ ********************************* @
  Proc (1) = EVALE6(hat1,hat3,dy0);
@ ********************************* 
* E6: Joint event of no recession and inflation target 
* **************************************************** @
local n,thr1,thr3,flag,ti;

thr1 = zeros(2,1);  @ inflation percent per annum @
thr1[1] = 1.5; thr1[2] = 3.5;
thr3 = 0;           @ output growth @
thr1 = thr1./400; thr3 = thr3./400;   @ quarterly rates @
n = rows(hat1);  @ forecast horizon @
flag = zeros(n,1);
ti = 1;
do until ti > n;
   IF ti eq 1;
      if hat3[ti] < thr3;   @ dy < 0% @
         if dy0 < thr3;     @ dy(-1) < 0% @
         else;
            if hat1[ti] > thr1[1];     @ dp > 1.5% @
               if hat1[ti] < thr1[2];  @ dp < 3.5% @
                  flag[ti] = 1;
               endif;
            endif;
         endif;
      else;   @ dy > 0% @
         if hat1[ti] > thr1[1];     @ dp > 1.5% @
            if hat1[ti] < thr1[2];  @ dp < 3.5% @
               flag[ti] = 1;
            endif;
         endif;
      endif;
   ELSE;
      if hat3[ti] < thr3;       @ dy < 0% @
         if hat3[ti-1] < thr3;  @ dy(-1) < 0% @
         else;
            if hat1[ti] > thr1[1];     @ dp > 1.5% @
               if hat1[ti] < thr1[2];  @ dp < 3.5% @
                  flag[ti] = 1;
               endif;
            endif;
         endif;
      else;   @ dy > 0% @
         if hat1[ti] > thr1[1];     @ dp > 1.5% @
            if hat1[ti] < thr1[2];  @ dp < 3.5% @
               flag[ti] = 1;
            endif;
         endif;
      endif;
   ENDIF;
   ti = ti + 1;
endo;

RETP(flag);ENDP;

@ ***************************** @
  Proc (1) = EVALE7(hat1,hat2);
@ ***************************** 
* E7: Joint event of good growth and BOE inflation target:
* Pr(4 qt MA dy > 1% and 1.5% < 4-qt dp < 3.5%) 
* ********************************************* @
local n,thr1,thr2,flag,ti;

thr1 = zeros(2,1);  @ inflation percent per annum @
thr1[1] = 1.5; thr1[2] = 3.5;
thr2 = 1;           @ output growth @
thr1 = thr1./400; thr2 = thr2./400;   @ quarterly rates @
n = rows(hat1);  @ forecast horizon @
flag = zeros(n,1);
ti = 1;
do until ti > n;
   if hat2[ti] < thr2;
   else;  @ dy > 1% @
          if hat1[ti] > thr1[1];  @ dp > 1.5% @
             if hat1[ti] < thr1[2]; flag[ti] = 1;   @ dp < 3.5% @
             endif;
          endif;
   endif;
   ti = ti + 1;
endo;

RETP(flag);ENDP;

@ ********************************* @
  Proc (7) = EMPDIST(ForZ,nh,m,ns);
@ ********************************* @
@ Empirical mean, variance and CIs of central forecasts @
@ ***************************************************** @
Local nhm,mean,med,std,L90,U90,L95,U95,j,tmp;

nhm = nh*m;
mean = zeros(nhm,1); @ empirical mean: n*m by 1 @
med = zeros(nhm,1);  @ empirical median @
std = zeros(nhm,1);  @ empirical standard deviation @
L90 = zeros(nhm,1);  @ 90% lower confidence bound @
U90 = zeros(nhm,1);  @ 90% upper confidence bound @
L95 = zeros(nhm,1);  @ 95% lower confidence bound @
U95 = zeros(nhm,1);  @ 95% upper confidence bound @

j=1;
do until j > rows(ForZ);
   tmp = ForZ[j,.]';
   { mean[j], med[j], std[j], l90[j], u90[j], l95[j], u95[j] } 
   = EMPD(tmp,ns);
   j=j+1;
endo;

mean = TRMAT(mean,nh,m);
med = TRMAT(med,nh,m);
std = TRMAT(std,nh,m);
L90 = TRMAT(l90,nh,m);
U90 = TRMAT(u90,nh,m);
L95 = TRMAT(l95,nh,m);
U95 = TRMAT(u95,nh,m);

RETP(mean,med,std,L90,U90,L95,U95);ENDP;

@ ************************ @
  Proc (7) = EMPD(tmp,n1);
@ ************************ @
local mean,med,std,fr50,l90,u90,l95,u95,stmp;

l90 = n1*0.05;   l90 = ceil(l90);
u90 = n1*0.95;   u90 = floor(u90);
l95 = n1*0.025;  l95 = ceil(l95);
u95 = n1*0.975;  u95 = floor(u95);
fr50 = n1*0.5;   fr50 = round(fr50); 

mean = meanc(tmp);
std = stdc(tmp);

stmp = sortc(tmp,1);
l90 = stmp[l90];
u90 = stmp[u90];
l95 = stmp[l95];
u95 = stmp[u95];
med = stmp[fr50];

RETP(mean,med,std,l90,u90,l95,u95);ENDP;

@ ***************************** @
  Proc (1) = TRMAT(vect,ro,co);
@ ***************************** 
* Converting nm by 1 vector to n by m matrix 
* ****************************************** @
local matri,smj,rj,cj;

matri = zeros(ro,co);
smj = 1;
do until smj > co;
   rj = (smj-1)*ro + 1;
   cj = smj*ro;
   matri[.,smj] = vect[rj:cj];
   smj = smj + 1;
endo;

RETP(Matri);ENDP;

@ ************************************** @
  proc (3) = BCREG(Zin,DX,p,case,br,jj);
@ ************************************************
* CONSTRUCTION of ZS1 and all other regressors, YY
* in recursive bootstrap
* jj: t-index, br: break point index
* ********************************** @
Local  tmp1,tmp2,tmp3,tmp4,tmp5,tmp6,tmp7,tmp8,tmp9,tmp10,
       DZIN,Z0,DZ0,DZlags,jj0,id1,td1,did,dtd,ZS0,YY;

DZIN = ZIN-lag(ZIN);
Z0 = ZIN[rows(ZIN),.]; DZ0 = DZIN[rows(DZIN),.]; @ 1 by m @

if p eq 1;
elseif p eq 2; 
   DZlags = DZin;
elseif p eq 3;
   tmp1 = lag(DZIN); DZlags = DZIN~tmp1;
elseif p eq 4;
   tmp1 = lag(DZIN); tmp2 = lag(tmp1);
   DZlags = DZIN~tmp1~tmp2;
elseif p eq 5;
   tmp1=lag(DZIN); tmp2=lag(tmp1); tmp3=lag(tmp2);
   DZlags = DZIN~tmp1~tmp2~tmp3;
elseif p eq 6;
   tmp1=lag(DZIN); tmp2=lag(tmp1); tmp3=lag(tmp2); tmp4=lag(tmp3);
   DZlags = DZIN~tmp1~tmp2~tmp3~tmp4;
elseif p eq 7;
   tmp1=lag(DZIN); tmp2=lag(tmp1); tmp3=lag(tmp2); tmp4=lag(tmp3);
   tmp5=lag(tmp4); DZlags = DZIN~tmp1~tmp2~tmp3~tmp4~tmp5;
elseif p eq 8;
   tmp1=lag(DZIN); tmp2=lag(tmp1); tmp3=lag(tmp2); tmp4=lag(tmp3);
   tmp5=lag(tmp4); tmp6=lag(tmp5); 
   DZlags = DZIN~tmp1~tmp2~tmp3~tmp4~tmp5~tmp6;
elseif p eq 9;
   tmp1=lag(DZIN); tmp2=lag(tmp1); tmp3=lag(tmp2); tmp4=lag(tmp3);
   tmp5=lag(tmp4); tmp6=lag(tmp5); tmp7=lag(tmp6); 
   DZlags = DZIN~tmp1~tmp2~tmp3~tmp4~tmp5~tmp6~tmp7;
elseif p eq 10;
   tmp1=lag(DZIN); tmp2=lag(tmp1); tmp3=lag(tmp2); tmp4=lag(tmp3);
   tmp5=lag(tmp4); tmp6=lag(tmp5); tmp7=lag(tmp6); tmp8=lag(tmp7);
   DZlags = DZIN~tmp1~tmp2~tmp3~tmp4~tmp5~tmp6~tmp7~tmp8;
elseif p eq 11;
   tmp1=lag(DZIN); tmp2=lag(tmp1); tmp3=lag(tmp2); tmp4=lag(tmp3);
   tmp5=lag(tmp4); tmp6=lag(tmp5); tmp7=lag(tmp6); tmp8=lag(tmp7);
   tmp9=lag(tmp8);
   DZlags = DZIN~tmp1~tmp2~tmp3~tmp4~tmp5~tmp6~tmp7~tmp8~tmp9;
elseif p eq 12;
   tmp1=lag(DZIN); tmp2=lag(tmp1); tmp3=lag(tmp2); tmp4=lag(tmp3);
   tmp5=lag(tmp4); tmp6=lag(tmp5); tmp7=lag(tmp6); tmp8=lag(tmp7);
   tmp9=lag(tmp8);tmp10=lag(tmp9);
   DZlags = DZIN~tmp1~tmp2~tmp3~tmp4~tmp5~tmp6~tmp7~tmp8~tmp9~tmp10;
endif;
@* sample adjustment @
DZlags = DZlags[rows(DZlags),.]; @ 1 by m(p-1) @

jj0 = jj-1; @ t-1 @
id1 = 0; td1 = 0; @ intercept and tr dummies @
did = zeros(p,1); dtd = zeros(p,1); @ differences @
if jj < br; 
else; id1 = 1; td1 = 1+(jj-br); dtd[1] = 1;
endif;

if jj eq br; did[1] = 1;
elseif jj eq br+1; did[2] = 1; 
endif;

if jj < br+1; 
else; dtd[2] = 1;
endif;

IF     case eq 1; ZS0 = Z0'; 
Elseif case eq 2; ZS0 = Z0'|1;
Elseif case eq 3; ZS0 = Z0';
Elseif case eq 4; ZS0 = Z0'|jj0;
Elseif case eq 6; ZS0 = Z0'|jj0|id1;
Elseif case eq 7; ZS0 = Z0'|jj0|td1;
ENDIF;

@ DX: contemporaneous: mx by sample @

if p eq 1; YY = DX[.,jj];
else; YY = DZLAGS'|DX[.,jj];
endif;

IF     case eq 3; YY = YY|1;
ELSEIF case eq 4; YY = YY|1;
ELSEIF case eq 5; YY = YY|jj;
ELSEIF case eq 6; YY = YY|1|did;
ELSEIF case eq 7; YY = YY|1|dtd; @ ?? @
ENDIF;

Z0 = Z0'; @ m by 1 @
RETP(Z0,ZS0,YY);ENDP;


@* Long-run structural FIML estimation @ 

@ **************************************** @
  proc (1) = INICOIN(ms,r,bet,Rmat,bvec);
@ ****************************************
* Initial estimates of restricted cointegrating vectors:
* Transform Johansen beta using the whole restrictions
* ****************************************************** @
local  Q1,Q2,Ir,IQ2,theta0,beta0;

Ir = eye(rcoin);
Q1 = Ir.*.bet; Q2 = Rmat*Q1;
IQ2 = INV(Q2'Q2);
theta0 = Q1*IQ2*Q2'bvec;
beta0 = TRMAT(theta0,ms,rcoin); @ ms by r @

retp(beta0); endp;

@ ********************************************* @
  proc (3) = AOest(rbeta,DY,ZS1,YY,m,sample,r);
@ ********************************************* @
local  Ralpha,Romega,MYY,vDY,bzb,ibzb,bz,rkapa,rPAI,E0,MLL;

MYY = eye(sample) - YY'invpd(YY*YY')*YY; @ idempotent @
vDY = vec(DY');
bzb = rbeta'ZS1*MYY*ZS1'rbeta;
ibzb = inv(bzb); @ invpd ?? @
bz = rbeta'ZS1*MYY;
rkapa = eye(m).*.(ibzb*bz)*vDY;

ralpha = TRMAT(rkapa,r,m); @ r by m: @
ralpha = ralpha'; @ since kapa = vec(alpha') @

rPAI = ralpha*rbeta'; 
E0 = DY' - ZS1'rPAI';
rOmega = E0'MYY*E0./sample;

MLL = -(m*sample/2)*(1+ln(2*pi))-(sample/2)*ln(det(romega));

retp(Ralpha,Romega,MLL);endp;

@ ******************************************************** @
  proc (2) = Best(Ralpha,ROmega,DY,ZS1,YY,ms,r,Rmat,bvec);
@ ******************************************************** @
@* estimation of beta subject to restrictions @
local  Rbeta,U11,MYY,vDY,aoa,z1Mz1,Atheta,aom,z1M,Btheta,Jtheta,IJth,
       rth,rtheta;

MYY = eye(sample) - YY'invpd(YY*YY')*YY; @ idempotent @
vDY = vec(DY');
aoa = Ralpha'ROmega*Ralpha;
z1Mz1 = ZS1*MYY*ZS1';
Atheta = aoa.*.z1Mz1;
aom = Ralpha'ROmega;
z1M = ZS1*MYY;
Btheta = aom.*.z1M;
Jtheta = Atheta + Rmat'Rmat;
IJth = inv(Jtheta);
rth = IJth*Btheta*vDY;
rtheta = rth - IJth*Rmat'inv(Rmat*IJth*Rmat')*(Rmat*rth-bvec); 
rbeta = TRMAT(rtheta,ms,r); @ ms by r @
U11 = IJth - IJth*Rmat'Inv(Rmat*IJth*Rmat')*Rmat*IJth;

retp(Rbeta,U11);endp;


@ ************************ @
  proc (2) = RESTR1(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           0  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    *    *    *      t
* A total of 30 restrictions, so 14 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 30;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 30 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 0;  
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  Rmatrix[10,ms+10] = 1; bvec[10] = 0;    
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+4] = 1;  bvec[12] = 0; 
Rmatrix[13,ms+6] = 1;  bvec[13] = -1;
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[17,2*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,2*ms+2] = 1;  Rmatrix[18,2*ms+10] = 1; bvec[18] = 0;    
Rmatrix[19,2*ms+4] = 1;  bvec[19] = 0; 
Rmatrix[20,2*ms+5] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+6] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+9] = 1;  bvec[22] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[23,3*ms+1] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+2] = 1;  Rmatrix[24,3*ms+10] = 1; bvec[24] = 0;    
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;  
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;    
Rmatrix[27,3*ms+6] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+7] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+8] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+9] = 1;  bvec[30] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

RETP(RMatrix,bvec); ENDP;

@ ************************ @
  proc (2) = RESTR2(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           0  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    0    *    *      t
* A total of 31 restrictions, so 15 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 31;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 31 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 0;  
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 9 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  Rmatrix[10,ms+10] = 1; bvec[10] = 0;    
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+4] = 1;  bvec[12] = 0; 
Rmatrix[13,ms+6] = 1;  bvec[13] = -1;
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[18,2*ms+1] = 1;  bvec[18] = 0;  
Rmatrix[19,2*ms+2] = 1;  Rmatrix[19,2*ms+10] = 1; bvec[19] = 0;    
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[24,3*ms+1] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+2] = 1;  Rmatrix[25,3*ms+10] = 1; bvec[25] = 0;    
Rmatrix[26,3*ms+3] = 1;  bvec[26] = 0;  
Rmatrix[27,3*ms+5] = 1;  bvec[27] = -1;    
Rmatrix[28,3*ms+6] = 1;  bvec[28] = 0; 
Rmatrix[29,3*ms+7] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+8] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+9] = 1;  bvec[31] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR3(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           0  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    *    *    0      t
* A total of 31 restrictions, so 15 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 31;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 31 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 0;  
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  Rmatrix[10,ms+10] = 1; bvec[10] = 0;    
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+4] = 1;  bvec[12] = 0; 
Rmatrix[13,ms+6] = 1;  bvec[13] = -1;
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[17,2*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,2*ms+2] = 1;  Rmatrix[18,2*ms+10] = 1; bvec[18] = 0;    
Rmatrix[19,2*ms+4] = 1;  bvec[19] = 0; 
Rmatrix[20,2*ms+5] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+6] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+9] = 1;  bvec[22] = -1;

@ 4th vector: Inflation: 9 res @
Rmatrix[23,3*ms+1] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+2] = 1;  Rmatrix[24,3*ms+10] = 1; bvec[24] = 0;    
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;  
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;    
Rmatrix[27,3*ms+6] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+7] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+8] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+9] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+11] = 1;  bvec[31] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR4(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           0  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    0    *    0      t

* A total of 32 restrictions, so 16 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;

kk = 32;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 32 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 0;  
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 9 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  Rmatrix[10,ms+10] = 1; bvec[10] = 0;    
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+4] = 1;  bvec[12] = 0; 
Rmatrix[13,ms+6] = 1;  bvec[13] = -1;
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[18,2*ms+1] = 1;  bvec[18] = 0;  
Rmatrix[19,2*ms+2] = 1;  Rmatrix[19,2*ms+10] = 1; bvec[19] = 0;    
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 9 res @
Rmatrix[24,3*ms+1] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+2] = 1;  Rmatrix[25,3*ms+10] = 1; bvec[25] = 0;    
Rmatrix[26,3*ms+3] = 1;  bvec[26] = 0;  
Rmatrix[27,3*ms+5] = 1;  bvec[27] = -1;    
Rmatrix[28,3*ms+6] = 1;  bvec[28] = 0; 
Rmatrix[29,3*ms+7] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+8] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+9] = 1;  bvec[31] = 0;
Rmatrix[32,3*ms+11] = 1;  bvec[32] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR5(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           1  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    *    *    *      t
* A total of 30 restrictions, so 14 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 30;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 30 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;  
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  Rmatrix[10,ms+10] = 1; bvec[10] = 0;    
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+4] = 1;  bvec[12] = 0; 
Rmatrix[13,ms+6] = 1;  bvec[13] = -1;
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[17,2*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,2*ms+2] = 1;  Rmatrix[18,2*ms+10] = 1; bvec[18] = 0;    
Rmatrix[19,2*ms+4] = 1;  bvec[19] = 0; 
Rmatrix[20,2*ms+5] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+6] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+9] = 1;  bvec[22] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[23,3*ms+1] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+2] = 1;  Rmatrix[24,3*ms+10] = 1; bvec[24] = 0;    
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;  
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;    
Rmatrix[27,3*ms+6] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+7] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+8] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+9] = 1;  bvec[30] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR6(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           1  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    0    *    *      t
* A total of 31 restrictions, so 15 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 31;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 31 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;  
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 9 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  Rmatrix[10,ms+10] = 1; bvec[10] = 0;    
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+4] = 1;  bvec[12] = 0; 
Rmatrix[13,ms+6] = 1;  bvec[13] = -1;
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[18,2*ms+1] = 1;  bvec[18] = 0;  
Rmatrix[19,2*ms+2] = 1;  Rmatrix[19,2*ms+10] = 1; bvec[19] = 0;    
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[24,3*ms+1] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+2] = 1;  Rmatrix[25,3*ms+10] = 1; bvec[25] = 0;    
Rmatrix[26,3*ms+3] = 1;  bvec[26] = 0;  
Rmatrix[27,3*ms+5] = 1;  bvec[27] = -1;    
Rmatrix[28,3*ms+6] = 1;  bvec[28] = 0; 
Rmatrix[29,3*ms+7] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+8] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+9] = 1;  bvec[31] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR7(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           1  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    *    *    0      t
* A total of 31 restrictions, so 15 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 31;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 31 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;  
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  Rmatrix[10,ms+10] = 1; bvec[10] = 0;    
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+4] = 1;  bvec[12] = 0; 
Rmatrix[13,ms+6] = 1;  bvec[13] = -1;
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[17,2*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,2*ms+2] = 1;  Rmatrix[18,2*ms+10] = 1; bvec[18] = 0;    
Rmatrix[19,2*ms+4] = 1;  bvec[19] = 0; 
Rmatrix[20,2*ms+5] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+6] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+9] = 1;  bvec[22] = -1;

@ 4th vector: Inflation: 9 res @
Rmatrix[23,3*ms+1] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+2] = 1;  Rmatrix[24,3*ms+10] = 1; bvec[24] = 0;    
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;  
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;    
Rmatrix[27,3*ms+6] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+7] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+8] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+9] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+11] = 1;  bvec[31] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR8(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           1  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    0    *    0      t

* A total of 32 restrictions, so 16 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;

kk = 32;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 32 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;  
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 9 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  Rmatrix[10,ms+10] = 1; bvec[10] = 0;    
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+4] = 1;  bvec[12] = 0; 
Rmatrix[13,ms+6] = 1;  bvec[13] = -1;
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[18,2*ms+1] = 1;  bvec[18] = 0;  
Rmatrix[19,2*ms+2] = 1;  Rmatrix[19,2*ms+10] = 1; bvec[19] = 0;    
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 9 res @
Rmatrix[24,3*ms+1] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+2] = 1;  Rmatrix[25,3*ms+10] = 1; bvec[25] = 0;    
Rmatrix[26,3*ms+3] = 1;  bvec[26] = 0;  
Rmatrix[27,3*ms+5] = 1;  bvec[27] = -1;    
Rmatrix[28,3*ms+6] = 1;  bvec[28] = 0; 
Rmatrix[29,3*ms+7] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+8] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+9] = 1;  bvec[31] = 0;
Rmatrix[32,3*ms+11] = 1;  bvec[32] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR9(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    *    *    *      t
* A total of 29 restrictions, so 13 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 29;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 29 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+2] = 1;  Rmatrix[9,ms+10] = 1; bvec[9] = 0;    
Rmatrix[10,ms+3] = 1;  bvec[10] = 0;   
Rmatrix[11,ms+4] = 1;  bvec[11] = 0; 
Rmatrix[12,ms+6] = 1;  bvec[12] = -1;
Rmatrix[13,ms+7] = 1;  bvec[13] = 0;
Rmatrix[14,ms+8] = 1;  bvec[14] = 0;
Rmatrix[15,ms+9] = 1;  bvec[15] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[16,2*ms+1] = 1;  bvec[16] = 0;  
Rmatrix[17,2*ms+2] = 1;  Rmatrix[17,2*ms+10] = 1; bvec[17] = 0;    
Rmatrix[18,2*ms+4] = 1;  bvec[18] = 0; 
Rmatrix[19,2*ms+5] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+6] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+9] = 1;  bvec[21] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[22,3*ms+1] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+2] = 1;  Rmatrix[23,3*ms+10] = 1; bvec[23] = 0;    
Rmatrix[24,3*ms+3] = 1;  bvec[24] = 0;  
Rmatrix[25,3*ms+5] = 1;  bvec[25] = -1;    
Rmatrix[26,3*ms+6] = 1;  bvec[26] = 0; 
Rmatrix[27,3*ms+7] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+8] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+9] = 1;  bvec[29] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR10(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    0    *    *      t
* A total of 30 restrictions, so 14 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 30;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 30 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 9 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+2] = 1;  Rmatrix[9,ms+10] = 1; bvec[9] = 0;    
Rmatrix[10,ms+3] = 1;  bvec[10] = 0;   
Rmatrix[11,ms+4] = 1;  bvec[11] = 0; 
Rmatrix[12,ms+6] = 1;  bvec[12] = -1;
Rmatrix[13,ms+7] = 1;  bvec[13] = 0;
Rmatrix[14,ms+8] = 1;  bvec[14] = 0;
Rmatrix[15,ms+9] = 1;  bvec[15] = 0;
Rmatrix[16,ms+10] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[17,2*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,2*ms+2] = 1;  Rmatrix[18,2*ms+10] = 1; bvec[18] = 0;    
Rmatrix[19,2*ms+4] = 1;  bvec[19] = 0; 
Rmatrix[20,2*ms+5] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+6] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+9] = 1;  bvec[22] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[23,3*ms+1] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+2] = 1;  Rmatrix[24,3*ms+10] = 1; bvec[24] = 0;    
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;  
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;    
Rmatrix[27,3*ms+6] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+7] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+8] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+9] = 1;  bvec[30] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR11(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    *    *    0      t
* A total of 30 restrictions, so 14 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 30;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 30 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+2] = 1;  Rmatrix[9,ms+10] = 1; bvec[9] = 0;    
Rmatrix[10,ms+3] = 1;  bvec[10] = 0;   
Rmatrix[11,ms+4] = 1;  bvec[11] = 0; 
Rmatrix[12,ms+6] = 1;  bvec[12] = -1;
Rmatrix[13,ms+7] = 1;  bvec[13] = 0;
Rmatrix[14,ms+8] = 1;  bvec[14] = 0;
Rmatrix[15,ms+9] = 1;  bvec[15] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[16,2*ms+1] = 1;  bvec[16] = 0;  
Rmatrix[17,2*ms+2] = 1;  Rmatrix[17,2*ms+10] = 1; bvec[17] = 0;    
Rmatrix[18,2*ms+4] = 1;  bvec[18] = 0; 
Rmatrix[19,2*ms+5] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+6] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+9] = 1;  bvec[21] = -1;

@ 4th vector: Inflation: 9 res @
Rmatrix[22,3*ms+1] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+2] = 1;  Rmatrix[23,3*ms+10] = 1; bvec[23] = 0;    
Rmatrix[24,3*ms+3] = 1;  bvec[24] = 0;  
Rmatrix[25,3*ms+5] = 1;  bvec[25] = -1;    
Rmatrix[26,3*ms+6] = 1;  bvec[26] = 0; 
Rmatrix[27,3*ms+7] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+8] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+9] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+11] = 1;  bvec[30] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR12(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    0    *    0      t

* A total of 31 restrictions, so 15 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;

kk = 31;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 31 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 9 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+2] = 1;  Rmatrix[9,ms+10] = 1; bvec[9] = 0;    
Rmatrix[10,ms+3] = 1;  bvec[10] = 0;   
Rmatrix[11,ms+4] = 1;  bvec[11] = 0; 
Rmatrix[12,ms+6] = 1;  bvec[12] = -1;
Rmatrix[13,ms+7] = 1;  bvec[13] = 0;
Rmatrix[14,ms+8] = 1;  bvec[14] = 0;
Rmatrix[15,ms+9] = 1;  bvec[15] = 0;
Rmatrix[16,ms+10] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[17,2*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,2*ms+2] = 1;  Rmatrix[18,2*ms+10] = 1; bvec[18] = 0;    
Rmatrix[19,2*ms+4] = 1;  bvec[19] = 0; 
Rmatrix[20,2*ms+5] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+6] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+9] = 1;  bvec[22] = -1;

@ 4th vector: Inflation: 9 res @
Rmatrix[23,3*ms+1] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+2] = 1;  Rmatrix[24,3*ms+10] = 1; bvec[24] = 0;    
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;  
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;    
Rmatrix[27,3*ms+6] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+7] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+8] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+9] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+11] = 1;  bvec[31] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR13(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
           *    *    0   -1      inf2
           *   -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    *    *    *      t
* A total of 28 restrictions, so 12 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 28;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 28 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+2] = 1;  Rmatrix[8,ms+10] = 1; bvec[8] = 0;    
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;   
Rmatrix[10,ms+4] = 1;  bvec[10] = 0; 
Rmatrix[11,ms+6] = 1;  bvec[11] = -1;
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[15,2*ms+1] = 1;  bvec[15] = 0;  
Rmatrix[16,2*ms+2] = 1;  Rmatrix[16,2*ms+10] = 1; bvec[16] = 0;    
Rmatrix[17,2*ms+4] = 1;  bvec[17] = 0; 
Rmatrix[18,2*ms+5] = 1;  bvec[18] = 0;
Rmatrix[19,2*ms+6] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+9] = 1;  bvec[20] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[21,3*ms+1] = 1;  bvec[21] = 0;
Rmatrix[22,3*ms+2] = 1;  Rmatrix[22,3*ms+10] = 1; bvec[22] = 0;    
Rmatrix[23,3*ms+3] = 1;  bvec[23] = 0;  
Rmatrix[24,3*ms+5] = 1;  bvec[24] = -1;    
Rmatrix[25,3*ms+6] = 1;  bvec[25] = 0; 
Rmatrix[26,3*ms+7] = 1;  bvec[26] = 0;
Rmatrix[27,3*ms+8] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+9] = 1;  bvec[28] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR14(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
           *    *    0   -1      inf2
           *   -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    0    *    *      t
* A total of 29 restrictions, so 13 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 29;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 29 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 9 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+2] = 1;  Rmatrix[8,ms+10] = 1; bvec[8] = 0;    
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;   
Rmatrix[10,ms+4] = 1;  bvec[10] = 0; 
Rmatrix[11,ms+6] = 1;  bvec[11] = -1;
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;
Rmatrix[15,ms+10] = 1;  bvec[15] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[16,2*ms+1] = 1;  bvec[16] = 0;  
Rmatrix[17,2*ms+2] = 1;  Rmatrix[17,2*ms+10] = 1; bvec[17] = 0;    
Rmatrix[18,2*ms+4] = 1;  bvec[18] = 0; 
Rmatrix[19,2*ms+5] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+6] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+9] = 1;  bvec[21] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[22,3*ms+1] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+2] = 1;  Rmatrix[23,3*ms+10] = 1; bvec[23] = 0;    
Rmatrix[24,3*ms+3] = 1;  bvec[24] = 0;  
Rmatrix[25,3*ms+5] = 1;  bvec[25] = -1;    
Rmatrix[26,3*ms+6] = 1;  bvec[26] = 0; 
Rmatrix[27,3*ms+7] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+8] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+9] = 1;  bvec[29] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR15(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
           *    *    0   -1      inf2
           *   -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    *    *    0      t
* A total of 29 restrictions, so 13 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 29;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 29 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+2] = 1;  Rmatrix[8,ms+10] = 1; bvec[8] = 0;    
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;   
Rmatrix[10,ms+4] = 1;  bvec[10] = 0; 
Rmatrix[11,ms+6] = 1;  bvec[11] = -1;
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[15,2*ms+1] = 1;  bvec[15] = 0;  
Rmatrix[16,2*ms+2] = 1;  Rmatrix[16,2*ms+10] = 1; bvec[16] = 0;    
Rmatrix[17,2*ms+4] = 1;  bvec[17] = 0; 
Rmatrix[18,2*ms+5] = 1;  bvec[18] = 0;
Rmatrix[19,2*ms+6] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+9] = 1;  bvec[20] = -1;

@ 4th vector: Inflation: 9 res @
Rmatrix[21,3*ms+1] = 1;  bvec[21] = 0;
Rmatrix[22,3*ms+2] = 1;  Rmatrix[22,3*ms+10] = 1; bvec[22] = 0;    
Rmatrix[23,3*ms+3] = 1;  bvec[23] = 0;  
Rmatrix[24,3*ms+5] = 1;  bvec[24] = -1;    
Rmatrix[25,3*ms+6] = 1;  bvec[25] = 0; 
Rmatrix[26,3*ms+7] = 1;  bvec[26] = 0;
Rmatrix[27,3*ms+8] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+9] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+11] = 1;  bvec[29] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR16(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *  -B22 -B34 -B43     yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
           *    *    0   -1      inf2
           *   -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1   B22  B34  B43     y
           *    0    *    0      t

* A total of 30 restrictions, so 14 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;

kk = 30;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 30 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 9 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+2] = 1;  Rmatrix[8,ms+10] = 1; bvec[8] = 0;    
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;   
Rmatrix[10,ms+4] = 1;  bvec[10] = 0; 
Rmatrix[11,ms+6] = 1;  bvec[11] = -1;
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;
Rmatrix[15,ms+10] = 1;  bvec[15] = 0;

@ 3rd vector: Investment function: 6 res @
Rmatrix[16,2*ms+1] = 1;  bvec[16] = 0;  
Rmatrix[17,2*ms+2] = 1;  Rmatrix[17,2*ms+10] = 1; bvec[17] = 0;    
Rmatrix[18,2*ms+4] = 1;  bvec[18] = 0; 
Rmatrix[19,2*ms+5] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+6] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+9] = 1;  bvec[21] = -1;

@ 4th vector: Inflation: 9 res @
Rmatrix[22,3*ms+1] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+2] = 1;  Rmatrix[23,3*ms+10] = 1; bvec[23] = 0;    
Rmatrix[24,3*ms+3] = 1;  bvec[24] = 0;  
Rmatrix[25,3*ms+5] = 1;  bvec[25] = -1;    
Rmatrix[26,3*ms+6] = 1;  bvec[26] = 0; 
Rmatrix[27,3*ms+7] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+8] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+9] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+11] = 1;  bvec[30] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR17(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    *      yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    *      y
           *    *    *    *      t
* A total of 26 restrictions, so 10 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 26;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 26 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 7 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;   
Rmatrix[10,ms+4] = 1;  bvec[10] = 0; 
Rmatrix[11,ms+6] = 1;  bvec[11] = -1;
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;

@ 3rd vector: Investment function: 7 res @
Rmatrix[15,2*ms+1] = 1;  bvec[15] = 0;  
Rmatrix[16,2*ms+4] = 1;  bvec[16] = 0; 
Rmatrix[17,2*ms+5] = 1;  bvec[17] = 0;
Rmatrix[18,2*ms+6] = 1;  bvec[18] = 0;
Rmatrix[19,2*ms+9] = 1;  bvec[19] = -1;

@ 4th vector: Inflation: 7 res @
Rmatrix[20,3*ms+1] = 1;  bvec[20] = 0;
Rmatrix[21,3*ms+3] = 1;  bvec[21] = 0;  
Rmatrix[22,3*ms+5] = 1;  bvec[22] = -1;    
Rmatrix[23,3*ms+6] = 1;  bvec[23] = 0; 
Rmatrix[24,3*ms+7] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+8] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+9] = 1;  bvec[26] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR18(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    *      yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    *      y
           *    0    *    *      t
* A total of 27 restrictions, so 11 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 27;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 27 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;   
Rmatrix[10,ms+4] = 1;  bvec[10] = 0; 
Rmatrix[11,ms+6] = 1;  bvec[11] = -1;
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;
Rmatrix[15,ms+11] = 1;  bvec[15] = 0;

@ 3rd vector: Investment function: 7 res @
Rmatrix[16,2*ms+1] = 1;  bvec[16] = 0;  
Rmatrix[17,2*ms+4] = 1;  bvec[17] = 0; 
Rmatrix[18,2*ms+5] = 1;  bvec[18] = 0;
Rmatrix[19,2*ms+6] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+9] = 1;  bvec[20] = -1;

@ 4th vector: Inflation: 7 res @
Rmatrix[21,3*ms+1] = 1;  bvec[21] = 0;
Rmatrix[22,3*ms+3] = 1;  bvec[22] = 0;  
Rmatrix[23,3*ms+5] = 1;  bvec[23] = -1;    
Rmatrix[24,3*ms+6] = 1;  bvec[24] = 0; 
Rmatrix[25,3*ms+7] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+8] = 1;  bvec[26] = 0;
Rmatrix[27,3*ms+9] = 1;  bvec[27] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR19(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    *      yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    *      y
           *    *    *    0      t
* A total of 27 restrictions, so 11 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 27;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 27 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 7 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;   
Rmatrix[10,ms+4] = 1;  bvec[10] = 0; 
Rmatrix[11,ms+6] = 1;  bvec[11] = -1;
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;

@ 3rd vector: Investment function: 7 res @
Rmatrix[15,2*ms+1] = 1;  bvec[15] = 0;  
Rmatrix[16,2*ms+4] = 1;  bvec[16] = 0; 
Rmatrix[17,2*ms+5] = 1;  bvec[17] = 0;
Rmatrix[18,2*ms+6] = 1;  bvec[18] = 0;
Rmatrix[19,2*ms+9] = 1;  bvec[19] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[20,3*ms+1] = 1;  bvec[20] = 0;
Rmatrix[21,3*ms+3] = 1;  bvec[21] = 0;  
Rmatrix[22,3*ms+5] = 1;  bvec[22] = -1;    
Rmatrix[23,3*ms+6] = 1;  bvec[23] = 0; 
Rmatrix[24,3*ms+7] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+8] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+9] = 1;  bvec[26] = 0;
Rmatrix[27,3*ms+11] = 1;  bvec[27] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR20(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    *      yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    0      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    *      y
           *    0    *    0      t
* A total of 28 restrictions, so 12 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 28;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 28 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;   
Rmatrix[10,ms+4] = 1;  bvec[10] = 0; 
Rmatrix[11,ms+6] = 1;  bvec[11] = -1;
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;
Rmatrix[15,ms+11] = 1;  bvec[15] = 0;

@ 3rd vector: Investment function: 5 res @
Rmatrix[16,2*ms+1] = 1;  bvec[16] = 0;  
Rmatrix[17,2*ms+4] = 1;  bvec[17] = 0; 
Rmatrix[18,2*ms+5] = 1;  bvec[18] = 0;
Rmatrix[19,2*ms+6] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+9] = 1;  bvec[20] = -1;

@ 4th vector: Inflation: 8 res @
Rmatrix[21,3*ms+1] = 1;  bvec[21] = 0;
Rmatrix[22,3*ms+3] = 1;  bvec[22] = 0;  
Rmatrix[23,3*ms+5] = 1;  bvec[23] = -1;    
Rmatrix[24,3*ms+6] = 1;  bvec[24] = 0; 
Rmatrix[25,3*ms+7] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+8] = 1;  bvec[26] = 0;
Rmatrix[27,3*ms+9] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+11] = 1;  bvec[28] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;

@ ************************ @
  proc (2) = RESTR21(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    *      yscobb3
           *    *    *    *      tobq
           0    0    0    *      winfe
         -B11   *  -B31  -1      inf2
          B11  -1   B31   *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    *      y
           *    *    *    *      t
* A total of 21 restrictions, so 5 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 21;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 21 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,4] = 1;  bvec[2] = 0;
Rmatrix[3,5] = 1;  Rmatrix[3,6] = 1; bvec[3] = 0;    
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 6 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+4] = 1;  bvec[8] = 0;   
Rmatrix[9,ms+6] = 1;  bvec[9] = -1; 
Rmatrix[10,ms+7] = 1;  bvec[10] = 0;
Rmatrix[11,ms+8] = 1;  bvec[11] = 0;
Rmatrix[12,ms+9] = 1;  bvec[12] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[13,2*ms+1] = 1;  bvec[13] = 0;  
Rmatrix[14,2*ms+4] = 1;  bvec[14] = 0; 
Rmatrix[15,2*ms+5] = 1;  Rmatrix[15,2*ms+6] = 1; bvec[15] = 0;
Rmatrix[16,2*ms+9] = 1;  bvec[16] = -1;

@ 4th vector: Inflation: 5 res @
Rmatrix[17,3*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,3*ms+5] = 1;  bvec[18] = -1;  
Rmatrix[19,3*ms+7] = 1;  bvec[19] = 0;    
Rmatrix[20,3*ms+8] = 1;  bvec[20] = 0; 
Rmatrix[21,3*ms+9] = 1;  bvec[21] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR22(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    *      yscobb3
           *    *    *    *      tobq
           0    0    0    *      winfe
         -B11   *  -B31  -1      inf2
          B11  -1   B31   *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    *      y
           *    0    *    *      t
* A total of 22 restrictions, so 6 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 22;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 22 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,4] = 1;  bvec[2] = 0;
Rmatrix[3,5] = 1;  Rmatrix[3,6] = 1; bvec[3] = 0;    
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+4] = 1;  bvec[8] = 0;   
Rmatrix[9,ms+6] = 1;  bvec[9] = -1; 
Rmatrix[10,ms+7] = 1;  bvec[10] = 0;
Rmatrix[11,ms+8] = 1;  bvec[11] = 0;
Rmatrix[12,ms+9] = 1;  bvec[12] = 0;
Rmatrix[13,ms+11] = 1;  bvec[13] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[14,2*ms+1] = 1;  bvec[14] = 0;  
Rmatrix[15,2*ms+4] = 1;  bvec[15] = 0; 
Rmatrix[16,2*ms+5] = 1;  Rmatrix[16,2*ms+6] = 1; bvec[16] = 0;
Rmatrix[17,2*ms+9] = 1;  bvec[17] = -1;

@ 4th vector: Inflation: 5 res @
Rmatrix[18,3*ms+1] = 1;  bvec[18] = 0;  
Rmatrix[19,3*ms+5] = 1;  bvec[19] = -1;  
Rmatrix[20,3*ms+7] = 1;  bvec[20] = 0;    
Rmatrix[21,3*ms+8] = 1;  bvec[21] = 0; 
Rmatrix[22,3*ms+9] = 1;  bvec[22] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR23(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    *      yscobb3
           *    *    *    *      tobq
           0    0    0    *      winfe
         -B11   *  -B31  -1      inf2
          B11  -1   B31   *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    *      y
           *    *    *    0      t
* A total of 22 restrictions, so 6 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 22;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 22 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,4] = 1;  bvec[2] = 0;
Rmatrix[3,5] = 1;  Rmatrix[3,6] = 1; bvec[3] = 0;    
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 7 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+4] = 1;  bvec[8] = 0;   
Rmatrix[9,ms+6] = 1;  bvec[9] = -1; 
Rmatrix[10,ms+7] = 1;  bvec[10] = 0;
Rmatrix[11,ms+8] = 1;  bvec[11] = 0;
Rmatrix[12,ms+9] = 1;  bvec[12] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[13,2*ms+1] = 1;  bvec[13] = 0;  
Rmatrix[14,2*ms+4] = 1;  bvec[14] = 0; 
Rmatrix[15,2*ms+5] = 1;  Rmatrix[15,2*ms+6] = 1; bvec[15] = 0;
Rmatrix[16,2*ms+9] = 1;  bvec[16] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[17,3*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,3*ms+5] = 1;  bvec[18] = -1;  
Rmatrix[19,3*ms+7] = 1;  bvec[19] = 0;    
Rmatrix[20,3*ms+8] = 1;  bvec[20] = 0; 
Rmatrix[21,3*ms+9] = 1;  bvec[21] = 0;
Rmatrix[22,3*ms+11] = 1;  bvec[22] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR24(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    *      yscobb3
           *    *    *    *      tobq
           0    0    0    *      winfe
         -B11   *  -B31  -1      inf2
          B11  -1   B31   *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    *      y
           *    0    *    0      t
* A total of 23 restrictions, so 7 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 23;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 23 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,4] = 1;  bvec[2] = 0;
Rmatrix[3,5] = 1;  Rmatrix[3,6] = 1; bvec[3] = 0;    
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+4] = 1;  bvec[8] = 0;   
Rmatrix[9,ms+6] = 1;  bvec[9] = -1; 
Rmatrix[10,ms+7] = 1;  bvec[10] = 0;
Rmatrix[11,ms+8] = 1;  bvec[11] = 0;
Rmatrix[12,ms+9] = 1;  bvec[12] = 0;
Rmatrix[13,ms+11] = 1;  bvec[13] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[14,2*ms+1] = 1;  bvec[14] = 0;  
Rmatrix[15,2*ms+4] = 1;  bvec[15] = 0; 
Rmatrix[16,2*ms+5] = 1;  Rmatrix[16,2*ms+6] = 1; bvec[16] = 0;
Rmatrix[17,2*ms+9] = 1;  bvec[17] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[18,3*ms+1] = 1;  bvec[18] = 0;  
Rmatrix[19,3*ms+5] = 1;  bvec[19] = -1;  
Rmatrix[20,3*ms+7] = 1;  bvec[20] = 0;    
Rmatrix[21,3*ms+8] = 1;  bvec[21] = 0; 
Rmatrix[22,3*ms+9] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+11] = 1;  bvec[23] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR25(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    0    *    0      yscobb3
           *    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    0    *    0      t
* A total of 31 restrictions, so 15 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 31;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 31 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,4] = 1;  bvec[2] = 0;
Rmatrix[3,5] = 1;  Rmatrix[3,6] = 1; bvec[3] = 0;    
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+2] = 1;  bvec[8] = 0;
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;
Rmatrix[10,ms+4] = 1;  bvec[10] = 0;   
Rmatrix[11,ms+6] = 1;  bvec[11] = -1; 
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;
Rmatrix[15,ms+10] = 1;  bvec[15] = 0;
Rmatrix[16,ms+11] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[17,2*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,2*ms+4] = 1;  bvec[18] = 0; 
Rmatrix[19,2*ms+5] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+6] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+9] = 1;  bvec[21] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[22,3*ms+1] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+2] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+3] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+4] = 1;  bvec[25] = 1;
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;  
Rmatrix[27,3*ms+7] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+8] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+9] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+10] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+11] = 1;  bvec[31] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR26(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    0    *    0      yscobb3
           0    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    0    *    0      t
* A total of 32 restrictions, so 16 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 32;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 32 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+2] = 1;  bvec[9] = 0;
Rmatrix[10,ms+3] = 1;  bvec[10] = 0;
Rmatrix[11,ms+4] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+6] = 1;  bvec[12] = -1; 
Rmatrix[13,ms+7] = 1;  bvec[13] = 0;
Rmatrix[14,ms+8] = 1;  bvec[14] = 0;
Rmatrix[15,ms+9] = 1;  bvec[15] = 0;
Rmatrix[16,ms+10] = 1;  bvec[16] = 0;
Rmatrix[17,ms+11] = 1;  bvec[17] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[18,2*ms+1] = 1;  bvec[18] = 0;  
Rmatrix[19,2*ms+4] = 1;  bvec[19] = 0; 
Rmatrix[20,2*ms+5] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+6] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+9] = 1;  bvec[22] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[23,3*ms+1] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+2] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+4] = 1;  bvec[26] = 1;
Rmatrix[27,3*ms+5] = 1;  bvec[27] = -1;  
Rmatrix[28,3*ms+7] = 1;  bvec[28] = 0; 
Rmatrix[29,3*ms+8] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+9] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+10] = 1;  bvec[31] = 0;
Rmatrix[32,3*ms+11] = 1;  bvec[32] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR27(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    0    *    0      yscobb3
           0    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    *    *    0      t
* A total of 31 restrictions, so 15 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 31;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 31 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+2] = 1;  bvec[9] = 0;
Rmatrix[10,ms+3] = 1;  bvec[10] = 0;
Rmatrix[11,ms+4] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+6] = 1;  bvec[12] = -1; 
Rmatrix[13,ms+7] = 1;  bvec[13] = 0;
Rmatrix[14,ms+8] = 1;  bvec[14] = 0;
Rmatrix[15,ms+9] = 1;  bvec[15] = 0;
Rmatrix[16,ms+10] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[17,2*ms+1] = 1;  bvec[17] = 0;  
Rmatrix[18,2*ms+4] = 1;  bvec[18] = 0; 
Rmatrix[19,2*ms+5] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+6] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+9] = 1;  bvec[21] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[22,3*ms+1] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+2] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+3] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+4] = 1;  bvec[25] = 1;
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;  
Rmatrix[27,3*ms+7] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+8] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+9] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+10] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+11] = 1;  bvec[31] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR28(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    0    *    0      yscobb3
           0    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    0    *    *      t
* A total of 31 restrictions, so 15 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 31;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 31 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+2] = 1;  bvec[9] = 0;
Rmatrix[10,ms+3] = 1;  bvec[10] = 0;
Rmatrix[11,ms+4] = 1;  bvec[11] = 0;   
Rmatrix[12,ms+6] = 1;  bvec[12] = -1; 
Rmatrix[13,ms+7] = 1;  bvec[13] = 0;
Rmatrix[14,ms+8] = 1;  bvec[14] = 0;
Rmatrix[15,ms+9] = 1;  bvec[15] = 0;
Rmatrix[16,ms+10] = 1;  bvec[16] = 0;
Rmatrix[17,ms+11] = 1;  bvec[17] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[18,2*ms+1] = 1;  bvec[18] = 0;  
Rmatrix[19,2*ms+4] = 1;  bvec[19] = 0; 
Rmatrix[20,2*ms+5] = 1;  bvec[20] = 0;
Rmatrix[21,2*ms+6] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+9] = 1;  bvec[22] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[23,3*ms+1] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+2] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+4] = 1;  bvec[26] = 1;
Rmatrix[27,3*ms+5] = 1;  bvec[27] = -1;  
Rmatrix[28,3*ms+7] = 1;  bvec[28] = 0; 
Rmatrix[29,3*ms+8] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+9] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+10] = 1;  bvec[31] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR29(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           *    *    *    0      yscobb3
           0    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    *    *    0      y
           *    0    *    *      t
* A total of 31 restrictions, so 13 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 29;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 29 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 7 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,3] = 1;  bvec[2] = 0;
Rmatrix[3,4] = 1;  bvec[3] = 0;
Rmatrix[4,5] = 1;  Rmatrix[4,6] = 1; bvec[4] = 0;    
Rmatrix[5,7] = 1;  bvec[5] = 0; 
Rmatrix[6,8] = 1;  bvec[6] = 0;
Rmatrix[7,10] = 1;  bvec[7] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[8,ms+1] = 1;  bvec[8] = 0;
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;
Rmatrix[10,ms+4] = 1;  bvec[10] = 0;   
Rmatrix[11,ms+6] = 1;  bvec[11] = -1; 
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;
Rmatrix[15,ms+11] = 1;  bvec[15] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[16,2*ms+1] = 1;  bvec[16] = 0;  
Rmatrix[17,2*ms+4] = 1;  bvec[17] = 0; 
Rmatrix[18,2*ms+5] = 1;  bvec[18] = 0;
Rmatrix[19,2*ms+6] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+9] = 1;  bvec[20] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[21,3*ms+1] = 1;  bvec[21] = 0;
Rmatrix[22,3*ms+2] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+3] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+4] = 1;  bvec[24] = 1;
Rmatrix[25,3*ms+5] = 1;  bvec[25] = -1;  
Rmatrix[26,3*ms+7] = 1;  bvec[26] = 0; 
Rmatrix[27,3*ms+8] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+9] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+10] = 1;  bvec[29] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR30(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           1    0    *    0      yscobb3
           0    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    0    *    *      t
* A total of 32 restrictions, so 16 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 32;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 32 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  bvec[10] = 0;
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;
Rmatrix[12,ms+4] = 1;  bvec[12] = 0;   
Rmatrix[13,ms+6] = 1;  bvec[13] = -1; 
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;
Rmatrix[18,ms+11] = 1;  bvec[18] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[19,2*ms+1] = 1;  bvec[19] = 0;  
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[24,3*ms+1] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+2] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+3] = 1;  bvec[26] = 0;
Rmatrix[27,3*ms+4] = 1;  bvec[27] = 1;
Rmatrix[28,3*ms+5] = 1;  bvec[28] = -1;  
Rmatrix[29,3*ms+7] = 1;  bvec[29] = 0; 
Rmatrix[30,3*ms+8] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+9] = 1;  bvec[31] = 0;
Rmatrix[32,3*ms+10] = 1;  bvec[32] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR31(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    *      poil 
           1    0    *    0      yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    0    *    *      t
* A total of 30 restrictions, so 14 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 30;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 30 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  bvec[10] = 0;
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;
Rmatrix[12,ms+4] = 1;  bvec[12] = 0;   
Rmatrix[13,ms+6] = 1;  bvec[13] = -1; 
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;
Rmatrix[18,ms+11] = 1;  bvec[18] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[19,2*ms+1] = 1;  bvec[19] = 0;  
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[24,3*ms+2] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;  
Rmatrix[27,3*ms+7] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+8] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+9] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+10] = 1;  bvec[30] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR32(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           1    0    *    *      yscobb3
           0    0    *    0      tobq
           0    0    0    *      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    *      y
           *    0    *    *      t
* A total of 29 restrictions, so 13 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 29;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 29 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  bvec[10] = 0;
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;
Rmatrix[12,ms+4] = 1;  bvec[12] = 0;   
Rmatrix[13,ms+6] = 1;  bvec[13] = -1; 
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;
Rmatrix[18,ms+11] = 1;  bvec[18] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[19,2*ms+1] = 1;  bvec[19] = 0;  
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[24,3*ms+1] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+3] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+5] = 1;  bvec[26] = -1;  
Rmatrix[27,3*ms+7] = 1;  bvec[27] = 0; 
Rmatrix[28,3*ms+8] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+9] = 1;  bvec[29] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR33(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    *    0      poil 
           *    0    *    0      yscobb3
           *    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    0    *    0      t
* A total of 30 restrictions, so 14 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 30;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 30 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 6 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,4] = 1;  bvec[2] = 0;
Rmatrix[3,5] = 1;  Rmatrix[3,6] = 1; bvec[3] = 0;    
Rmatrix[4,7] = 1;  bvec[4] = 0; 
Rmatrix[5,8] = 1;  bvec[5] = 0;
Rmatrix[6,10] = 1;  bvec[6] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[7,ms+1] = 1;  bvec[7] = 0;
Rmatrix[8,ms+2] = 1;  bvec[8] = 0;
Rmatrix[9,ms+3] = 1;  bvec[9] = 0;
Rmatrix[10,ms+4] = 1;  bvec[10] = 0;   
Rmatrix[11,ms+6] = 1;  bvec[11] = -1; 
Rmatrix[12,ms+7] = 1;  bvec[12] = 0;
Rmatrix[13,ms+8] = 1;  bvec[13] = 0;
Rmatrix[14,ms+9] = 1;  bvec[14] = 0;
Rmatrix[15,ms+10] = 1;  bvec[15] = 0;
Rmatrix[16,ms+11] = 1;  bvec[16] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[17,2*ms+4] = 1;  bvec[17] = 0; 
Rmatrix[18,2*ms+5] = 1;  bvec[18] = 0;
Rmatrix[19,2*ms+6] = 1;  bvec[19] = 0;
Rmatrix[20,2*ms+9] = 1;  bvec[20] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[21,3*ms+1] = 1;  bvec[21] = 0;
Rmatrix[22,3*ms+2] = 1;  bvec[22] = 0;
Rmatrix[23,3*ms+3] = 1;  bvec[23] = 0;
Rmatrix[24,3*ms+4] = 1;  bvec[24] = 1;
Rmatrix[25,3*ms+5] = 1;  bvec[25] = -1;  
Rmatrix[26,3*ms+7] = 1;  bvec[26] = 0; 
Rmatrix[27,3*ms+8] = 1;  bvec[27] = 0;
Rmatrix[28,3*ms+9] = 1;  bvec[28] = 0;
Rmatrix[29,3*ms+10] = 1;  bvec[29] = 0;
Rmatrix[30,3*ms+11] = 1;  bvec[30] = 0;


dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;


@ ************************ @
  proc (2) = RESTR34(ms,r);
@ ************************ @
@* e.g. m = 10 (+1, trend) and rcoin = 4
  We have the following full theory restrictions:
  beta = { 0    0    0    0      poil 
           1    0    *    0      yscobb3
           0    0    *    0      tobq
           0    0    0    1      winfe
         -B11   *    0   -1      inf2
          B11  -1    0    *      rus
           0    0    *    0      dscr2
           0    0    *    0      cf2
           *    0   -1    0      invest
          -1    0    *    0      y
           *    0    *    0      t
* A total of 33 restrictions, so 17 over-identifying restrictions
* Later consider co-trending and co-breaking  @
local RMatrix,bvec,kk;
        
kk = 33;
Rmatrix = zeros(kk,ms*rcoin); @ k by msr known matrix: 33 by 44 @
bvec = zeros(kk,1); @ k by 1 known vector @

@ 1st vector: Aggregate Demand: 8 res @
Rmatrix[1,1] = 1;  bvec[1] = 0;
Rmatrix[2,2] = 1;  bvec[2] = 1;
Rmatrix[3,3] = 1;  bvec[3] = 0;
Rmatrix[4,4] = 1;  bvec[4] = 0;
Rmatrix[5,5] = 1;  Rmatrix[5,6] = 1; bvec[5] = 0;    
Rmatrix[6,7] = 1;  bvec[6] = 0; 
Rmatrix[7,8] = 1;  bvec[7] = 0;
Rmatrix[8,10] = 1;  bvec[8] = -1;
    
@ 2nd vector: Taylor rule: 8 res @
Rmatrix[9,ms+1] = 1;  bvec[9] = 0;
Rmatrix[10,ms+2] = 1;  bvec[10] = 0;
Rmatrix[11,ms+3] = 1;  bvec[11] = 0;
Rmatrix[12,ms+4] = 1;  bvec[12] = 0;   
Rmatrix[13,ms+6] = 1;  bvec[13] = -1; 
Rmatrix[14,ms+7] = 1;  bvec[14] = 0;
Rmatrix[15,ms+8] = 1;  bvec[15] = 0;
Rmatrix[16,ms+9] = 1;  bvec[16] = 0;
Rmatrix[17,ms+10] = 1;  bvec[17] = 0;
Rmatrix[18,ms+11] = 1;  bvec[18] = 0;

@ 3rd vector: Investment function: 4 res @
Rmatrix[19,2*ms+1] = 1;  bvec[19] = 0;  
Rmatrix[20,2*ms+4] = 1;  bvec[20] = 0; 
Rmatrix[21,2*ms+5] = 1;  bvec[21] = 0;
Rmatrix[22,2*ms+6] = 1;  bvec[22] = 0;
Rmatrix[23,2*ms+9] = 1;  bvec[23] = -1;

@ 4th vector: Inflation: 6 res @
Rmatrix[24,3*ms+1] = 1;  bvec[24] = 0;
Rmatrix[25,3*ms+2] = 1;  bvec[25] = 0;
Rmatrix[26,3*ms+3] = 1;  bvec[26] = 0;
Rmatrix[27,3*ms+4] = 1;  bvec[27] = 1;
Rmatrix[28,3*ms+5] = 1;  bvec[28] = -1;  
Rmatrix[29,3*ms+7] = 1;  bvec[29] = 0; 
Rmatrix[30,3*ms+8] = 1;  bvec[30] = 0;
Rmatrix[31,3*ms+9] = 1;  bvec[31] = 0;
Rmatrix[32,3*ms+10] = 1;  bvec[32] = 0;
Rmatrix[33,3*ms+11] = 1;  bvec[33] = 0;

dimR = rows(Rmatrix); OVDOF = dimR - rcoin^2;

retp(RMatrix,bvec); endp;
